/*
 * Copyright 1999-2006 University of Chicago
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 *  user functions.  used by control.c or DSI implementation if it choses.
 */
#include "globus_i_gridftp_server.h"
#include <grp.h>

#define FTP_SERVICE_NAME "file"

static
char * 
globus_gfs_acl_cas_action_to_string(
    globus_gfs_acl_action_t             action)
{
    switch(action)
    {
        case GFS_ACL_ACTION_DELETE:
            return "delete";
        case GFS_ACL_ACTION_WRITE:
            return "write";
        case GFS_ACL_ACTION_CREATE:
            return "create";
        case GFS_ACL_ACTION_READ:
            return "read";
        case GFS_ACL_ACTION_LOOKUP:
            return "lookup";
        case GFS_ACL_ACTION_AUTHZ_ASSERT:
            return "authz_assert";
        default:
            return NULL;
    }
}

static
void
globus_gfs_acl_cas_cb(
    void *                              callback_arg,
    globus_gsi_authz_handle_t           handle,
    globus_result_t                     result)
{
    globus_i_gfs_acl_handle_t *         acl_handle;
    GlobusGFSName(globus_gfs_acl_cas_cb);
    GlobusGFSDebugEnter();

    acl_handle = (globus_i_gfs_acl_handle_t *) callback_arg;
    globus_gfs_acl_authorized_finished(acl_handle, result);

    GlobusGFSDebugExit();
}

static
int
globus_gfs_acl_cas_init(
    void **                             out_handle,
    globus_gfs_acl_info_t *             acl_info,
    globus_gfs_acl_handle_t             acl_handle,
    globus_result_t *                   out_res)
{
    globus_gsi_authz_handle_t           cas_handle;
    GlobusGFSName(globus_gfs_acl_cas_init);
    GlobusGFSDebugEnter();

    if(acl_info->context == NULL)
    {
        goto err;
    }
    *out_res = globus_gsi_authz_handle_init(
        &cas_handle,
        FTP_SERVICE_NAME,
        acl_info->context,
        globus_gfs_acl_cas_cb,
        acl_handle);
    if(*out_res != GLOBUS_SUCCESS)
    {
        goto err;
    }
    *out_handle = cas_handle;
    GlobusGFSDebugExit();
    return GLOBUS_GFS_ACL_WOULD_BLOCK;

  err:
    GlobusGFSDebugExitWithError();
    return GLOBUS_GFS_ACL_COMPLETE;
}


static
int
mdtm_group_find(
    gid_t       gid,
    gid_t *     groups,
    int         ngroups)
{
  int i, found = 0;

  for(i = 0; i  < ngroups; i++) {
      if(groups[i] == gid) {
        found = 1;
        break;
      }
  }
  return found;
}

#define MDTM_GROUPS_DEFAULT             16

static
globus_result_t
mdtm_gfs_pwfile_authorize(
    const char *                        name,
    globus_gfs_acl_action_t             action,
    globus_gfs_acl_info_t *             acl_info,
    globus_gfs_acl_handle_t             acl_handle
    )
{
  struct stat           sb;
  uid_t                 f_uid;
  gid_t                 f_gid;
  struct passwd *       pwd;
  gid_t *               groups;
  int                   n, ngroups;
  globus_result_t       rc = GLOBUS_FAILURE;
  int                   result = -1;
  char *                path;

  // get file's owner and group
  if((path = strdup(name)) == NULL)
    return GLOBUS_FAILURE;

  while(result < 0) {
      result = lstat(path, &sb);
      if(result < 0) {
          if(errno == ENOENT && action == GFS_ACL_ACTION_CREATE) {
              if(strlen(path) > 0 && strcmp(path, "/")) {
                  char * p;
                  if((p = strrchr(path, '/')) != NULL) {
                      *p = '\0';
                      continue;
                  }
              }
          }
          free(path);
          return GLOBUS_FAILURE;
      }
  }
  free(path);
  f_uid = sb.st_uid;
  f_gid = sb.st_gid;

  // get user's id
  if((pwd = getpwnam(acl_info->username)) == NULL)
    return GLOBUS_FAILURE;
  if(pwd->pw_uid == f_uid) {
      // user is the owner of the file
      if(action == GFS_ACL_ACTION_READ ||
         action == GFS_ACL_ACTION_LOOKUP) {
          if(sb.st_mode & S_IRUSR)
            rc = GLOBUS_SUCCESS;
          else
            rc = GLOBUS_FAILURE;
      }
      else if (action == GFS_ACL_ACTION_CREATE ||
               action == GFS_ACL_ACTION_WRITE ||
               action == GFS_ACL_ACTION_DELETE) {
          if(sb.st_mode & S_IWUSR)
            rc =GLOBUS_SUCCESS;
          else
            rc = GLOBUS_FAILURE;
      }
      return rc;
  }

  // get user's groups
  n = MDTM_GROUPS_DEFAULT;
  if((groups = (gid_t *) globus_malloc(n * sizeof(gid_t))) == NULL)
      return GLOBUS_FAILURE;
  ngroups = getgrouplist(acl_info->username,
      pwd->pw_gid,
      groups,
      &n);
  if(ngroups < 0) {
      groups = (gid_t *) globus_realloc(groups, n * sizeof(gid_t));
      if(groups)
        ngroups = getgrouplist(acl_info->username,
          pwd->pw_gid,
          groups,
          &n);
      if(groups == NULL || ngroups < 0) {
          if(groups) globus_free(groups);
          return GLOBUS_FAILURE;
      }
  }
  if(pwd->pw_gid == f_gid || mdtm_group_find(f_gid, groups, ngroups)) {
      // user belongs to the group of file
      if(action == GFS_ACL_ACTION_READ ||
         action == GFS_ACL_ACTION_LOOKUP) {
          if(sb.st_mode & S_IRGRP)
            rc = GLOBUS_SUCCESS;
          else
            rc = GLOBUS_FAILURE;
      }
      else if (action == GFS_ACL_ACTION_CREATE ||
               action == GFS_ACL_ACTION_WRITE ||
               action == GFS_ACL_ACTION_DELETE) {
          if(sb.st_mode & S_IWGRP)
            rc =GLOBUS_SUCCESS;
          else
            rc = GLOBUS_FAILURE;
      }
  }
  else {
      // user is among the others
      if(action == GFS_ACL_ACTION_READ ||
         action == GFS_ACL_ACTION_LOOKUP) {
          if(sb.st_mode & S_IROTH)
            rc = GLOBUS_SUCCESS;
          else
            rc = GLOBUS_FAILURE;
      }
      else if (action == GFS_ACL_ACTION_CREATE ||
               action == GFS_ACL_ACTION_WRITE ||
               action == GFS_ACL_ACTION_DELETE) {
          if(sb.st_mode & S_IWOTH)
            rc =GLOBUS_SUCCESS;
          else
            rc = GLOBUS_FAILURE;
      }
  }
  globus_free(groups);
  return rc;
}

static
int
globus_gfs_acl_cas_authorize(
    void *                              out_handle,
    globus_gfs_acl_action_t             action,
    globus_gfs_acl_object_desc_t *      object,
    globus_gfs_acl_info_t *             acl_info,
    globus_gfs_acl_handle_t             acl_handle,
    globus_result_t *                   out_res)
{
    globus_gsi_authz_handle_t           cas_handle;
    char *                              full_object;
    char *                              action_str;
    GlobusGFSName(globus_gfs_acl_cas_authorize);
    GlobusGFSDebugEnter();

    cas_handle = (globus_gsi_authz_handle_t) out_handle;
    if(acl_info->context == NULL)
    {
        // LZ: mdtmFTP authorization using pwd file
        *out_res = mdtm_gfs_pwfile_authorize(
            object->name,
            action,
            acl_info,
            acl_handle);
        // end: mdtmFTP authorization using pwd file
        goto err;
    }

    action_str = globus_gfs_acl_cas_action_to_string(action);
    if(!action_str)
    {
        *out_res = GLOBUS_SUCCESS;
        GlobusGFSDebugExit();
        return GLOBUS_GFS_ACL_COMPLETE;
    }
    
    /*
     * If the action is "authz_assert" then the object contains the assertions
     * received over the gridftp control channel - just pass it unmodified to
     * the authz callout
     */
    if(action == GFS_ACL_ACTION_AUTHZ_ASSERT)
    {
        full_object = globus_libc_strdup(object->name);
    }
    else
    {
        full_object = globus_common_create_string(
            "ftp://%s%s", acl_info->hostname, object->name);
    }    

    *out_res = globus_gsi_authorize(
        cas_handle,
        action_str,
        full_object,
        globus_gfs_acl_cas_cb,
        acl_handle);
    globus_free(full_object);
    if(*out_res != GLOBUS_SUCCESS)
    {
        goto err;
    }

    GlobusGFSDebugExit();
    return GLOBUS_GFS_ACL_WOULD_BLOCK;

  err:
    GlobusGFSDebugExitWithError();
    return GLOBUS_GFS_ACL_COMPLETE;
}

static
void
globus_gfs_acl_cas_destroy_cb(
    void *                              callback_arg,
    globus_gsi_authz_handle_t           handle,
    globus_result_t                     result)
{
    GlobusGFSName(globus_gfs_acl_cas_destroy_cb);
    GlobusGFSDebugEnter();


    GlobusGFSDebugExit();
}

static
void
globus_gfs_acl_cas_destroy(
    void *                              out_handle)
{
    globus_gsi_authz_handle_t           cas_handle;
    GlobusGFSName(globus_gfs_acl_cas_destroy);
    GlobusGFSDebugEnter();

    cas_handle = (globus_gsi_authz_handle_t) out_handle;
    globus_gsi_authz_handle_destroy(
        cas_handle, globus_gfs_acl_cas_destroy_cb, NULL);

    GlobusGFSDebugExit();
}

globus_gfs_acl_module_t                 globus_gfs_acl_cas_module = 
{
    globus_gfs_acl_cas_init,
    globus_gfs_acl_cas_authorize,
    globus_gfs_acl_cas_destroy,
    NULL
};

