/*
 * Copyright 1999-2006 University of Chicago
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "globus_gridftp_server.h"
#include "globus_xio.h"
#include "globus_xio_file_driver.h"
#include "openssl/md5.h"
#include "version.h"
#include <grp.h>
#include <utime.h>
#include <sys/syscall.h>

#include <stdio.h>
#include <string.h>

#include <arpa/inet.h>

#include <fcntl.h>

#include "mdtm.h"
#include "mdtm_io_thread.h"
#include "mdtm_rrd.h"
#include "../../mdtmmonitor/mdtm_man_thread.h"
#include "../../mdtmmonitor/mdtm_monitor.h"

#ifndef MAXPATHLEN
#define MAXPATHLEN 4096
#endif
#define HAVE_SPLICE
//#define MDTM_SPLICE_PIPES_PER_NUMA              32
#define MDTM_DIRECTIO_WATERMARK_DEFAULT         100000000UL
#define MDTM_DIRECTIO_MULTIPLE                  4096            // aligned to 4K by defalut

#define MDTM_OPEN_FILE_IN_CACHE_MAX            10000
#define MDTM_OPEN_FILE_IN_CACHE_DEFAULT        2500

GlobusDebugDeclare(GLOBUS_GRIDFTP_SERVER_FILE);

#define GlobusGFSFileDebugPrintf(level, message)                             \
    GlobusDebugPrintf(GLOBUS_GRIDFTP_SERVER_FILE, level, message)

#define GlobusGFSFileDebugEnter()                                            \
    GlobusGFSFileDebugPrintf(                                                \
        GLOBUS_GFS_DEBUG_TRACE,                                              \
        ("[%s] Entering\n", _gfs_name))
        
#define GlobusGFSFileDebugExit()                                             \
    GlobusGFSFileDebugPrintf(                                                \
        GLOBUS_GFS_DEBUG_TRACE,                                              \
        ("[%s] Exiting\n", _gfs_name))

#define GlobusGFSFileDebugExitWithError()                                    \
    GlobusGFSFileDebugPrintf(                                                \
        GLOBUS_GFS_DEBUG_TRACE,                                              \
        ("[%s] Exiting with error\n", _gfs_name))

GlobusDebugDefine(GLOBUS_GRIDFTP_SERVER_FILE);

typedef void
(*globus_l_gfs_file_cksm_cb_t)(
    globus_result_t                     result,
    char *                              cksm,
    void *                              user_arg);

typedef struct globus_l_gfs_file_cksm_monitor_s
{
    globus_gfs_operation_t              op;
    globus_off_t                        offset;
    globus_off_t                        length;
    globus_off_t                        count;
    globus_off_t                        read_left;
    globus_size_t                       block_size;
    globus_l_gfs_file_cksm_cb_t         internal_cb;
    void *                              internal_cb_arg;
    
    globus_callback_handle_t            marker_handle;
    int                                 marker_freq;
    globus_bool_t                       send_marker;
    globus_off_t                        total_bytes;
    
    MD5_CTX                             mdctx;
    globus_byte_t                       buffer[1];
} globus_l_gfs_file_cksm_monitor_t;

typedef struct 
{
    gss_cred_id_t                       cred;
    char *                              sbj;
    char *                              username;
    char *                              pw;
} gfs_l_file_session_t;

typedef struct
{
  pthread_mutex_t                     lock;
  globus_fifo_t                       q_;
} mdtm_monitor_queue_t;

/*
 * When transferring big file, it may be split into multiple segments to
 * achieve parallelism. This construct contains information on the file
 * and segments. Especially, the "consumed" is used for directory transfer
 * which uses the big virtual file concept.
 */
typedef struct
{
  long                  total_segments;         // number of segments
  long                  remain_segments;        // segments not transfered
  size_t                segment_size;           // segment size in bytes
  size_t                completed_file_size;    // transfered file in bytes
  size_t                file_size;              // total file size in bytes
  size_t                consumed;               // offset in the virtual file
  char *                pathname;               // file full path
} mdtm_file_segment_info_t;

typedef struct
{
  pthread_mutex_t                     lock;
  void                                *dir_q;
  mdtm_file_segment_info_t *          pending_segments;
  globus_off_t                        consumed;
} mdtm_dirq_info_t;

typedef struct globus_l_file_monitor_s
{
    globus_mutex_t                      lock;
    globus_memory_t                     mem;
    globus_priority_q_t                 queue;
    globus_list_t *                     buffer_list;
    globus_gfs_operation_t              op;
    char *                              pathname;
    char *                              pathroot;
    globus_xio_handle_t                 file_handle;
    globus_off_t                        file_offset;
    globus_off_t                        read_offset;
    globus_off_t                        read_length;
    int                                 pending_writes;
    int                                 pending_reads;
    globus_size_t                       block_size;
    int                                 optimal_count;
    int                                 node_ndx;
    globus_object_t *                   error;
    globus_bool_t                       first_read;
    globus_bool_t                       eof;
    globus_bool_t                       aborted;
    int                                 concurrency_check;
    int                                 concurrency_check_interval;
    char *                              expected_cksm;
    char *                              expected_cksm_alg;
    /* added for multicast stuff, but cold be genreally useful */
    gfs_l_file_session_t *              session;

    globus_result_t                     finish_result;

    /* LZ: splice */
#ifdef HAVE_SPLICE
    mdtm_pipefd_t *                     mdtm_pipefds;
    mdtm_pipemem_t                      pipemem;
#endif
    pthread_mutex_t                     plock;

    int                                 file_numa;
//    globus_l_file_monitor_t             *parent, *child;
    globus_off_t                        multithrd_offset;
    globus_off_t                        multithrd_length;
    globus_off_t                        consumed;
    mdtm_io_thrd_mgr_t *                reader_mgr;
    mdtm_io_thrd_mgr_t *                writer_mgr;
    unsigned int                        iothreads_idx;
    int                                 mt_idx;         //LZ: multithread
    mdtm_monitor_queue_t *              monitor_q;
    void *                              monitor_pool;
    int                                 parallel;
    int                                 parallel_max;
    mdtm_dirq_info_t                    *dirq_info;
    void                                *transfer_info;
    globus_bool_t                       splice_on;
    globus_bool_t                       directio_on;            //LZ: directio flag is set
    globus_bool_t                       directio_allowed;       //LZ: qualified for directio
    globus_off_t                        directio_watermark;     //LZ: min. file size for directio
    int                                 directio_sectorsize;    //LZ: device sector size
    int                                 directio_bytes_padded;
    globus_off_t                        alloc_size;         //LZ: the size of whole file
    char *                              file_system;

    globus_hashtable_t                  url2handle_hash;
    globus_hashtable_t                  handle2info_hash;
    globus_hashtable_t                  handle2handle_hash;
    int                                 n_opens;
    int                                 n_closes;

    globus_bool_t                       mdtm_is_writer;

    mdtm_xmit_progress_t *              xmit_progress;
    globus_xio_file_flag_t              open_flags;
} globus_l_file_monitor_t;

typedef struct
{
    globus_byte_t *                     buffer;
    globus_off_t                        offset;
    globus_size_t                       length;
} globus_l_buffer_info_t;

typedef struct gfs_l_file_stack_entry_s
{
    globus_xio_driver_t                 driver;
    char *                              driver_name;
    char *                              opts;
} gfs_l_file_stack_entry_t;

typedef struct {
  void *                io_handle;
  size_t                bytes_left;
  char *                url;
  void *                handle;
  globus_bool_t         directio_on;
  int                   directio_bytes_padded;
} mdtm_file_info_t;     //LZ:TODO consolidate with globus_gass_copy.c

typedef struct
{
  globus_mutex_t                      mutex;
  globus_cond_t                       cond;
  volatile globus_bool_t              done;
  globus_bool_t                       use_err;
  globus_object_t *                   err;
  globus_l_file_monitor_t *           monitor;
  char *                              pathname;
} mdtm_blocking_info_t;


typedef struct {
  globus_gfs_operation_t              op;
  globus_result_t                     finish_result;
  mdtm_xmit_progress_t *              xmit_progress;
 }mdtm_finish_trans_info_t;

typedef struct {
   pthread_mutex_t               lock;
   int                           offset;
   int                           capacity;
   globus_l_file_monitor_t **    monitors;
} mdtm_monitor_pool_t;

static globus_xio_driver_t              globus_l_gfs_file_driver;

//LZ: stats

static struct timeval   start_stamp;
static struct timeval   open_stamp;
static struct timeval   read_stamp;
static struct timeval   close_stamp;
static double           intervel_stamps[4];
// end of stats

static
void
mdtm_l_gfs_file_dispatch_read(void *user_arg);

//static
//void
//globus_l_gfs_file_close_mt(void *user_arg);

//static
//void
//mdtm_l_gfs_file_dispatch_read_locked(
//    void *           arg);
//
//static
//void
//mdtm_l_gfs_file_dispatch_write_locked(
//    void *           arg);

static
globus_result_t
globus_l_gfs_file_open(
    globus_xio_handle_t *               file_handle,
    const char *                        pathname,
    globus_xio_file_flag_t              open_flags,
    void *                              arg);

static
globus_result_t
mdtm_l_gfs_file_open(
    globus_xio_handle_t *               file_handle,
    const char *                        pathname,
    globus_xio_file_flag_t              open_flags,
    void *                              arg);

static
globus_result_t
mdtm_l_gfs_file_close2(
    globus_xio_handle_t                 file_handle,
    globus_result_t                     in_result);

static
globus_result_t
globus_l_gfs_file_cksm(
    globus_gfs_operation_t              op,
    const char *                        pathname,
    const char *                        algorithm,
    globus_off_t                        offset,
    globus_off_t                        length,
    globus_l_gfs_file_cksm_cb_t         internal_cb,
    void *                              internal_cb_arg);

static
void
mdtm_report_progress_kickout(void *data);

static
void
mdtm_report_progress_finished_kickout(void *data);

static
mdtm_xmit_progress_t *
mdtm_xmit_progress_alloc(char* user_arg, globus_gfs_operation_t op, int direction);

static
void
mdtm_monitor_pool_destroy(mdtm_monitor_pool_t * pool);

static
void
mdtm_l_gfs_file_dispatch_close_kickout(void *user_arg);

static
int roundUp(int numToRound, int multiple)
{
    if (multiple == 0)
        return numToRound;

    int remainder = numToRound % multiple;
    if (remainder == 0)
        return numToRound;

    return numToRound + multiple - remainder;
}

/**
 * Return the mdtm_io_thrd_mgr_t object for given type and filename
 *
 * param@filename	file name; only for storage type; otherwise set to NULL
 * param@type		type of device: MDTM_DEVICE_CLASS_STORAGE, MDTM_DEVICE_CLASS_NETWORK
 *
 * return			the pointer to the io_manager
 */
static
mdtm_io_thrd_mgr_t *
mdtm_io_manager(const char *filename, const int type)
{
	mdtm_io_thrd_mgr_t *	io_mgr = NULL;
    char                    devname[PATH_MAX];

    if(type == MDTM_DEVICE_CLASS_STORAGE) {
    	int rc;
    	mdtm_lv_seg_t * 		segments;

    	if(!strcmp(filename, "/dev/zero") || !strcmp(filename, "/dev/random"))
    		return mdtm_io_thrd_mgr_list_get2(NULL, MDTM_DEVICE_CLASS_STORAGE);

    	rc = mdtm_map_file2(filename, 0, &segments);
    	if(rc < 0) {
    		if(rc == MDTM_ERR_INVADL)
    			globus_gfs_log_message(
    					GLOBUS_GFS_LOG_ERR,
						"mdtm_map_file2 failed @ %s:%d: cannot map %s duo to file system parsing failure.\n",  __FILE__,__LINE__, (filename==NULL)? "":filename);
    		if(rc == MDTM_ERR_NOTSUPPORT)
    			globus_gfs_log_message(
    					GLOBUS_GFS_LOG_ERR,
						"mdtm_map_file2 failed @ %s:%d: cannot map %s due to lack of system functions like lvm.\n",  __FILE__,__LINE__, (filename==NULL)? "":filename);
    		if(rc == MDTM_ERR_NOMEM)
    			globus_gfs_log_message(
    					GLOBUS_GFS_LOG_ERR,
						"mdtm_map_file2 failed @ %s:%d: cannot map %s due to insufficient memory.\n",  __FILE__,__LINE__, (filename==NULL)? "":filename);
    		if(rc == MDTM_ERR_NOEXTENT)
    			globus_gfs_log_message(
    					GLOBUS_GFS_LOG_ERR,
						"mdtm_map_file2 failed @ %s:%d: cannot map %s due to inability to find extents.\n",  __FILE__,__LINE__, (filename==NULL)? "":filename);

    		return NULL;
    	}
    	else if (rc == 0) {
    		// could be sparse file
    		io_mgr = mdtm_io_thrd_mgr_list_get2(NULL, MDTM_DEVICE_CLASS_STORAGE);
    	}
    	else {
    		strcpy(devname, segments[0].dev);
    		mdtm_map_file_free(segments, rc);

    		io_mgr = mdtm_io_thrd_mgr_list_get(devname);
    		if( io_mgr == 0)
    			io_mgr = mdtm_io_thrd_mgr_list_get2(NULL, MDTM_DEVICE_CLASS_STORAGE);
    	}
    }
    else if(type == MDTM_DEVICE_CLASS_NETWORK) {
        memset(devname, 0, sizeof(devname));
        if(mdtm_config_get_netif(devname, PATH_MAX) < 0) {
            globus_gfs_log_message(
                GLOBUS_GFS_LOG_ERR,
                "Can not find network device %s in config file.\n", devname);
            return NULL;
        }
        io_mgr = mdtm_io_thrd_mgr_list_get((const char*)devname);
        if( io_mgr == 0)
        	io_mgr = mdtm_io_thrd_mgr_list_get2(NULL, MDTM_DEVICE_CLASS_NETWORK);
    }
    else {
        globus_gfs_log_message(
            GLOBUS_GFS_LOG_ERR,
            "Can not find network device %s in config file.\n", devname);
        return NULL;
    }

	return io_mgr;
}

/**
 * Return the parallel number
 *
 * @param A_mgr		[IN] io manager
 * @param B_mgr		[IN] io manager
 * @param dirq_info	[IN] directory queue
 * @param size		[IN] file size
 * @param par		[OUT] parallel value
 * @param sgzie		[OUT] segment size
 */
static
globus_result_t
mdtm_l_get_parallel(
		mdtm_io_thrd_mgr_t 		*A_mgr,
		mdtm_io_thrd_mgr_t 		*B_mgr,
		mdtm_dirq_info_t 		*dirq_info,
		globus_off_t 			size,
		int 					*par,
		globus_off_t 			*sgsize)
{
    int totalA, totalB, parallel;
    globus_off_t segsize;

    if(dirq_info) {
        // if directory
        int dirq_size = 0;
        totalA = mdtm_io_thrd_mgr_getthrdnum(A_mgr);
        totalB = mdtm_io_thrd_mgr_getthrdnum(B_mgr);
        parallel = (totalB < totalA) ? totalB : totalA;
        dirq_size = mdtm_dir_queue_size(dirq_info->dir_q, MDTM_DIR_QUEUE_SIZE_FILE);
        if(parallel > dirq_size) {
        	// if number of files is less than the current parallel number,
        	// check if there is large file that can be transfered in
        	// multiple threads
        	size_t	max_filesize, max_cfg_filesize;

        	max_filesize = mdtm_dir_queue_maxfilesize(dirq_info->dir_q);
        	max_cfg_filesize = (size_t)mdtm_config_get_maxfilesize();
        	if(max_filesize < max_cfg_filesize) {
        		// if no large file, each file use one thread
        		parallel = dirq_size;
        	}
        	else {
        		// if existing large file, calculate equivalent number to use multiple threads
        		size_t	s_size, sum;
            	int count, adj_dirq_size;

        		mdtm_util_round((long long)max_cfg_filesize, (long long*)&s_size);
        		for(count = 0, sum = 0; sum < max_filesize; count++) {
        			sum += s_size;
        		}
        		adj_dirq_size = dirq_size + count;
        		parallel = (parallel < adj_dirq_size) ? parallel : adj_dirq_size;
        	}
        }

        if(sgsize)
        	segsize = size;
    }
    else if(size >= mdtm_config_get_maxfilesize()){
        // if single large file
        totalA = mdtm_io_thrd_mgr_getthrdnum(A_mgr);
        totalB = mdtm_io_thrd_mgr_getthrdnum(B_mgr);
        parallel = (totalB < totalA) ? totalB : totalA;
        if(sgsize)
        	mdtm_util_round((long long)size/parallel, (long long*)&segsize);
    }
    else{
        // if single small file
        parallel = 1;
        if(sgsize)
        	segsize = size;
    }

    *par = parallel;
    if(sgsize)
    	*sgsize = segsize;

    return GLOBUS_SUCCESS;
}

static
globus_result_t
globus_l_gfs_file_make_stack(
    globus_gfs_operation_t              op,
    globus_l_file_monitor_t *           mon,
    globus_xio_attr_t                   attr,
    globus_xio_stack_t                  stack,
    gfs_l_file_session_t *              session_h)
{
    globus_xio_driver_t                 driver;
    globus_result_t                     result;
    globus_list_t *                     driver_list = NULL;
    globus_xio_driver_list_ent_t *      ent;
    GlobusGFSName(globus_l_gfs_file_make_stack);

    globus_gfs_data_get_file_stack_list(op, &driver_list);

    /* set the cred in case anyone wants it */
    globus_xio_attr_cntl(
        attr,
        NULL,
        GLOBUS_XIO_ATTR_SET_CREDENTIAL,
        session_h->cred,
        session_h->sbj,
        session_h->username,
        session_h->pw);

    if(driver_list == NULL)
    {
        result = globus_xio_stack_push_driver(
            stack, globus_l_gfs_file_driver);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusGFSErrorWrapFailed(
                "globus_xio_stack_push_driver", result);
            goto error_push;
        }
    }
    else
    {
        while(!globus_list_empty(driver_list))
        {
            ent = (globus_xio_driver_list_ent_t *)
                globus_list_remove(&driver_list, driver_list);

            if(strcmp(ent->driver_name, "file") == 0)
            {
                driver = globus_l_gfs_file_driver;
                result = globus_xio_stack_push_driver(
                    stack, globus_l_gfs_file_driver);
            }
            else
            {
                driver = ent->driver;
                result = globus_xio_stack_push_driver(stack, ent->driver);
            }
            if(result != GLOBUS_SUCCESS)
            {
                result = GlobusGFSErrorWrapFailed(
                    "globus_xio_stack_push_driver", result);
                goto error_push;
            }

            if(ent->opts != NULL)
            {
                /* ignore error */
                globus_xio_attr_cntl(
                    attr,
                    driver,
                    GLOBUS_XIO_SET_STRING_OPTIONS,
                    ent->opts);
            }
        }
    }

    return GLOBUS_SUCCESS;

error_push:
    return result;
}



/*
 * if priority_1 comes after priority_2, return > 0
 * else if priority_1 comes before priority_2, return < 0
 * else return 0
 */
 
static
int
globus_l_gfs_file_queue_compare(
    void *                              priority_1,
    void *                              priority_2)
{
    globus_l_buffer_info_t *            buf_info1;
    globus_l_buffer_info_t *            buf_info2;
    int                                 rc = 0;
    GlobusGFSName(globus_l_gfs_file_queue_compare);
    GlobusGFSFileDebugEnter();

    buf_info1 = (globus_l_buffer_info_t *) priority_1;
    buf_info2 = (globus_l_buffer_info_t *) priority_2;
    
    /* the void * are really just offsets */
    if(buf_info1->offset > buf_info2->offset)
    {
        rc = 1;
    }
    if(buf_info1->offset < buf_info2->offset)
    {
        rc = -1;
    }
    
    GlobusGFSFileDebugExit();
    return rc;
}

#if 0
static
globus_result_t
globus_l_gfs_file_monitor_init(
    globus_l_file_monitor_t **          u_monitor,
    globus_size_t                       block_size,
    int                                 optimal_count)
{
    globus_l_file_monitor_t *           monitor;
    globus_result_t                     result;
    int                                 rc;
    GlobusGFSName(globus_l_gfs_file_monitor_init);
    GlobusGFSFileDebugEnter();
        
    monitor = (globus_l_file_monitor_t *) globus_malloc(
        sizeof(globus_l_file_monitor_t));
    if(!monitor)
    {
        result = GlobusGFSErrorMemory("monitor");
        goto error_alloc;
    }

    rc = globus_memory_init(&monitor->mem, block_size, optimal_count);
    if(!rc)
    {
        globus_free(monitor);
        result = GlobusGFSErrorMemory("buffer");
        goto error_alloc;
    }

    globus_mutex_init(&monitor->lock, NULL);
    globus_priority_q_init(
        &monitor->queue, globus_l_gfs_file_queue_compare);
    monitor->buffer_list = NULL;
    monitor->op = NULL;
    monitor->file_handle = NULL;
    monitor->pending_reads = 0;
    monitor->pending_writes = 0;
    monitor->file_offset = 0;
    monitor->block_size = block_size;
    monitor->optimal_count = optimal_count;
    monitor->error = NULL;
    monitor->eof = GLOBUS_FALSE;
    monitor->aborted = GLOBUS_FALSE;
    monitor->concurrency_check = 2;
    monitor->concurrency_check_interval = 2;
    monitor->expected_cksm = NULL;
    monitor->expected_cksm_alg = NULL;
    monitor->pathname = NULL;

    *u_monitor = monitor;

    GlobusGFSFileDebugExit();
    return GLOBUS_SUCCESS;

error_alloc:
    GlobusGFSFileDebugExitWithError();
    return result;
}
#endif

static
globus_result_t
globus_l_gfs_file_monitor_init2(
    globus_l_file_monitor_t **          u_monitor,
    globus_size_t                       block_size,
    int                                 optimal_count,
    int                                 numa)
{
    globus_l_file_monitor_t *           monitor;
    globus_result_t                     result;

    GlobusGFSName(globus_l_gfs_file_monitor_init2);
    GlobusGFSFileDebugEnter();

    monitor = (globus_l_file_monitor_t *) globus_malloc(
        sizeof(globus_l_file_monitor_t));
    if(!monitor)
    {
        result = GlobusGFSErrorMemory("monitor");
        goto error_alloc;
    }
    
    globus_mutex_init(&monitor->lock, NULL);
    globus_priority_q_init(
        &monitor->queue, globus_l_gfs_file_queue_compare);
    monitor->buffer_list = NULL;
    monitor->op = NULL;
    monitor->file_handle = NULL;
    monitor->pending_reads = 0;
    monitor->pending_writes = 0;
    monitor->file_offset = 0;
    monitor->block_size = block_size;
    monitor->optimal_count = optimal_count;
    monitor->error = NULL;
    monitor->eof = GLOBUS_FALSE;
    monitor->aborted = GLOBUS_FALSE;
    monitor->concurrency_check = 2;
    monitor->concurrency_check_interval = 2;
    monitor->expected_cksm = NULL;
    monitor->expected_cksm_alg = NULL;
    monitor->pathname = NULL;

    monitor->file_numa = numa;
    monitor->dirq_info = NULL;
    monitor->parallel = 0;
    monitor->parallel_max = 0;
    monitor->monitor_q = NULL;
    monitor->consumed = 0;
    monitor->directio_on = GLOBUS_FALSE;
    monitor->directio_allowed = GLOBUS_FALSE;
    monitor->directio_bytes_padded = 0;
    monitor->mdtm_is_writer = GLOBUS_FALSE;
    monitor->file_system = NULL;
    monitor->handle2handle_hash = NULL;
    monitor->handle2info_hash = NULL;
    monitor->url2handle_hash = NULL;
    monitor->reader_mgr = NULL;
    monitor->writer_mgr = NULL;
    monitor->splice_on = GLOBUS_FALSE;

    *u_monitor = monitor;
    
    //LZ: added
    if(pthread_mutex_init(&monitor->plock, NULL) != 0)
      printf("DEBUG: %s:%d: mutex failed\n",__FILE__,__LINE__);

    GlobusGFSFileDebugExit();
    return GLOBUS_SUCCESS;

error_alloc:
    GlobusGFSFileDebugExitWithError();
    return result;
}

// LZ: FIXME
static void
mdtm_filehandle_free(
    void *           entry)
{
//  globus_xio_handle_t file_handle = (globus_xio_handle_t)entry;

//  mdtm_l_gfs_file_close2(
//      file_handle,
//      GLOBUS_SUCCESS);
  return;
}

static void
mdtm_fileinfo_free(
    void *           entry)
{
  mdtm_file_info_t *info = (mdtm_file_info_t*)entry;
  if(info) {
      globus_free(info->url);
      globus_free(info);
  }
  return;
}

static void
mdtm_url_free(
    void *           entry)
{
  char *url = (char*)entry;

  if(url) free(url);
  return;
}

static
void
globus_l_gfs_file_monitor_destroy(
    globus_l_file_monitor_t *           monitor)
{
    globus_l_buffer_info_t *            buf_info;
    globus_list_t *                     list;
    globus_byte_t *                     buffer;
    GlobusGFSName(globus_l_gfs_file_monitor_destroy);
    GlobusGFSFileDebugEnter();
    
    while(!globus_priority_q_empty(&monitor->queue))
    {
        buf_info = (globus_l_buffer_info_t *)
            globus_priority_q_dequeue(&monitor->queue);
        if(buf_info)
        {
            if(buf_info->buffer)
            {
                globus_memory_push_node(&monitor->mem, buf_info->buffer);
            }
            globus_free(buf_info);
        }
    }
    
    for(list = monitor->buffer_list;
        !globus_list_empty(list);
        list = globus_list_rest(list))
    {
        buffer = (globus_byte_t *) globus_list_first(list);
        if(monitor->splice_on == GLOBUS_TRUE)
          mdtm_pipemem_push_node(&monitor->pipemem, buffer);
        else
          globus_memory_push_node(&monitor->mem, buffer);
    }
    
    if(monitor->pathname && !monitor->mdtm_is_writer)
    {
        globus_free(monitor->pathname);
    }
    
    if(monitor->expected_cksm)
    {
        globus_free(monitor->expected_cksm);
    }
    if(monitor->expected_cksm_alg)
    {
        globus_free(monitor->expected_cksm_alg);
    }
    if(monitor->file_system)
    {
        globus_free(monitor->file_system);
    }
    
    globus_priority_q_destroy(&monitor->queue);
    globus_list_free(monitor->buffer_list);
    if(monitor->splice_on == GLOBUS_TRUE)
      mdtm_pipemem_destroy(&monitor->pipemem);
    else
      globus_memory_destroy(&monitor->mem);

    if(monitor->url2handle_hash)
      globus_hashtable_destroy_all(
          &monitor->url2handle_hash,
          mdtm_filehandle_free);        // LZ: files already closed in globus_l_gfs_file_write_cb_mt()
    if(monitor->handle2info_hash)
      globus_hashtable_destroy_all(
          &monitor->handle2info_hash,
          mdtm_fileinfo_free);
    if(monitor->handle2handle_hash)
      globus_hashtable_destroy_all(
          &monitor->handle2handle_hash,
          mdtm_url_free);

    globus_mutex_destroy(&monitor->lock);
    globus_free(monitor);

    GlobusGFSFileDebugExit();
}


static
void
globus_l_gfs_file_cksm_verify(
    globus_result_t                     result,
    char *                              cksm,
    void *                              user_arg)
{
    globus_l_file_monitor_t *           monitor;
    GlobusGFSName(globus_l_gfs_file_cksm_verify);
    GlobusGFSFileDebugEnter();

    monitor = (globus_l_file_monitor_t *) user_arg;
    
    if(result != GLOBUS_SUCCESS)
    {
        monitor->finish_result = 
            GlobusGFSErrorWrapFailed("checksum verification", result);
    }
    else if(strcmp(monitor->expected_cksm, cksm) != 0)
    {
        char *                          msg = NULL;
        msg = globus_common_create_string(
            "Actual checksum of %s does not match expected checksum of %s.",
            cksm, monitor->expected_cksm);
        monitor->finish_result = GlobusGFSErrorGeneric(msg);
        if(msg)
        {
            globus_free(msg);
        }                       
    }

    globus_gridftp_server_finished_transfer(
        monitor->op, monitor->finish_result);
    
    globus_l_gfs_file_monitor_destroy(monitor);

    GlobusGFSFileDebugExit();
}

static 
globus_bool_t
globus_l_gfs_file_timeout_cb(
    globus_xio_handle_t                 handle,
    globus_xio_operation_type_t         type,
    void *                              user_arg)
{
    GlobusGFSName(globus_l_gfs_file_timeout_cb);

    GlobusGFSFileDebugEnter();

    globus_gfs_log_message(
        GLOBUS_GFS_LOG_WARN,
        "A file access timeout has occurred.\n");
   
    GlobusGFSFileDebugExit();

    return GLOBUS_TRUE;
}

//typedef void * (*globus_thread_func_t)(void *);

static
void
mdtm_server_finished_transfer(
    void* user_arg)
{
  mdtm_finish_trans_info_t *           info;

  if(!user_arg)
    return;

  info = (mdtm_finish_trans_info_t *) user_arg;

  // stop the periodic polling
  if(mdtm_monitor_enabled())
  {
    globus_result_t     res;

    res = globus_callback_unregister(
                    *((globus_callback_handle_t *)info->xmit_progress->callback_handle),
                    NULL,
                    NULL,
                    NULL);
    if(res != GLOBUS_SUCCESS ) {
        free(info);
        return;
    }
  }

  globus_gridftp_server_finished_transfer_mt(info->op, info->finish_result);

  if(mdtm_monitor_enabled())
    mdtm_report_progress_finished_kickout((void*)info);

  free(info);
  return;
}


static
void
globus_l_gfs_file_close_cb(
    globus_xio_handle_t                 handle,
    globus_result_t                     result,
    void *                              user_arg)
{
    globus_l_file_monitor_t *           monitor;
    GlobusGFSName(globus_l_gfs_file_close_cb);

    monitor = (globus_l_file_monitor_t *) user_arg;
    struct stat sb;


    GlobusGFSFileDebugEnter();

    if(monitor != NULL)
    {
        /* must lock/unlock or we can get to the destroy before the
           registeration locak has actually released... crazy threads */
        globus_mutex_lock(&monitor->lock);
        pthread_mutex_lock(&monitor->plock);
        {
            if(monitor->finish_result == GLOBUS_SUCCESS)
            {
                monitor->finish_result = result;
            }
        }
        pthread_mutex_unlock(&monitor->plock);
        globus_mutex_unlock(&monitor->lock);

        if(monitor->finish_result == GLOBUS_SUCCESS &&
            monitor->expected_cksm != NULL)
        {
            /* verify file before finishing */
            result = globus_l_gfs_file_cksm(
                NULL,
                monitor->pathname,
                monitor->expected_cksm_alg,
                0,
                -1,
                globus_l_gfs_file_cksm_verify,
                monitor);
        }
        else
        {
            //LZ: Modified
#if 1
            int queue_size;

            // lock for operations on monitor queue
            pthread_mutex_lock(&monitor->monitor_q->lock);
            queue_size = globus_fifo_size(&monitor->monitor_q->q_);

            // lock dirq if existing
            if(monitor->dirq_info != NULL)
              pthread_mutex_lock(&monitor->dirq_info->lock);

            if((queue_size >= (monitor->parallel - 1)) &&
               ( monitor->mdtm_is_writer ||
                 (monitor->dirq_info == NULL) ||
                 (mdtm_dir_queue_empty(monitor->dirq_info->dir_q)&&!monitor->dirq_info->pending_segments->remain_segments) ||
                 (!monitor->mdtm_is_writer && monitor->error)))
            {
                // All monitors are reclaimed and no files waiting in the directory queue,
                // it is time to finish the current transfer.

                // If direct IO is turned on for writing files,
                // check if truncation is in need
                if(monitor->mdtm_is_writer &&
                    monitor->directio_on &&
                    monitor->pathname &&
                    strcmp(monitor->pathname, "/dev/null"))
                {
                    struct stat sb;
                    int rc;

                    if(lstat(monitor->pathname, &sb) < 0)
                      globus_gfs_log_message(
                          GLOBUS_GFS_LOG_WARN,
                          "%s: lstat failed on %s: %s\n",
                          __FUNCTION__, monitor->pathname, strerror(errno));
                    else
                      if(sb.st_size > ((globus_gfs_transfer_info_t *)monitor->transfer_info)->alloc_size) {
                          rc = truncate(monitor->pathname, ((globus_gfs_transfer_info_t *)monitor->transfer_info)->alloc_size);
                          if(rc < 0)
                            globus_gfs_log_message(
                                GLOBUS_GFS_LOG_WARN,
                                "truncated %s to %lld. rc=%d:%s\n",
                                monitor->pathname,
                                (long long)(((globus_gfs_transfer_info_t *)monitor->transfer_info)->alloc_size),
                                rc, strerror(errno));
                      }
                }

                // unlock dirq if existing
                if(monitor->dirq_info != NULL)
                  pthread_mutex_unlock(&monitor->dirq_info->lock);

                // Reclaim the monitor
                globus_fifo_enqueue(&monitor->monitor_q->q_, monitor);
                pthread_mutex_unlock(&monitor->monitor_q->lock);

                // Dispatch the finish transfer function.
                {
                  mdtm_finish_trans_info_t *    info;

                  info = (mdtm_finish_trans_info_t *)
                      globus_malloc(sizeof(mdtm_finish_trans_info_t));
                  assert(info);
                  info->op = monitor->op;
                  info->finish_result = monitor->finish_result;
                  info->xmit_progress = monitor->xmit_progress;

                  mdtm_thrd_dispatch_task(
                      (mdtm_event_queue_t *)globus_gridftp_server_get_thread_data(monitor->op,0),
                      mdtm_server_finished_transfer,
                      (void *) info);
                }

                // Release resources
                {
                  mdtm_monitor_queue_t *        monitor_q = monitor->monitor_q;
                  void *                        m;

                  if(monitor->dirq_info) {
                      pthread_mutex_destroy(&monitor->dirq_info->lock);
                      mdtm_dir_queue_destroy(monitor->dirq_info->dir_q);
                      if(monitor->dirq_info->pending_segments)
                        globus_free(monitor->dirq_info->pending_segments);
                      globus_free(monitor->dirq_info);
                  }

                  mdtm_monitor_pool_destroy((mdtm_monitor_pool_t*)(monitor->monitor_pool));
                  while( (m = globus_fifo_dequeue(&monitor_q->q_)) != NULL)
                    globus_l_gfs_file_monitor_destroy((globus_l_file_monitor_t *)m);

                  globus_fifo_destroy(&monitor_q->q_);
                  pthread_mutex_destroy(&monitor_q->lock);
                  globus_free(monitor_q);
                }
            }
            else if(monitor->mdtm_is_writer ||
                    monitor->dirq_info == NULL ||
                    (mdtm_dir_queue_empty(monitor->dirq_info->dir_q)&&!monitor->dirq_info->pending_segments->remain_segments) ||
                    (!monitor->mdtm_is_writer && monitor->error)) {
                // One monitor is done and no files waiting in the directory queue for service,
                // then reclaim the monitor and wait for the rest monitors to finish their jobs

                // If direct IO is turned on for writing files,
                // check if truncation is in need
                if( monitor->mdtm_is_writer &&
                    monitor->directio_on &&
                    monitor->pathname &&
                    strcmp(monitor->pathname, "/dev/null"))
                {
                    struct stat sb;
                    int rc;

                    if(lstat(monitor->pathname, &sb) < 0)
                      globus_gfs_log_message(
                          GLOBUS_GFS_LOG_WARN,
                          "%s: lstat failed on %s: %s\n",
                          __FUNCTION__, monitor->pathname, strerror(errno));
                    else
                      if(sb.st_size > monitor->alloc_size) {
                          rc = truncate(monitor->pathname, monitor->alloc_size);
                          if(rc < 0)
                            globus_gfs_log_message(
                                GLOBUS_GFS_LOG_WARN,
                                "truncated %s to %lld. rc=%d:%s\n",
                                monitor->pathname,
                                (long long)monitor->alloc_size, rc, strerror(errno));
                      }
                }

                // unlock dirq if existing
                if(monitor->dirq_info != NULL)
                  pthread_mutex_unlock(&monitor->dirq_info->lock);

                // Reclaim the monitor
                globus_fifo_enqueue(&monitor->monitor_q->q_, monitor);
                pthread_mutex_unlock(&monitor->monitor_q->lock);
            }
            else {
                // there are either files waiting in the directory queue or pending file segments to transfer,
                // then assign it to the current monitor
                // Note: this is only for file reading.
                char                            filename[PATH_MAX];
                globus_off_t                    size, consumed = 0;
                globus_xio_file_flag_t          open_flags;
                mdtm_dirq_info_t *              dirq_info = monitor->dirq_info;
                mdtm_file_segment_info_t *      filesegments = dirq_info->pending_segments;

                consumed = dirq_info->consumed;

                // check if there are pending file segments to transfer...
                if(filesegments->remain_segments == 0) {
                    // no pending file segments, then look for the next file in the directory queue
                    assert(mdtm_dir_queue_dequeue2(dirq_info->dir_q, filename));
                    lstat(filename, &sb);
                    size = sb.st_size;

                    // Check if it is a big file...
                    if(size >= mdtm_config_get_maxfilesize()) {
                        // see a big file, split into segments to transfer in parallel.
                        long        count;
                        size_t      sum;

                        mdtm_util_round((long long)mdtm_config_get_maxfilesize(), (long long*)&filesegments->segment_size);
                        for(count = 0, sum = 0; sum < size; count++) {
                            sum += filesegments->segment_size;
                        }
                        filesegments->total_segments =  count;
                        filesegments->remain_segments = count - 1;
                        filesegments->file_size = size;
                        filesegments->completed_file_size = filesegments->segment_size;
                        filesegments->consumed = consumed;
                        filesegments->pathname = globus_libc_strdup(filename);

                        dirq_info->consumed += filesegments->segment_size;
                        monitor->consumed = filesegments->consumed;
                        monitor->multithrd_offset = 0;
                        monitor->multithrd_length = filesegments->segment_size;
                        monitor->alloc_size = size;
                    }
                    else {
                        // see a normal file, transfer the whole file.
                        dirq_info->consumed += size;
                        monitor->consumed = consumed;
                        monitor->multithrd_offset = 0;
                        monitor->multithrd_length = -1;
                        monitor->alloc_size = size;
                    }
                }
                else {
                    // exist pending file segments, handle those first
                    strcpy(filename, filesegments->pathname);
                    filesegments->remain_segments--;
                    if(filesegments->remain_segments) {
                        // not last segment
                        dirq_info->consumed += filesegments->segment_size;
                        monitor->multithrd_offset = filesegments->completed_file_size;
                        monitor->multithrd_length = filesegments->segment_size;
                        monitor->consumed = filesegments->consumed;
                        monitor->alloc_size = filesegments->file_size;
                        filesegments->completed_file_size += filesegments->segment_size;
                    }
                    else {
                        // last segment
                        dirq_info->consumed += (filesegments->file_size - filesegments->completed_file_size);
                        monitor->multithrd_offset = filesegments->completed_file_size;
                        monitor->multithrd_length = -1;
                        monitor->consumed = filesegments->consumed;
                        monitor->alloc_size = filesegments->file_size;
                        filesegments->completed_file_size += filesegments->segment_size;
                    }
                }
                // end of check

                pthread_mutex_unlock(&monitor->dirq_info->lock);
                pthread_mutex_unlock(&monitor->monitor_q->lock);

                // reuse the current monitor
                // TODO: compare with the globus_l_gfs_file_monitor_init2()
                {
                  monitor->pending_reads = 0;
                  monitor->pending_writes = 0;
                  monitor->file_offset = 0;
                  //                monitor->block_size = block_size;
                  //                monitor->optimal_count = optimal_count;
                  monitor->error = NULL;
                  monitor->eof = GLOBUS_FALSE;
                  monitor->aborted = GLOBUS_FALSE;
                  monitor->concurrency_check = 2;
                  monitor->concurrency_check_interval = 2;
                  monitor->expected_cksm = NULL;
                  monitor->expected_cksm_alg = NULL;
                  //                monitor->file_numa = numa;
                  if(monitor->pathname)
                    globus_free(monitor->pathname);
                  monitor->pathname = globus_libc_strdup(filename);
//                  monitor->multithrd_offset = 0;
//                  monitor->multithrd_length = -1; //(parallel <= 1)? -1 : segsize;
//                  monitor->mt_idx = 0;
//                  monitor->iothreads_idx = mdtm_util_nthsetbit(readers, 1) - 1;       //TODO:
//                  monitor->parallel = parallel;
//                  monitor->monitor_q = monitor_q;
                  monitor->read_length = -1;      //LZ: TODO: check again
//                  monitor->dirq_info = dirq_info;
//                  monitor->transfer_info = (void*)transfer_info;
//                  monitor->consumed = consumed;
                }

                if(globus_gfs_config_get_bool("mdtm_debug")) {
                    gettimeofday(&start_stamp,NULL);
                    intervel_stamps[3] += (start_stamp.tv_sec - close_stamp.tv_sec) * 1000 +
                        (double)(start_stamp.tv_usec - close_stamp.tv_usec) / 1000;
                }

                open_flags = GLOBUS_XIO_FILE_BINARY | GLOBUS_XIO_FILE_RDONLY;
                result = globus_l_gfs_file_open(
                    &monitor->file_handle, filename, open_flags, monitor);
                if(result != GLOBUS_SUCCESS) {
                    result = GlobusGFSErrorWrapFailed("globus_l_gfs_file_open", result);
                    goto error_open;
                }
            } // if(queue_size >= monitor->parallel...)
#else
            globus_gridftp_server_finished_transfer(
                monitor->op, monitor->finish_result);
            globus_l_gfs_file_monitor_destroy(monitor);
#endif
            // LZ: End
        }
    }
error_open:
//error_alloc:
    GlobusGFSFileDebugExit();
}

static
void
globus_l_gfs_file_close_kickout(
    void *                              arg)
{
    globus_l_file_monitor_t *           monitor;

    monitor = (globus_l_file_monitor_t *) arg;

#if 0
    globus_gridftp_server_finished_transfer(
        monitor->op, monitor->finish_result);
    globus_l_gfs_file_monitor_destroy(monitor);
#else
    // Dispatch the finish transfer function.
//    if(!monitor->dirq_info) {
//        globus_gridftp_server_finished_transfer(
//            monitor->op, monitor->finish_result);
//        globus_l_gfs_file_monitor_destroy(monitor);
//    }
//    else
    //LZ: comparable to globus_l_gfs_file_close_cb()
    {
        int queue_size;

        // lock for operations on monitor queue
        pthread_mutex_lock(&monitor->monitor_q->lock);
        queue_size = globus_fifo_size(&monitor->monitor_q->q_);

        // lock dirq if existing
        //                     pthread_mutex_lock(&monitor->dirq_info->lock);
        if(queue_size >= (monitor->parallel - 1)){

            // Reclaim the monitor
            globus_fifo_enqueue(&monitor->monitor_q->q_, monitor);
            pthread_mutex_unlock(&monitor->monitor_q->lock);

            {
              mdtm_finish_trans_info_t *    info;

              info = (mdtm_finish_trans_info_t *)
                    globus_malloc(sizeof(mdtm_finish_trans_info_t));
              assert(info);
              info->op = monitor->op;
              info->finish_result = monitor->finish_result;
              info->xmit_progress = monitor->xmit_progress;

              mdtm_thrd_dispatch_task(
                  (mdtm_event_queue_t *)globus_gridftp_server_get_thread_data(monitor->op,0),
                  mdtm_server_finished_transfer,
                  (void *) info);
            }
        }
        else {

            // Reclaim the monitor
            globus_fifo_enqueue(&monitor->monitor_q->q_, monitor);
            pthread_mutex_unlock(&monitor->monitor_q->lock);
        }

    }
#endif
}

static
void
globus_l_gfs_file_close(
    globus_l_file_monitor_t *           monitor,
    globus_result_t                     in_result)
{
    globus_bool_t                       oneshot = GLOBUS_FALSE;
    globus_result_t                     result;

    monitor->finish_result = in_result;
    if(monitor->file_handle)
    {
        result = globus_xio_register_close(
            monitor->file_handle,
            NULL,
            globus_l_gfs_file_close_cb,
            monitor);
        if(result != GLOBUS_SUCCESS)
        {
            oneshot = GLOBUS_TRUE;

            if(monitor->finish_result == GLOBUS_SUCCESS)
            {
                monitor->finish_result = result;
            }
        }
    }
    else
    {
        oneshot = GLOBUS_TRUE;
    }

    if(oneshot)
    {
        globus_callback_register_oneshot(
            NULL,
            NULL,
            globus_l_gfs_file_close_kickout,
            monitor);
    }
}

//static
//void
//mdtm_l_gfs_file_close_cb(
//    globus_xio_handle_t                 handle,
//    globus_result_t                     result,
//    void *                              user_arg)
//{
//    mdtm_blocking_info_t *              blocking_info;
//
//    GlobusGFSName(globus_l_gfs_file_close_cb);
//
//    blocking_info = (mdtm_blocking_info_t *) user_arg;
//
//    GlobusGFSFileDebugEnter();
//
//
//    globus_mutex_lock(&blocking_info->mutex);
//    blocking_info->done = GLOBUS_TRUE;
////    if(error != GLOBUS_NULL)
////    {
////        blocking_info->err = globus_object_copy(error);
////        blocking_info->use_err = GLOBUS_TRUE;
////    }
//    globus_cond_signal(&blocking_info->cond);
//    globus_mutex_unlock(&blocking_info->mutex);
//
//    GlobusGFSFileDebugExit();
//}

typedef struct
{
  char *                filename;
  globus_off_t          filesz;
} mdtm_directio_pad_info_t;

static
void
mdtm_l_gfs_file_close_cb(
    globus_xio_handle_t                 handle,
    globus_result_t                     result,
    void *                              user_arg)
{
  mdtm_directio_pad_info_t *              file_info;

  GlobusGFSName(mdtm_l_gfs_file_close_cb);
  GlobusGFSFileDebugEnter();

  if(result == GLOBUS_SUCCESS && user_arg != NULL) {
      file_info = (mdtm_directio_pad_info_t *) user_arg;
      truncate(file_info->filename, file_info->filesz);
//      fprintf(stderr, "Truncate %s to %d\n", file_info->filename, (int)file_info->filesz);
      globus_free(file_info->filename);
      globus_free(file_info);
  }
  else
    globus_gfs_log_message(
        GLOBUS_GFS_LOG_WARN,
        "%s failed or invalid argument.\n",
        __FUNCTION__);

  GlobusGFSFileDebugExit();
}

static
void
mdtm_l_gfs_file_close(
    globus_l_file_monitor_t *           monitor,
    globus_result_t                     in_result)
{
    globus_bool_t                       oneshot = GLOBUS_FALSE;
    globus_result_t                     result;

    monitor->finish_result = in_result;
    if(monitor->file_handle)
    {
        if(monitor->directio_bytes_padded > 0) {
            mdtm_directio_pad_info_t *          file_info;
            file_info = globus_calloc(1, sizeof(mdtm_directio_pad_info_t));
            file_info->filename = strdup(monitor->pathname);
            file_info->filesz = monitor->alloc_size;
            result = globus_xio_register_close(
                monitor->file_handle,
                NULL,
                mdtm_l_gfs_file_close_cb,
                (void*)file_info);
            monitor->directio_bytes_padded = 0;
        }
        else
          result = globus_xio_register_close(
              monitor->file_handle,
              NULL,
              NULL,       //mdtm_l_gfs_file_close_cb,
              NULL);      //&blocking_info);
        if(result != GLOBUS_SUCCESS)
        {
            oneshot = GLOBUS_TRUE;

            if(monitor->finish_result == GLOBUS_SUCCESS)
            {
                monitor->finish_result = result;
            }
        }
    }
    else
    {
        oneshot = GLOBUS_TRUE;
    }

    if(oneshot)
    {
        globus_callback_register_oneshot(
            NULL,
            NULL,
            globus_l_gfs_file_close_kickout,
            monitor);
    }
}

// LZ: fixme: need to clean up
static
globus_result_t
mdtm_l_gfs_file_close2(
    globus_xio_handle_t                 file_handle,
    globus_result_t                     in_result)
{
//    globus_bool_t                       oneshot = GLOBUS_FALSE;
    globus_result_t                     result;

//    monitor->finish_result = in_result;
    result = in_result;
    if(file_handle)
    {
//        mdtm_blocking_info_t        blocking_info;
//        globus_condattr_t           condattr;
//
//        globus_mutex_init(&blocking_info.mutex, GLOBUS_NULL);
//        globus_condattr_init(&condattr);
//        globus_condattr_setspace(&condattr, mdtm_thread_get_space());
//        globus_cond_init(&blocking_info.cond, &condattr);
//        blocking_info.done = GLOBUS_FALSE;
//        blocking_info.err = GLOBUS_NULL;
//        blocking_info.use_err = GLOBUS_FALSE;

        result = globus_xio_register_close(
            file_handle,
            NULL,
            NULL,       //mdtm_l_gfs_file_close_cb,
            NULL);      //&blocking_info);
        if(result != GLOBUS_SUCCESS)
        {
//            oneshot = GLOBUS_TRUE;
//
//            if(monitor->finish_result == GLOBUS_SUCCESS)
//            {
//                monitor->finish_result = result;
//            }
        }

//        /* wait on cond_wait() for completion */
//        globus_mutex_lock(&blocking_info.mutex);
//
//        while(!blocking_info.done)
//          {
//            globus_cond_wait(&blocking_info.cond, &blocking_info.mutex);
//          }
//
//        globus_mutex_unlock(&blocking_info.mutex);
//
//        globus_mutex_destroy(&blocking_info.mutex);
//        globus_cond_destroy(&blocking_info.cond);
    }
//    else
//    {
//        oneshot = GLOBUS_TRUE;
//    }
//
//    if(oneshot)
//    {
//        globus_callback_register_oneshot(
//            NULL,
//            NULL,
//            globus_l_gfs_file_close_kickout,
//            monitor);
//    }

    return result;
}

/**
 * stat calls
 */
 
/* basepath and filename must be MAXPATHLEN long 
 * the pathname may be absolute or relative, basepath will be the same */
static
void
globus_l_gfs_file_partition_path(
    const char *                        pathname,
    char *                              basepath,
    char *                              filename)
{
    char                                buf[MAXPATHLEN];
    char *                              filepart;
    GlobusGFSName(globus_l_gfs_file_partition_path);
    GlobusGFSFileDebugEnter();
    
    strncpy(buf, pathname, MAXPATHLEN);
    buf[MAXPATHLEN - 1] = '\0';
    
    filepart = strrchr(buf, '/');
    while(filepart && !*(filepart + 1) && filepart != buf)
    {
        *filepart = '\0';
        filepart = strrchr(buf, '/');
    }

    if(!filepart)
    {
        strcpy(filename, buf);
        basepath[0] = '\0';
    }
    else
    {
        if(filepart == buf)
        {
            if(!*(filepart + 1))
            {
                basepath[0] = '\0';
                filename[0] = '/';
                filename[1] = '\0';
            }
            else
            {
                *filepart++ = '\0';
                basepath[0] = '/';
                basepath[1] = '\0';
                strcpy(filename, filepart);
            }
        }
        else
        {                
            *filepart++ = '\0';
            strcpy(basepath, buf);
            strcpy(filename, filepart);
        }
    }    

    GlobusGFSFileDebugExit();
}

static
globus_result_t
globus_l_gfs_file_realpath(
    const char *                        in_path,
    char **                             out_path,
    void *                              user_arg)
{
    globus_result_t                     result;
    char                                resolved[MAXPATHLEN];
    GlobusGFSName(globus_l_gfs_file_stat);
    GlobusGFSFileDebugEnter();
    
    if(realpath(in_path, resolved) == NULL)
    {
        result = GlobusGFSErrorSystemError("realpath", errno);
        goto error;
    }
    *out_path = globus_libc_strdup(resolved);

    GlobusGFSFileDebugExit();
    return GLOBUS_SUCCESS;

error:
    GlobusGFSFileDebugExitWithError();
    return result;
}

static
void
globus_l_gfs_file_copy_stat(
    globus_gfs_stat_t *                 stat_object,
    struct stat *                       stat_buf,
    const char *                        filename,
    const char *                        symlink_target,
    int                                 link_mode,
    globus_gridftp_server_control_stat_error_t  error)
{
    GlobusGFSName(globus_l_gfs_file_copy_stat);
    GlobusGFSFileDebugEnter();

    stat_object->mode     = stat_buf->st_mode;
    stat_object->nlink    = stat_buf->st_nlink;
    stat_object->uid      = stat_buf->st_uid;
    stat_object->gid      = stat_buf->st_gid;
    stat_object->size     = stat_buf->st_size;
    stat_object->mtime    = stat_buf->st_mtime;
    stat_object->atime    = stat_buf->st_atime;
    stat_object->ctime    = stat_buf->st_ctime;
    stat_object->dev      = stat_buf->st_dev;
    stat_object->ino      = stat_buf->st_ino;
    stat_object->link_mode = link_mode;
    stat_object->error    = error;
    
    if(filename && *filename)
    {
        stat_object->name = globus_libc_strdup(filename);
    }
    else
    {
        stat_object->name = NULL;
    }
    if(symlink_target && *symlink_target)
    {
        stat_object->symlink_target = globus_libc_strdup(symlink_target);
    }
    else
    {
        stat_object->symlink_target = NULL;
    }

    GlobusGFSFileDebugExit();
}
static
void
globus_l_gfs_file_destroy_stat(
    globus_gfs_stat_t *                 stat_array,
    int                                 stat_count)
{
    int                                 i;
    GlobusGFSName(globus_l_gfs_file_destroy_stat);
    GlobusGFSFileDebugEnter();
        
    for(i = 0; i < stat_count; i++)
    {
        if(stat_array[i].name != NULL)
        {
            globus_free(stat_array[i].name);
        }        
        if(stat_array[i].symlink_target != NULL)
        {
            globus_free(stat_array[i].symlink_target);
        }
    }
    globus_free(stat_array);     

    GlobusGFSFileDebugExit();
}
    

#define GFS_STAT_COUNT_CHECK 100
#define GFS_STAT_COUNT_MAX 1000
#define GFS_STAT_TIME 10

static
void
globus_l_gfs_file_stat(
    globus_gfs_operation_t              op,
    globus_gfs_stat_info_t *            stat_info,
    void *                              user_arg)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    struct stat                         stat_buf;
    struct stat                         link_stat_buf;
    globus_gfs_stat_t *                 stat_array;
    int                                 stat_count = 0;
    int                                 total_stat_count = 0;
//    DIR *                               dir;
    char                                basepath[MAXPATHLEN];
    char                                filename[MAXPATHLEN];
    char                                symlink_target[MAXPATHLEN];
    globus_gridftp_server_control_stat_error_t  base_error = GLOBUS_GRIDFTP_SERVER_CONTROL_STAT_SUCCESS;
    
    GlobusGFSName(globus_l_gfs_file_stat);
    GlobusGFSFileDebugEnter();
    
    /* lstat is the same as stat when not operating on a link */
    if(lstat(stat_info->pathname, &stat_buf) != 0)
    {
        result = GlobusGFSErrorSystemError("stat", errno);
        goto error_stat1;
    }
    /* if this is a link we still need to stat to get the info we are 
        interested in and then use realpath() to get the full path of 
        the symlink target */
    *symlink_target = '\0';
    if(S_ISLNK(stat_buf.st_mode))
    {
        int stat_result = 0;
        
        if(stat_info->use_symlink_info)
        {
            memset(&link_stat_buf, 0, sizeof(struct stat));
            stat_result = stat(stat_info->pathname, &link_stat_buf);
        } 
        else if(stat(stat_info->pathname, &stat_buf) != 0)
        {
            result = GlobusGFSErrorSystemError("stat", errno);
            goto error_stat1;
        }
        
        if(stat_result < 0 || realpath(stat_info->pathname, symlink_target) == NULL)
        {
            int nchars = readlink(stat_info->pathname, symlink_target, MAXPATHLEN);
            if (nchars < 0) 
        {
            result = GlobusGFSErrorSystemError("realpath", errno);
            goto error_stat1;
        }
            symlink_target[nchars] = '\0';
            base_error = GLOBUS_GRIDFTP_SERVER_CONTROL_STAT_INVALIDLINK;
        }
    }    
    globus_l_gfs_file_partition_path(stat_info->pathname, basepath, filename);
    
    // mdtm: create and send metedata for folder
    if(stat_info->is_meta) {
        // generating metadata for folder to transfer
        void 	*q;
        char 	*buff;
        size_t 	buffSize, used;
        long	nblocks;

        q = (void*)mdtm_dir_queue_create(stat_info->pathname);
        nblocks = mdtm_dir_queue_blocks(q, 0);
        buffSize = nblocks * MDTM_META_ENTRY_SIZE + 1024;
        if((buff = (char*)malloc(buffSize)) == NULL) {
        	globus_gfs_log_message(
        			GLOBUS_GFS_LOG_ERR,
					"%s: memory allocation failed for meta data: %lld\n", __FUNCTION__, buffSize);
            GlobusGFSFileDebugExit();
        	return;
        }
        memset(buff,0, buffSize);
        used = mdtm_dir_queue_create_meta2(q, (void*) buff, buffSize);

        globus_gfs_log_message(
            GLOBUS_GFS_LOG_INFO,
            "Creating metadata for folder: files: %lld, dirs: %lld, blocks: %lld, metadata size: %lld.\n",
			mdtm_dir_queue_size(q, MDTM_DIR_QUEUE_SIZE_FILE),
			mdtm_dir_queue_size(q, MDTM_DIR_QUEUE_SIZE_DIR),
			nblocks,
			used);

        globus_gridftp_server_finished_stat(
            op, result, (globus_gfs_stat_t *)buff, used);
        //        globus_l_gfs_file_destroy_stat(stat_array, stat_count);
        //        free(buff);
        mdtm_dir_queue_destroy(q);
        GlobusGFSFileDebugExit();
        return;
    }
    else if(stat_info->is_publish) {
        globus_gridftp_server_finished_stat(
            op, result, NULL, 0);

        GlobusGFSFileDebugExit();
        return;
    }
    // mdtm: end of creating and sending metadata for folder

    if(!S_ISDIR(stat_buf.st_mode) || stat_info->file_only)
    {
        stat_array = (globus_gfs_stat_t *)
            globus_malloc(sizeof(globus_gfs_stat_t));
        if(!stat_array)
        {
            result = GlobusGFSErrorMemory("stat_array");
            goto error_alloc1;
        }
        
        globus_l_gfs_file_copy_stat(
            stat_array, &stat_buf, filename, symlink_target, link_stat_buf.st_mode, base_error);
        stat_count = 1;
    }
    else
    {
        struct dirent *                 dir_entry;
        struct dirent **                entries = NULL;
        int                             i;
        int                             j;
        char                            dir_path[MAXPATHLEN];
        int                             stat_limit_check = GFS_STAT_COUNT_CHECK;
        int                             stat_limit_max = GFS_STAT_COUNT_MAX;
        time_t                          stat_limit_time;
        globus_bool_t                   check_cdir = GLOBUS_TRUE;

        stat_limit_time = time(NULL) + GFS_STAT_TIME;

        total_stat_count = scandir(
            stat_info->pathname,
            &entries,
            NULL,
            (getenv("FTPNOSORT") ? NULL : alphasort));
        if(total_stat_count < 0)
        {
            result = GlobusGFSErrorSystemError("scandir", errno);
            goto error_open;
        }

        stat_array = (globus_gfs_stat_t *) globus_malloc(
            sizeof(globus_gfs_stat_t) *
            (GLOBUS_MIN(stat_limit_max, total_stat_count) + 1));
        if(!stat_array)
        {
            result = GlobusGFSErrorMemory("stat_array");
            goto error_alloc2;
        }

        snprintf(
            dir_path,
            sizeof(dir_path),
            "%s/%s",
            (basepath[0] != '/' || basepath[1] != '\0') ? basepath : "",
            filename);

        dir_path[MAXPATHLEN - 1] = '\0';

        i = 0;
        j = 0;
        while(j < total_stat_count)
        {
            char                        path[MAXPATHLEN];

            dir_entry = entries[j++];
            snprintf(path, sizeof(path), "%s/%s", dir_path, dir_entry->d_name);
            path[MAXPATHLEN - 1] = '\0';

            /* lstat is the same as stat when not operating on a link */
            if(lstat(path, &stat_buf) != 0)
            {
                result = GlobusGFSErrorSystemError("lstat", errno);
                globus_free(dir_entry);
                /* just skip invalid entries */
                continue;
            }
            /* if this is a link we still need to stat to get the info we are
                interested in and then use realpath() to get the full path of
                the symlink target */
            *symlink_target = '\0';
            if(S_ISLNK(stat_buf.st_mode))
            {
                int stat_result = 0;

                if(stat_info->use_symlink_info)
                {
                    memset(&link_stat_buf, 0, sizeof(struct stat));
                    stat_result = stat(path, &link_stat_buf);
                }
                else if(stat(path, &stat_buf) != 0)
                {
                    globus_free(dir_entry);
                    /* just skip invalid entries */
                    continue;
                }
                if(stat_result < 0 || realpath(path, symlink_target) == NULL)
                {
                    int nchars = readlink(path, symlink_target, MAXPATHLEN);
                    if (nchars < 0)
                    {
                        globus_free(dir_entry);
                        /* just skip invalid entries */
                        continue;
                    }
                    symlink_target[nchars] = '\0';
                    base_error = GLOBUS_GRIDFTP_SERVER_CONTROL_STAT_INVALIDLINK;
                }
            }

            globus_l_gfs_file_copy_stat(
                    &stat_array[i], &stat_buf, dir_entry->d_name, symlink_target, link_stat_buf.st_mode, base_error);

            /* set nlink to total files in dir for . entry */
            if(check_cdir && dir_entry->d_name &&
                dir_entry->d_name[0] == '.' && dir_entry->d_name[1] == '\0')
            {
                check_cdir = GLOBUS_FALSE;
                stat_array[i].nlink = total_stat_count;
            }

            i++;
            globus_free(dir_entry);

            /* send updates every GFS_STAT_TIME, checked every GFS_STAT_CHECK */
            if(i >= stat_limit_check)
            {
                time_t                  tmp_time;
                globus_bool_t           send_stats = GLOBUS_FALSE;

                tmp_time = time(NULL);
                if(i >= stat_limit_max || tmp_time > stat_limit_time)
                {
                    send_stats = GLOBUS_TRUE;
                }
                else
                {
                    stat_limit_check += GFS_STAT_COUNT_CHECK;
                }

                if(send_stats)
                {
                    stat_count = i;
                    stat_limit_check = GFS_STAT_COUNT_CHECK;
                    stat_limit_time = tmp_time + GFS_STAT_TIME;

                    i = 0;

                    globus_gridftp_server_finished_stat_partial(
                        op, GLOBUS_SUCCESS, stat_array, stat_count);

                    globus_l_gfs_file_destroy_stat(stat_array, stat_count);

                    stat_array = (globus_gfs_stat_t *) globus_malloc(
                        sizeof(globus_gfs_stat_t) * (stat_limit_max + 1));
                    if(!stat_array)
                    {
                        result = GlobusGFSErrorMemory("stat_array");
                        goto error_alloc2;
                    }

                    stat_count = 0;
                }
            }
        }
        stat_count = i;

        if(entries)
        {
            globus_free(entries);
        }
    }

    globus_gridftp_server_finished_stat(
        op, result, stat_array, stat_count);
    globus_l_gfs_file_destroy_stat(stat_array, stat_count);

    GlobusGFSFileDebugExit();
    return;

error_alloc2:
//    closedir(dir);

error_open:
error_alloc1:
error_stat1:
    globus_gridftp_server_finished_stat(op, result, NULL, 0);

    GlobusGFSFileDebugExitWithError();
}


/* LZ: added
 * Generate the meta file for the directory
 * NOTE: not used
 */
//static
//void
//globus_l_gfs_file_meta(
//    globus_gfs_operation_t              op,
//    globus_gfs_stat_info_t *            stat_info,
//    void *                              user_arg)
//{
//    globus_result_t                     result = GLOBUS_SUCCESS;
//    struct stat                         stat_buf;
//    struct stat                         link_stat_buf;
//    globus_gfs_stat_t *                 stat_array;
//    int                                 stat_count = 0;
////    int                                 total_stat_count = 0;
////    DIR *                               dir;
//    char                                basepath[MAXPATHLEN];
//    char                                filename[MAXPATHLEN];
//    char                                symlink_target[MAXPATHLEN];
//    globus_gridftp_server_control_stat_error_t  base_error = GLOBUS_GRIDFTP_SERVER_CONTROL_STAT_SUCCESS;
//
//    GlobusGFSName(globus_l_gfs_file_meta);
//    GlobusGFSFileDebugEnter();
//
//    /* lstat is the same as stat when not operating on a link */
//    if(lstat(stat_info->pathname, &stat_buf) != 0)
//    {
//        result = GlobusGFSErrorSystemError("stat", errno);
//        goto error_stat1;
//    }
//    /* if this is a link we still need to stat to get the info we are
//        interested in and then use realpath() to get the full path of
//        the symlink target */
//    *symlink_target = '\0';
//    if(S_ISLNK(stat_buf.st_mode))
//    {
//        int stat_result = 0;
//
//        if(stat_info->use_symlink_info)
//        {
//            memset(&link_stat_buf, 0, sizeof(struct stat));
//            stat_result = stat(stat_info->pathname, &link_stat_buf);
//        }
//        else if(stat(stat_info->pathname, &stat_buf) != 0)
//        {
//            result = GlobusGFSErrorSystemError("stat", errno);
//            goto error_stat1;
//        }
//
//        if(stat_result < 0 || realpath(stat_info->pathname, symlink_target) == NULL)
//        {
//            int nchars = readlink(stat_info->pathname, symlink_target, MAXPATHLEN);
//            if (nchars < 0)
//        {
//            result = GlobusGFSErrorSystemError("realpath", errno);
//            goto error_stat1;
//        }
//            symlink_target[nchars] = '\0';
//            base_error = GLOBUS_GRIDFTP_SERVER_CONTROL_STAT_INVALIDLINK;
//        }
//    }
//    globus_l_gfs_file_partition_path(stat_info->pathname, basepath, filename);
//
//    if(!S_ISDIR(stat_buf.st_mode) || stat_info->file_only)
//    {
//        stat_array = (globus_gfs_stat_t *)
//            globus_malloc(sizeof(globus_gfs_stat_t));
//        if(!stat_array)
//        {
//            result = GlobusGFSErrorMemory("stat_array");
//            goto error_alloc1;
//        }
//
//        globus_l_gfs_file_copy_stat(
//            stat_array, &stat_buf, filename, symlink_target, link_stat_buf.st_mode, base_error);
//        stat_count = 1;
//    }
//#if 0
//    else
//    {
//        struct dirent *                 dir_entry;
//        struct dirent **                entries = NULL;
//        int                             i;
//        int                             j;
//        char                            dir_path[MAXPATHLEN];
//        int                             stat_limit_check = GFS_STAT_COUNT_CHECK;
//        int                             stat_limit_max = GFS_STAT_COUNT_MAX;
//        time_t                          stat_limit_time;
//        globus_bool_t                   check_cdir = GLOBUS_TRUE;
//
//        stat_limit_time = time(NULL) + GFS_STAT_TIME;
//
//        total_stat_count = scandir(
//            stat_info->pathname,
//            &entries,
//            NULL,
//            (getenv("FTPNOSORT") ? NULL : alphasort));
//        if(total_stat_count < 0)
//        {
//            result = GlobusGFSErrorSystemError("scandir", errno);
//            goto error_open;
//        }
//
//        stat_array = (globus_gfs_stat_t *) globus_malloc(
//            sizeof(globus_gfs_stat_t) *
//            (GLOBUS_MIN(stat_limit_max, total_stat_count) + 1));
//        if(!stat_array)
//        {
//            result = GlobusGFSErrorMemory("stat_array");
//            goto error_alloc2;
//        }
//
//        snprintf(
//            dir_path,
//            sizeof(dir_path),
//            "%s/%s",
//            (basepath[0] != '/' || basepath[1] != '\0') ? basepath : "",
//            filename);
//
//        dir_path[MAXPATHLEN - 1] = '\0';
//
//        i = 0;
//        j = 0;
//        while(j < total_stat_count)
//        {
//            char                        path[MAXPATHLEN];
//
//            dir_entry = entries[j++];
//            snprintf(path, sizeof(path), "%s/%s", dir_path, dir_entry->d_name);
//            path[MAXPATHLEN - 1] = '\0';
//
//            /* lstat is the same as stat when not operating on a link */
//            if(lstat(path, &stat_buf) != 0)
//            {
//                result = GlobusGFSErrorSystemError("lstat", errno);
//                globus_free(dir_entry);
//                /* just skip invalid entries */
//                continue;
//            }
//            /* if this is a link we still need to stat to get the info we are
//                interested in and then use realpath() to get the full path of
//                the symlink target */
//            *symlink_target = '\0';
//            if(S_ISLNK(stat_buf.st_mode))
//            {
//                int stat_result = 0;
//
//                if(stat_info->use_symlink_info)
//                {
//                    memset(&link_stat_buf, 0, sizeof(struct stat));
//                    stat_result = stat(path, &link_stat_buf);
//                }
//                else if(stat(path, &stat_buf) != 0)
//                {
//                    globus_free(dir_entry);
//                    /* just skip invalid entries */
//                    continue;
//                }
//                if(stat_result < 0 || realpath(path, symlink_target) == NULL)
//                {
//                    int nchars = readlink(path, symlink_target, MAXPATHLEN);
//                    if (nchars < 0)
//                    {
//                        globus_free(dir_entry);
//                        /* just skip invalid entries */
//                        continue;
//                    }
//                    symlink_target[nchars] = '\0';
//                    base_error = GLOBUS_GRIDFTP_SERVER_CONTROL_STAT_INVALIDLINK;
//                }
//            }
//
//            globus_l_gfs_file_copy_stat(
//                    &stat_array[i], &stat_buf, dir_entry->d_name, symlink_target, link_stat_buf.st_mode, base_error);
//
//            /* set nlink to total files in dir for . entry */
//            if(check_cdir && dir_entry->d_name &&
//                dir_entry->d_name[0] == '.' && dir_entry->d_name[1] == '\0')
//            {
//                check_cdir = GLOBUS_FALSE;
//                stat_array[i].nlink = total_stat_count;
//            }
//
//            i++;
//            globus_free(dir_entry);
//
//            /* send updates every GFS_STAT_TIME, checked every GFS_STAT_CHECK */
//            if(i >= stat_limit_check)
//            {
//                time_t                  tmp_time;
//                globus_bool_t           send_stats = GLOBUS_FALSE;
//
//                tmp_time = time(NULL);
//                if(i >= stat_limit_max || tmp_time > stat_limit_time)
//                {
//                    send_stats = GLOBUS_TRUE;
//                }
//                else
//                {
//                    stat_limit_check += GFS_STAT_COUNT_CHECK;
//                }
//
//                if(send_stats)
//                {
//                    stat_count = i;
//                    stat_limit_check = GFS_STAT_COUNT_CHECK;
//                    stat_limit_time = tmp_time + GFS_STAT_TIME;
//
//                    i = 0;
//
//                    globus_gridftp_server_finished_stat_partial(
//                        op, GLOBUS_SUCCESS, stat_array, stat_count);
//
//                    globus_l_gfs_file_destroy_stat(stat_array, stat_count);
//
//                    stat_array = (globus_gfs_stat_t *) globus_malloc(
//                        sizeof(globus_gfs_stat_t) * (stat_limit_max + 1));
//                    if(!stat_array)
//                    {
//                        result = GlobusGFSErrorMemory("stat_array");
//                        goto error_alloc2;
//                    }
//
//                    stat_count = 0;
//                }
//            }
//        }
//        stat_count = i;
//
//        if(entries)
//        {
//            globus_free(entries);
//        }
//    }
//#endif
//    else {
//        // mdtm meta file generating
//        void *q;
//        char *buff;
//        size_t buffSize, used, nentries;
//
//        q = (void*)mdtm_dir_queue_create(stat_info->pathname);
//        nentries = mdtm_dir_queue_size(
//            q,
//            MDTM_DIR_QUEUE_SIZE_FILE | MDTM_DIR_QUEUE_SIZE_DIR);
//
//        buffSize = nentries * MDTM_META_ENTRY_SIZE;
//
//        buff = (char*)malloc(buffSize);
//        assert(buff);
//        memset(buff,0, buffSize);
//        used = mdtm_dir_queue_create_meta2(q, (void*) buff, buffSize);
//        //TODO:
////        globus_gridftp_server_finished_stat(
////            op, result, buff, buffSize);
////        globus_l_gfs_file_destroy_stat(stat_array, stat_count);
//
//        mdtm_dir_queue_destroy(q);
//    }
//
////    globus_gridftp_server_finished_stat(
////        op, result, stat_array, stat_count);
////    globus_l_gfs_file_destroy_stat(stat_array, stat_count);
//
//    GlobusGFSFileDebugExit();
//    return;
//
////error_alloc2:
////    closedir(dir);
//
////error_open:
//error_alloc1:
//error_stat1:
//    globus_gridftp_server_finished_stat(op, result, NULL, 0);
//
//    GlobusGFSFileDebugExitWithError();
//}

static
globus_result_t
globus_l_gfs_file_mkdir(
    globus_gfs_operation_t   op,
    const char *                        pathname)
{
    int                                 rc;
    globus_result_t                     result;
    GlobusGFSName(globus_l_gfs_file_mkdir);
    GlobusGFSFileDebugEnter();

    rc = mkdir(pathname, 0777);
    if(rc != 0)
    {
        result = GlobusGFSErrorSystemError("mkdir", errno);
        goto error;
    }
    
    globus_gridftp_server_finished_command(op, GLOBUS_SUCCESS, NULL);
        
    GlobusGFSFileDebugExit();
    return GLOBUS_SUCCESS;
    
error:
    GlobusGFSFileDebugExitWithError();
    return result;
}

static
globus_result_t
globus_l_gfs_file_rmdir(
    globus_gfs_operation_t   op,
    const char *                        pathname)
{
    int                                 rc;
    globus_result_t                     result;
    GlobusGFSName(globus_l_gfs_file_rmdir);
    GlobusGFSFileDebugEnter();

    rc = rmdir(pathname);
    if(rc != 0)
    {
        result = GlobusGFSErrorSystemError("rmdir", errno);
        goto error;
    }
    
    globus_gridftp_server_finished_command(op, GLOBUS_SUCCESS, NULL);
        
    GlobusGFSFileDebugExit();
    return GLOBUS_SUCCESS;
    
error:
    GlobusGFSFileDebugExitWithError();
    return result;
}

static
globus_result_t
globus_l_gfs_file_delete_dir(
    const char *                        pathname)
{

    globus_result_t                     result;
    int                                 rc;
    DIR *                               dir;
    struct stat                         stat_buf;
    struct dirent *                     dir_entry;
    int                                 i;
    char                                path[MAXPATHLEN];
    GlobusGFSName(globus_l_gfs_file_delete_dir);
    GlobusGFSFileDebugEnter();
    
    /* lstat is the same as stat when not operating on a link */
    if(lstat(pathname, &stat_buf) != 0)
    {
        result = GlobusGFSErrorSystemError("stat", errno);
        goto error_stat;
    }
    
    if(!S_ISDIR(stat_buf.st_mode))
    {
        /* remove anything that isn't a dir -- don't follow links */
        rc = unlink(pathname);       
        if(rc != 0)
        {
            result = GlobusGFSErrorSystemError("unlink", errno);
            goto error_unlink1;
        }
    }
    else
    {
        dir = globus_libc_opendir(pathname);
        if(!dir)
        {
            result = GlobusGFSErrorSystemError("opendir", errno);
            goto error_open;
        }
        
        for(i = 0;
            globus_libc_readdir_r(dir, &dir_entry) == 0 && dir_entry;
            i++)
        {   
            if(dir_entry->d_name[0] == '.' && 
                (dir_entry->d_name[1] == '\0' || 
                (dir_entry->d_name[1] == '.' && dir_entry->d_name[2] == '\0')))
            {
                globus_free(dir_entry);
                continue;
            }
            snprintf(path, sizeof(path), "%s/%s", pathname, dir_entry->d_name);
            path[MAXPATHLEN - 1] = '\0';
              
            /* lstat is the same as stat when not operating on a link */
            if(lstat(path, &stat_buf) != 0)
            {
                result = GlobusGFSErrorSystemError("lstat", errno);
                globus_free(dir_entry);
                /* just skip invalid entries */
                continue;
            }
            
            if(!S_ISDIR(stat_buf.st_mode))
            {
                /* remove anything that isn't a dir -- don't follow links */
                rc = unlink(path);       
                if(rc != 0)
                {
                    result = GlobusGFSErrorSystemError("unlink", errno);
                    goto error_unlink2;
                }
            }
            else
            {
                result = globus_l_gfs_file_delete_dir(path);
                if(result != GLOBUS_SUCCESS)
                {
                    goto error_recurse;
                }
            }

            globus_free(dir_entry);
        }

        closedir(dir);
        rc = rmdir(pathname);
        if(rc != 0)
        {
            result = GlobusGFSErrorSystemError("rmdir", errno);
            goto error_rmdir;
        }
    } 
    
    GlobusGFSFileDebugExit();
    return GLOBUS_SUCCESS;

error_recurse:
error_unlink2:
        closedir(dir);
        globus_free(dir_entry);
error_open: 
error_stat:
error_unlink1:
error_rmdir:    
    GlobusGFSFileDebugExitWithError();
    return result; 
}
    
static
globus_result_t
globus_l_gfs_file_delete(
    globus_gfs_operation_t              op,
    const char *                        pathname,
    globus_bool_t                       recurse)
{
    int                                 rc;
    globus_result_t                     result;
    GlobusGFSName(globus_l_gfs_file_delete);
    GlobusGFSFileDebugEnter();

    if(!recurse)
    {
        rc = unlink(pathname);
        if(rc != 0)
        {
            result = GlobusGFSErrorSystemError("unlink", errno);
            goto error;
        }
    }
    else
    {
        result = globus_l_gfs_file_delete_dir(pathname);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusGFSErrorWrapFailed("recursion", result);
            goto error;
        }
    }
    
    globus_gridftp_server_finished_command(op, GLOBUS_SUCCESS, NULL);
        
    GlobusGFSFileDebugExit();
    return GLOBUS_SUCCESS;
    
error:
    GlobusGFSFileDebugExitWithError();
    return result;
}

static
globus_result_t
globus_l_gfs_file_rename(
    globus_gfs_operation_t   op,
    const char *                        from_pathname,
    const char *                        to_pathname)
{
    int                                 rc;
    globus_result_t                     result;
    GlobusGFSName(globus_l_gfs_file_rename);
    GlobusGFSFileDebugEnter();

    rc = rename(from_pathname, to_pathname);
    if(rc != 0)
    {
        result = GlobusGFSErrorSystemError("rename", errno);
        goto error;
    }
    
    globus_gridftp_server_finished_command(op, GLOBUS_SUCCESS, NULL);
        
    GlobusGFSFileDebugExit();
    return GLOBUS_SUCCESS;
    
error:
    GlobusGFSFileDebugExitWithError();
    return result;
}

static
globus_result_t
globus_l_gfs_file_chgrp(
    globus_gfs_operation_t   op,
    const char *                        pathname,
    const char *                        group)
{
    int                                 rc;
    globus_result_t                     result;
    struct group *                      grp_info;
    int                                 grp_id;
    char*                               endpt;
    
    GlobusGFSName(globus_l_gfs_file_chgrp);
    GlobusGFSFileDebugEnter();

    grp_info = getgrnam(group);
    if(grp_info != NULL)
    {
        grp_id = grp_info->gr_gid;
    } 
    else
    {
        grp_id = strtol(group, &endpt, 10);
        if(*group == '\0' || *endpt != '\0')
        {
            result = GlobusGFSErrorSystemError("chgrp", EPERM);
            goto error;
        }
    }
    
    if(grp_id < 0)
    {
        result = GlobusGFSErrorSystemError("chgrp", EPERM);
        goto error;
    }
    
    rc = chown(pathname, -1, grp_id);
    if(rc != 0)
    {
        result = GlobusGFSErrorSystemError("chgrp", errno);
        goto error;
    }
    
    globus_gridftp_server_finished_command(op, GLOBUS_SUCCESS, NULL);
        
    GlobusGFSFileDebugExit();
    return GLOBUS_SUCCESS;
    
error:
    GlobusGFSFileDebugExitWithError();
    return result;
}

static
globus_result_t
globus_l_gfs_file_utime(
    globus_gfs_operation_t              op,
    const char *                        pathname,
    time_t                              modtime)
{
    int                                 rc;
    globus_result_t                     result;
    struct utimbuf                      ubuf;
    GlobusGFSName(globus_l_gfs_file_utime);
    GlobusGFSFileDebugEnter();

    ubuf.modtime = modtime;
    ubuf.actime = time(NULL);
    
    rc = utime(pathname, &ubuf);
    if(rc != 0)
    {
        result = GlobusGFSErrorSystemError("utime", errno);
        goto error;
    }
    
    globus_gridftp_server_finished_command(op, GLOBUS_SUCCESS, NULL);
        
    GlobusGFSFileDebugExit();
    return GLOBUS_SUCCESS;
    
error:
    GlobusGFSFileDebugExitWithError();
    return result;
}

static
globus_result_t
globus_l_gfs_file_symlink(
    globus_gfs_operation_t   op,
    const char *                        reference_path,
    const char *                        pathname)
{
    int                                 rc;
    globus_result_t                     result;
    GlobusGFSName(globus_l_gfs_file_symlink);
    GlobusGFSFileDebugEnter();

    rc = symlink(reference_path, pathname);
    if(rc != 0)
    {
        result = GlobusGFSErrorSystemError("symlink", errno);
        goto error;
    }
    
    globus_gridftp_server_finished_command(op, GLOBUS_SUCCESS, NULL);
        
    GlobusGFSFileDebugExit();
    return GLOBUS_SUCCESS;
    
error:
    GlobusGFSFileDebugExitWithError();
    return result;
}

static
globus_result_t
globus_l_gfs_file_chmod(
    globus_gfs_operation_t   op,
    const char *                        pathname,
    mode_t                              mode)
{
    int                                 rc;
    globus_result_t                     result;
    GlobusGFSName(globus_l_gfs_file_chmod);
    GlobusGFSFileDebugEnter();

    rc = chmod(pathname, mode);
    if(rc != 0)
    {
        result = GlobusGFSErrorSystemError("chmod", errno);
        goto error;
    }
    
    globus_gridftp_server_finished_command(op, GLOBUS_SUCCESS, NULL);
        
    GlobusGFSFileDebugExit();
    return GLOBUS_SUCCESS;
    
error:
    GlobusGFSFileDebugExitWithError();
    return result;
}

static
void
globus_l_gfs_file_cksm_read_cb(
    globus_xio_handle_t                 handle, 
    globus_result_t                     result,
    globus_byte_t *                     buffer,
    globus_size_t                       len,
    globus_size_t                       nbytes, 
    globus_xio_data_descriptor_t        data_desc,
    void *                              user_arg)
{
    globus_l_gfs_file_cksm_monitor_t *  monitor;
    globus_bool_t                       eof = GLOBUS_FALSE;
    char *                              md5ptr;
    unsigned char                       md[MD5_DIGEST_LENGTH];
    char                                md5sum[MD5_DIGEST_LENGTH * 2 + 1];
    int                                 i;    
    GlobusGFSName(globus_l_gfs_file_cksm_read_cb);
    GlobusGFSFileDebugEnter();
    
    monitor = (globus_l_gfs_file_cksm_monitor_t *) user_arg;

    if(result != GLOBUS_SUCCESS)
    {
        if(globus_xio_error_is_eof(result))
        {
            eof = GLOBUS_TRUE;
        }
        else
        {
            result = GlobusGFSErrorWrapFailed(
                "checksum read callback", result);
            goto error_read;
        }
    }        
    if(monitor->length >= 0)
    {
        monitor->read_left -= nbytes;
        monitor->count = (monitor->read_left > monitor->block_size) ? 
            monitor->block_size : monitor->read_left;
        if(monitor->read_left == 0)
        {
            eof = GLOBUS_TRUE;
        }
    }
    monitor->total_bytes += nbytes;
    
    MD5_Update(&monitor->mdctx, buffer, nbytes);

    if(!eof)
    {
        if(monitor->send_marker)
        {
            monitor->send_marker = GLOBUS_FALSE;
            
            char                        count[128];
            sprintf(count, "%"GLOBUS_OFF_T_FORMAT, monitor->total_bytes);
            
            globus_gridftp_server_intermediate_command(
                monitor->op, GLOBUS_SUCCESS, count);
        }

        result = globus_xio_register_read(
            handle,
            monitor->buffer,
            monitor->count,
            monitor->count,
            NULL,
            globus_l_gfs_file_cksm_read_cb,
            monitor);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusGFSErrorWrapFailed(
                "globus_xio_register_read", result);
            goto error_register;
        }
    }
    else
    {
        if(monitor->marker_handle)
        {
            globus_callback_unregister(
                monitor->marker_handle,
                NULL,
                NULL,
                NULL);
            monitor->marker_handle = NULL;
        }
        
        MD5_Final(md, &monitor->mdctx);
    
        globus_xio_register_close(
            handle,
            NULL,
            globus_l_gfs_file_close_cb,
            NULL);
            
        md5ptr = md5sum;
        for(i = 0; i < MD5_DIGEST_LENGTH; i++)
        {
           sprintf(md5ptr, "%02x", md[i]);
           md5ptr++;
           md5ptr++;
        }
        md5ptr = '\0';
            
        if(monitor->internal_cb)
        {
            monitor->internal_cb(
                GLOBUS_SUCCESS, md5sum, monitor->internal_cb_arg);
        }
        else
        {
            globus_gridftp_server_finished_command(
                monitor->op, GLOBUS_SUCCESS, md5sum);
        }   
        
        globus_free(monitor);
            
    }        
    GlobusGFSFileDebugExit();
    return;
        

error_register:
error_read:
    globus_xio_register_close(handle, NULL, NULL, NULL);
    handle = NULL;
    globus_gridftp_server_finished_command(monitor->op, result, NULL);    
    globus_free(monitor);
    
    GlobusGFSFileDebugExitWithError();
}


static void
globus_l_gfs_file_marker_cb(
    void *                              user_arg)
{
    globus_l_gfs_file_cksm_monitor_t *  monitor;
    
    monitor = (globus_l_gfs_file_cksm_monitor_t *) user_arg;
    monitor->send_marker = GLOBUS_TRUE;
}

static
void
globus_l_gfs_file_open_cksm_cb(
    globus_xio_handle_t                 handle,
    globus_result_t                     result,
    void *                              user_arg)
{  
    globus_l_gfs_file_cksm_monitor_t *  monitor;
//    char *                              freq;
    GlobusGFSName(globus_l_gfs_file_open_cksm_cb);
    GlobusGFSFileDebugEnter();
    
    monitor = (globus_l_gfs_file_cksm_monitor_t *) user_arg;

    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed(
            "open", result);
        goto error_open;  
    }  
    
    if(monitor->op)
    {
        globus_gridftp_server_get_update_interval(
            monitor->op, &monitor->marker_freq);
    }

    if(monitor->marker_freq)
    {
        globus_result_t                     res;
        globus_reltime_t                    delay;
        
        GlobusTimeReltimeSet(delay, monitor->marker_freq, 0);
        res = globus_callback_register_periodic(
            &monitor->marker_handle,
            &delay,
            &delay,
            globus_l_gfs_file_marker_cb,
            monitor);
        if(res != GLOBUS_SUCCESS)
        {
            
        }
    }
    
    if(monitor->length >= 0)
    {
        monitor->read_left = monitor->length;
        monitor->count = (monitor->read_left > monitor->block_size) ? 
            monitor->block_size : monitor->read_left;
    }
    else
    {
        monitor->count = monitor->block_size;
    }
    
    if(monitor->offset > 0)
    {
        result = globus_xio_handle_cntl(
            handle,
            GLOBUS_XIO_QUERY,
            GLOBUS_XIO_SEEK,
            monitor->offset,
            GLOBUS_XIO_FILE_SEEK_SET);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusGFSErrorWrapFailed(
                "globus_xio_handle_cntl", result);
            goto error_seek;
        }
    }
    
    MD5_Init(&monitor->mdctx);  
    
    result = globus_xio_register_read(
        handle,
        monitor->buffer,
        monitor->count,
        monitor->count,
        NULL,
        globus_l_gfs_file_cksm_read_cb,
        monitor);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed(
            "globus_xio_register_read", result);
        goto error_register;
    }

    GlobusGFSFileDebugExit();
    return;
        
error_register:
error_seek:
error_open:
    globus_xio_register_close(handle, NULL, NULL, NULL);
    handle = NULL;
    if(monitor->internal_cb)
    {
        monitor->internal_cb(result, NULL, monitor->internal_cb_arg);
    }
    else
    {
        globus_gridftp_server_finished_command(monitor->op, result, NULL);
    }   
    globus_free(monitor);
    
    GlobusGFSFileDebugExitWithError();
}


static
globus_result_t
globus_l_gfs_file_cksm(
    globus_gfs_operation_t              op,
    const char *                        pathname,
    const char *                        algorithm,
    globus_off_t                        offset,
    globus_off_t                        length,
    globus_l_gfs_file_cksm_cb_t         internal_cb,
    void *                              internal_cb_arg)
{
    globus_result_t                     result;
    globus_xio_attr_t                   attr;
    globus_xio_stack_t                  stack;
    globus_xio_handle_t                 file_handle;
    globus_l_gfs_file_cksm_monitor_t *  monitor;
    globus_size_t                       block_size;
    int                                 timeout;
    GlobusGFSName(globus_l_gfs_file_cksm);
    GlobusGFSFileDebugEnter();
    
    if(offset < 0)
    {
        result = GlobusGFSErrorGeneric("Invalid offset.");
        goto param_error;
    }
        
    result = globus_xio_attr_init(&attr);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed("globus_xio_attr_init", result);
        goto error_attr;
    }

    result = globus_xio_attr_cntl(
        attr,
        globus_l_gfs_file_driver,
        GLOBUS_XIO_FILE_SET_FLAGS,
        O_RDONLY);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed("globus_xio_attr_init", result);
        goto error_cntl;
    }

    timeout = globus_gfs_config_get_int("file_timeout");
    if(timeout > 0)
    {
        globus_reltime_t                delay;

        GlobusTimeReltimeSet(delay, timeout, 0);
        result = globus_xio_attr_cntl(
            attr,
            NULL,
            GLOBUS_XIO_ATTR_SET_TIMEOUT_ALL,
            globus_l_gfs_file_timeout_cb,
            &delay,
            NULL);
        if(result != GLOBUS_SUCCESS)
        {
            globus_gfs_log_message(
                GLOBUS_GFS_LOG_WARN,
                "Unable to set file access timeout of %d seconds\n", 
                timeout);
        }
    }
    
    result = globus_xio_stack_init(&stack, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed("globus_xio_stack_init", result);
        goto error_stack;
    }
    
    result = globus_xio_stack_push_driver(stack, globus_l_gfs_file_driver);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed(
            "globus_xio_stack_push_driver", result);
        goto error_push;
    }

    result = globus_xio_handle_create(&file_handle, stack);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed("globus_xio_handle_create", result);
        goto error_create;
    }

    globus_gridftp_server_get_block_size(op, &block_size);

    monitor = (globus_l_gfs_file_cksm_monitor_t *) globus_calloc(
        1, sizeof(globus_l_gfs_file_cksm_monitor_t) + block_size);
    if(monitor == NULL)
    {
        result = GlobusGFSErrorMemory("cheksum buffer");
        goto error_mem;
    }
    
    monitor->op = op;
    monitor->offset = offset;
    monitor->length = length;
    monitor->block_size = block_size;
    monitor->internal_cb = internal_cb;
    monitor->internal_cb_arg = internal_cb_arg;

    result = globus_xio_register_open(
        file_handle,
        pathname,
        attr,
        globus_l_gfs_file_open_cksm_cb,
        monitor);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed("globus_xio_register_open", result);
        goto error_register;
    }

    globus_xio_attr_destroy(attr);
    globus_xio_stack_destroy(stack);
    
    GlobusGFSFileDebugExit();
    return GLOBUS_SUCCESS;

error_register:
    globus_xio_register_close(file_handle, NULL, NULL, NULL);
    file_handle = NULL;
    globus_free(monitor);
    
error_mem:
error_create:
error_push:
    globus_xio_stack_destroy(stack);
    
error_stack:
error_cntl:    
    globus_xio_attr_destroy(attr);
    
error_attr:
param_error:
    GlobusGFSFileDebugExitWithError();
    return result;
}     

static
void
globus_l_gfs_file_command(
    globus_gfs_operation_t   op,
    globus_gfs_command_info_t *        cmd_info,
    void *                              user_arg)
{
    globus_result_t                     result;
    GlobusGFSName(globus_l_gfs_file_command);
    GlobusGFSFileDebugEnter();

    switch(cmd_info->command)
    {
      case GLOBUS_GFS_CMD_MKD:
        result = globus_l_gfs_file_mkdir(op, cmd_info->pathname);
        break;
      case GLOBUS_GFS_CMD_RMD:
        result = globus_l_gfs_file_rmdir(op, cmd_info->pathname);
        break;
      case GLOBUS_GFS_CMD_DELE:
        result = globus_l_gfs_file_delete(
            op, cmd_info->pathname, GLOBUS_FALSE);
        break;
      case GLOBUS_GFS_CMD_SITE_RDEL:
        result = globus_l_gfs_file_delete(
            op, cmd_info->pathname, GLOBUS_TRUE);
        break;
      case GLOBUS_GFS_CMD_RNTO:
        result = globus_l_gfs_file_rename(
            op, cmd_info->from_pathname, cmd_info->pathname);
        break;
      case GLOBUS_GFS_CMD_SITE_CHMOD:
        result = globus_l_gfs_file_chmod(
            op, cmd_info->pathname, cmd_info->chmod_mode);
        break;
      case GLOBUS_GFS_CMD_SITE_CHGRP:
        result = globus_l_gfs_file_chgrp(
            op, cmd_info->pathname, cmd_info->chgrp_group);
        break;
      case GLOBUS_GFS_CMD_SITE_UTIME:
        result = globus_l_gfs_file_utime(
            op, cmd_info->pathname, cmd_info->utime_time);
        break;
      case GLOBUS_GFS_CMD_SITE_SYMLINK:
        result = globus_l_gfs_file_symlink(
            op, cmd_info->from_pathname, cmd_info->pathname);
        break;
      case GLOBUS_GFS_CMD_CKSM:
        result = globus_l_gfs_file_cksm(
            op, 
            cmd_info->pathname, 
            cmd_info->cksm_alg,
            cmd_info->cksm_offset,
            cmd_info->cksm_length,
            NULL,
            NULL);
        break;
      
      default:
        result = GLOBUS_FAILURE;
        break;
    }

    if(result != GLOBUS_SUCCESS)
    {
        goto error;
    }
    
    GlobusGFSFileDebugExit();
    return;

error:
    globus_gridftp_server_finished_command(op, result, NULL);    

    GlobusGFSFileDebugExitWithError();
}

/**
 * recv calls
 */

static
void
globus_l_gfs_file_server_read_cb(
    globus_gfs_operation_t   op,
    globus_result_t                     result,
    globus_byte_t *                     buffer,
    globus_size_t                       nbytes,
    globus_off_t                        offset,
    globus_bool_t                       eof,
    void *                              user_arg);

static
void
globus_l_gfs_file_server_read_cb_mt(
    globus_gfs_operation_t   op,
    globus_result_t                     result,
    globus_byte_t *                     buffer,
    globus_size_t                       nbytes,
    globus_off_t                        offset,
    globus_bool_t                       eof,
    void *                              user_arg);

static
globus_result_t
globus_l_gfs_file_dispatch_write(
    globus_l_file_monitor_t *           monitor);
    
#if 0
static
void
globus_l_gfs_file_write_cb(
    globus_xio_handle_t                 xio_handle, 
    globus_result_t                     result,
    globus_byte_t *                     buffer,
    globus_size_t                       len,
    globus_size_t                       nbytes, 
    globus_xio_data_descriptor_t        data_desc,
    void *                              user_arg)
{
    globus_l_file_monitor_t *           monitor;
    GlobusGFSName(globus_l_gfs_file_write_cb);
    GlobusGFSFileDebugEnter();
    
    monitor = (globus_l_file_monitor_t *) user_arg;

    globus_mutex_lock(&monitor->lock);
    pthread_mutex_lock(&monitor->plock);
    { 
        monitor->pending_writes--;
        globus_gridftp_server_update_bytes_written(
            monitor->op, 
            monitor->file_offset,
            nbytes);
        monitor->file_offset += nbytes;

        if(result != GLOBUS_SUCCESS && monitor->error == NULL)
        {
            monitor->error = GlobusGFSErrorObjWrapFailed("callback", result);
        }
        if(monitor->error != NULL)
        {
            goto error;
        }
        
        if(!monitor->eof)
        {
            result = globus_gridftp_server_register_read(
                monitor->op,
                buffer,
                monitor->block_size,
                globus_l_gfs_file_server_read_cb,
                monitor);
            if(result != GLOBUS_SUCCESS)
            {
                monitor->error = GlobusGFSErrorObjWrapFailed(
                    "globus_gridftp_server_register_read", result);
                goto error;
            }
            
            monitor->pending_reads++;
        }
        else
        {
            globus_memory_push_node(&monitor->mem, buffer);
        }
        
        result = globus_l_gfs_file_dispatch_write(monitor);
        if(result != GLOBUS_SUCCESS)
        {
            monitor->error = GlobusGFSErrorObjWrapFailed(
                "globus_l_gfs_file_dispatch_write", result);
            goto error_dispatch;
        }
        
        if(monitor->pending_reads == 0 && monitor->pending_writes == 0)
        {
            globus_assert(monitor->eof || monitor->aborted);

            globus_l_gfs_file_close(monitor, GLOBUS_SUCCESS);
            pthread_mutex_unlock(&monitor->plock);
            globus_mutex_unlock(&monitor->lock);
        }
        else
        {
            pthread_mutex_unlock(&monitor->plock);
            globus_mutex_unlock(&monitor->lock);
        }
    }
    /* already unlocked */
    
    GlobusGFSFileDebugExit();
    return;

error:
    globus_memory_push_node(&monitor->mem, buffer);
error_dispatch:
    if(monitor->pending_reads != 0 || monitor->pending_writes != 0)
    {
        /* there are still outstanding callbacks, wait for them */
        pthread_mutex_unlock(&monitor->plock);
        globus_mutex_unlock(&monitor->lock);
    
        GlobusGFSFileDebugExitWithError();
        return;
    }

    globus_l_gfs_file_close(monitor, globus_error_put(monitor->error));
    pthread_mutex_unlock(&monitor->plock);
    globus_mutex_unlock(&monitor->lock);

    GlobusGFSFileDebugExitWithError();
}
#endif

//static void
//mdtm_handle2info_table_element_free(
//    void *           data)
//{
//  mdtm_file_info_t *info = (mdtm_file_info_t*)data;
//  if(info->url) free(info->url);
//  return;
//}
//
//static void
//mdtm_url2handle_table_element_free(
//    void *           data)
//{
//  char *url = (char*)data;
//  if(url) free(url);
//  return;
//}

static
void
globus_l_gfs_file_write_cb_mt(
    globus_xio_handle_t                 xio_handle,
    globus_result_t                     result,
    globus_byte_t *                     buffer,
    globus_size_t                       len,
    globus_size_t                       nbytes,
    globus_xio_data_descriptor_t        data_desc,
    void *                              user_arg)
{
    globus_l_file_monitor_t *           monitor;
    GlobusGFSName(globus_l_gfs_file_write_cb);
    GlobusGFSFileDebugEnter();

    monitor = (globus_l_file_monitor_t *) user_arg;

    globus_mutex_lock(&monitor->lock);
    pthread_mutex_lock(&monitor->plock);
    {
        monitor->pending_writes--;
        globus_gridftp_server_update_bytes_written(
            monitor->op,
            monitor->file_offset,
            nbytes);
        monitor->file_offset += nbytes;

        if(result != GLOBUS_SUCCESS && monitor->error == NULL)
        {
            monitor->error = GlobusGFSErrorObjWrapFailed("callback", result);
        }
        if(monitor->error != NULL)
        {
            goto error;
        }

        //LZ:  if directory transfer, close finished file or cached it
        if(monitor->dirq_info && nbytes > 0) {
            globus_hashtable_t *table;
            mdtm_file_info_t * dst_info;
            void *      entry;

//            state->dest.data.io_grp[*mt_idx].blocking = 0;
//            state->dest.data.io_grp[*mt_idx].blocking_info = GLOBUS_NULL;

            dst_info = globus_hashtable_lookup(
                &monitor->handle2info_hash,
                monitor->file_handle);
//            if(!dst_info)goto ssdone;
            assert(dst_info);

            if( dst_info->bytes_left <= nbytes ||
                monitor->directio_bytes_padded > 0) {
                /* the cached file is finished */
                entry =globus_hashtable_remove(
                    &monitor->handle2info_hash,
                    (void*)monitor->file_handle);
                assert(entry);
                free(((mdtm_file_info_t *) entry)->url);
                free(entry);

                entry = globus_hashtable_remove(
                    &monitor->url2handle_hash,
                    (void*)monitor->pathname);
                assert(entry);  // LZ: entry is the file handle to be closed below
//                if(!monitor->eof)
                mdtm_l_gfs_file_close(monitor, GLOBUS_SUCCESS);

                entry = globus_hashtable_remove(
                    &monitor->handle2handle_hash,
                    (void*)monitor->pathname);
                assert(entry);  // LZ: entry is the filename
                globus_free((char*)entry);
                monitor->pathname = NULL;
            }
            else {
                table = &monitor->handle2info_hash;

                if(globus_hashtable_size(table) < MDTM_OPEN_FILE_IN_CACHE_DEFAULT)  {
                    dst_info->bytes_left = dst_info->bytes_left - nbytes;
                    entry = globus_hashtable_update(
                        table,
                        (void*)monitor->file_handle,
                        (void*)dst_info);
                    assert(entry);
                }
                else {
                  entry = globus_hashtable_remove(
                      &monitor->handle2info_hash,
                      (void*)monitor->file_handle);
                  assert(entry);
                  free(((mdtm_file_info_t *)entry)->url);
                  free((mdtm_file_info_t *)entry);

                  entry =globus_hashtable_remove(
                      &monitor->url2handle_hash,
                      (void*)monitor->pathname);
                  assert(entry);
                  //                  if(!monitor->eof)
                  mdtm_l_gfs_file_close(monitor, GLOBUS_SUCCESS);

                  entry = globus_hashtable_remove(
                      &monitor->handle2handle_hash,
                      (void*)monitor->pathname);
                  assert(entry);// LZ: entry is the filename
                  globus_free((char*)entry);
                  monitor->pathname = NULL;
                }
            }
        } // end: directory transfer

        if(!monitor->eof)
        {
#if 0
            result = globus_gridftp_server_register_read_mt(
                monitor->op,
                buffer,
                monitor->block_size,
                globus_l_gfs_file_server_read_cb,
                monitor);
            if(result != GLOBUS_SUCCESS)
            {
                monitor->error = GlobusGFSErrorObjWrapFailed(
                    "globus_gridftp_server_register_read", result);
                goto error;
            }
#else
            result = mdtm_gridftp_server_register_read_dispatch(
                monitor->op,
                buffer,
                monitor->block_size,
                monitor->mt_idx,
                globus_l_gfs_file_server_read_cb_mt,
                monitor);
            if(result != GLOBUS_SUCCESS)
            {
                monitor->error = GlobusGFSErrorObjWrapFailed(
                    "globus_gridftp_server_register_read", result);
                goto error;
            }
#endif

            monitor->pending_reads++;
        }
        else
        {
            if(monitor->splice_on == GLOBUS_TRUE)
              mdtm_pipemem_push_node(&monitor->pipemem, buffer);//buffer = (globus_byte_t*)(monitor->mdtm_pipefds + pipecount++);
            else
              globus_memory_push_node(&monitor->mem, buffer);
        }

        result = globus_l_gfs_file_dispatch_write(monitor);
        if(result != GLOBUS_SUCCESS)
        {
            monitor->error = GlobusGFSErrorObjWrapFailed(
                "globus_l_gfs_file_dispatch_write", result);
            goto error_dispatch;
        }

        if(monitor->pending_reads == 0 && monitor->pending_writes == 0)
        {
            globus_assert(monitor->eof || monitor->aborted);

            // In case of folder transfer, replace with the fake file
            // at the end of transfer before closing file.
            if(monitor->dirq_info) {
                char                      filename[] = "/dev/null";
                globus_hashtable_t *      table;
                void *                    entry;
                globus_xio_handle_t       file_handle;

                monitor->file_handle = globus_hashtable_lookup(
                    &monitor->url2handle_hash,
                    (void *)filename);
                assert(monitor->file_handle);
//                if(monitor->pathname)
//                  globus_free(monitor->pathname);
                monitor->pathname = globus_hashtable_lookup(
                    &monitor->handle2handle_hash,
                    (void *)filename);

                entry = globus_hashtable_remove(
                    &monitor->handle2info_hash,
                    (void*)monitor->file_handle);
                free(((mdtm_file_info_t *) entry)->url);
                free(entry);
                globus_hashtable_remove(
                    &monitor->url2handle_hash,
                    (void*)monitor->pathname);
                globus_hashtable_remove(
                    &monitor->handle2handle_hash,
                    (void*)monitor->pathname);


                table = &monitor->url2handle_hash;
                for(entry = globus_hashtable_first(table);
                    entry != GLOBUS_NULL;
                    entry = globus_hashtable_next(table))
                {
                    file_handle = (globus_xio_handle_t)entry;

                    mdtm_l_gfs_file_close2(
                        file_handle,
                        GLOBUS_SUCCESS);
                }

                monitor->pathname = NULL;

//                table = &monitor->handle2info_hash;
//                for(entry = globus_hashtable_first(table);
//                    entry != GLOBUS_NULL;
//                    entry = globus_hashtable_next(table))
//                {
//                    mdtm_file_info_t* dst_info = (mdtm_file_info_t*)entry;
//                    if(dst_info)
//                      globus_free(dst_info);
//                }

//                globus_hashtable_destroy_all(
//                    &monitor->handle2info_hash,
//                    mdtm_handle2info_table_element_free);
//                globus_hashtable_destroy_all(
//                    &monitor->url2handle_hash,
//                    mdtm_url2handle_table_element_free);
//                globus_hashtable_destroy_all(
//                    &monitor->handle2handle_hash,
//                    mdtm_url2handle_table_element_free);
            }

            globus_l_gfs_file_close(monitor, GLOBUS_SUCCESS);
            pthread_mutex_unlock(&monitor->plock);
            globus_mutex_unlock(&monitor->lock);
        }
        else
        {
            pthread_mutex_unlock(&monitor->plock);
            globus_mutex_unlock(&monitor->lock);
        }
    }
    /* already unlocked */

    GlobusGFSFileDebugExit();
    return;

error:
    globus_memory_push_node(&monitor->mem, buffer);
error_dispatch:
    if(monitor->pending_reads != 0 || monitor->pending_writes != 0)
    {
        /* there are still outstanding callbacks, wait for them */
        pthread_mutex_unlock(&monitor->plock);
        globus_mutex_unlock(&monitor->lock);

        GlobusGFSFileDebugExitWithError();
        return;
    }

//    globus_l_gfs_file_close(monitor, globus_error_put(monitor->error));
    pthread_mutex_unlock(&monitor->plock);
    globus_mutex_unlock(&monitor->lock);
    mdtm_l_gfs_file_dispatch_close_kickout((void*)monitor);

    GlobusGFSFileDebugExitWithError();
}

typedef struct {
  char *                        filename;
  globus_l_file_monitor_t *     monitor;
  globus_byte_t *               buffer;
  globus_off_t                  offset;
  globus_size_t                 length;
  mdtm_file_entry_t             entry;
}mdtm_dispatch_write_t;

/*
 * open a new file and dispatch writing operation
 */
static
void
mdtm_l_gfs_file_dispatch_write_newfile(
    void *           arg)
{
  globus_result_t                     result;
  globus_xio_file_flag_t              open_flags;
  globus_l_file_monitor_t *           monitor = NULL;
  char *                              filename = NULL;
  mdtm_file_entry_t                   entry;
  mdtm_dispatch_write_t *             info;
  GlobusGFSName(mdtm_l_gfs_file_dispatch_write_newfile);
  GlobusGFSFileDebugEnter();

  info = (mdtm_dispatch_write_t *)arg;
  monitor  = info->monitor;
  filename = info->filename;
  entry    = info->entry;

//  globus_mutex_lock(&monitor->lock);
//  pthread_mutex_lock(&monitor->plock);
  {
    mdtm_file_info_t *                  dst_info;

    // reset the directio parameters
    monitor->directio_on = GLOBUS_FALSE;
    monitor->directio_allowed = GLOBUS_FALSE;
    monitor->directio_bytes_padded = 0;

    // open file --- blocking call
    open_flags = GLOBUS_XIO_FILE_BINARY |
         GLOBUS_XIO_FILE_CREAT |
         GLOBUS_XIO_FILE_WRONLY;
//    if(((globus_gfs_transfer_info_t *)monitor->transfer_info)->truncate)
//      open_flags |= GLOBUS_XIO_FILE_TRUNC;

    // check if qualified for directio
#ifdef O_DIRECT
    if(globus_gfs_config_get_bool("direct_io")) {
        if( strstr(monitor->file_system, "tmpfs") == NULL &&
            entry.size >= monitor->directio_watermark &&
            monitor->splice_on == GLOBUS_FALSE)
          monitor->directio_allowed = GLOBUS_TRUE;
        else
          monitor->directio_allowed = GLOBUS_FALSE;
    }
#endif

    result = mdtm_l_gfs_file_open(
        &monitor->file_handle,
        filename,
        open_flags,
        monitor
        );
    if (result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed(
                "mdtm_l_gfs_file_open", result);

        globus_gfs_log_message(
                GLOBUS_GFS_LOG_ERR,
                "Can not open file %s: %s\n",
                filename, globus_error_print_friendly(globus_error_peek(result)));

        goto error_dispatch_new_file;
    }

//    if(result == GLOBUS_SUCCESS && handle->partial_offset != -1)
//      {
//        result = globus_io_file_seek(
//            new_handle,
//            handle->partial_offset,
//            SEEK_SET);
//      }

    if(result == GLOBUS_SUCCESS) {
        result = globus_xio_handle_cntl(
            monitor->file_handle,
            GLOBUS_XIO_QUERY,
            GLOBUS_XIO_SEEK,
            entry.offset,
            GLOBUS_XIO_FILE_SEEK_SET);
    }
    if (result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed(
                "globus_xio_handle_cntl", result);

        globus_gfs_log_message(
                GLOBUS_GFS_LOG_ERR,
                "Can not SEEK file %s: %s\n",
                filename, globus_error_print_friendly(globus_error_peek(result)));

        goto error_dispatch_new_file;
    }

    if(result==GLOBUS_SUCCESS)
    {
//                        state->dest.status = GLOBUS_I_GASS_COPY_TARGET_READY;
#ifdef GLOBUS_I_GASS_COPY_DEBUG
        globus_libc_fprintf(stderr,
            "io_setup_put(): SUCCESS opening %s\n",parsed_url.url_path);
      }
    else
      {
        globus_libc_fprintf(stderr,
            "io_setup_put(): FAILED opening %s\n",parsed_url.url_path);
#endif
      }

//    globus_mutex_lock(&monitor->lock);
//    pthread_mutex_lock(&monitor->plock);

    monitor->n_opens++;
    monitor->pathname = filename;       //strdup(filename);

//    // update the hashtables
//    dst_info = (mdtm_file_info_t*)globus_malloc(sizeof(mdtm_file_info_t));
//    dst_info->bytes_left = entry.size; //roundUp(entry.size, MDTM_DIRECTIO_MULTIPLE);
//    dst_info->handle = (void*)monitor->file_handle;
//    dst_info->url = strdup(entry.path);
//    dst_info->directio_on = monitor->directio_on;
//
//    assert((globus_hashtable_insert(
//        &monitor->url2handle_hash,
//        monitor->pathname,
//        (void*)monitor->file_handle)) == GLOBUS_SUCCESS);
//    assert((globus_hashtable_insert(
//        &monitor->handle2info_hash,
//        (void*)monitor->file_handle,
//        dst_info)) == GLOBUS_SUCCESS);
//    assert((globus_hashtable_insert(
//        &monitor->handle2handle_hash,
//        monitor->pathname,
//        (void*)monitor->pathname)) == GLOBUS_SUCCESS);

    if(monitor->splice_on  == GLOBUS_TRUE)
      result = globus_xio_register_write_splice(
          monitor->file_handle,
          info->buffer,
          info->length,
          info->length,
          NULL,
          globus_l_gfs_file_write_cb_mt,
          monitor);
    else {
    	globus_size_t len = info->length;

    	if(monitor->directio_on) {
    		len = roundUp(
    		        info->length,
//                    monitor->directio_sectorsize> 0?
//                    monitor->directio_sectorsize : MDTM_DIRECTIO_MULTIPLE);
    				MDTM_DIRECTIO_MULTIPLE);
    		if(len == info->length) {
    			len = info->length;
    			monitor->directio_bytes_padded = 0;
    		}
    		else {
    			globus_gfs_log_message(
                    GLOBUS_GFS_LOG_INFO,
                    "Adjust %s len from %d to %d: multiple=%d\n",
                    monitor->pathname, (int)info->length, (int)len, MDTM_DIRECTIO_MULTIPLE);
    			monitor->directio_bytes_padded = len - info->length;
    		}
    	}

    	result = globus_xio_register_write(
    			monitor->file_handle,
				info->buffer,
				len,
				len,
				NULL,
				globus_l_gfs_file_write_cb_mt,			//globus_l_gfs_file_write_cb
				monitor);
    }

    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed(
            "globus_xio_register_write", result);

        globus_gfs_log_message(
            GLOBUS_GFS_LOG_ERR,
            "Can not write file %s: %s\n",
            monitor->pathname, globus_error_print_friendly(globus_error_peek(result)));

        goto error_dispatch_new_file;
    }

    // update the hashtables
    dst_info = (mdtm_file_info_t*)globus_malloc(sizeof(mdtm_file_info_t));
    dst_info->bytes_left = entry.size; //roundUp(entry.size, MDTM_DIRECTIO_MULTIPLE);
    dst_info->handle = (void*)monitor->file_handle;
    dst_info->url = strdup(entry.path);
    dst_info->directio_on = monitor->directio_on;
    dst_info->directio_bytes_padded = monitor->directio_bytes_padded;

    assert((globus_hashtable_insert(
        &monitor->url2handle_hash,
        monitor->pathname,
        (void*)monitor->file_handle)) == GLOBUS_SUCCESS);
    assert((globus_hashtable_insert(
        &monitor->handle2info_hash,
        (void*)monitor->file_handle,
        dst_info)) == GLOBUS_SUCCESS);
    assert((globus_hashtable_insert(
        &monitor->handle2handle_hash,
        monitor->pathname,
        (void*)monitor->pathname)) == GLOBUS_SUCCESS);
  }

  globus_free(info);

  GlobusGFSFileDebugExit();
  return;

error_dispatch_new_file:
  GlobusGFSFileDebugExitWithError();
  return;
}

typedef struct {
  globus_xio_handle_t                 xio_handle;
  globus_result_t                     result;
  globus_byte_t *                     buffer;
  globus_size_t                       len;
  globus_size_t                       nbytes;
  globus_xio_data_descriptor_t        data_desc;
  void *                              user_arg;
}mdtm_file_write_cb_kickout_arg_t;

static
void
mdtm_l_gfs_file_write_cb_kickout(
    void *                              user_arg)
{
  mdtm_file_write_cb_kickout_arg_t * arg;

  arg = (mdtm_file_write_cb_kickout_arg_t *)user_arg;
  globus_l_gfs_file_write_cb_mt(
      arg->xio_handle,
      arg->result,
      arg->buffer,
      arg->len,
      arg->nbytes,
      arg->data_desc,
      arg->user_arg);
  globus_free(arg);
  return;
}

/* Called LOCKED */
static
globus_result_t
globus_l_gfs_file_dispatch_write(
    globus_l_file_monitor_t *           monitor)
{
    globus_l_buffer_info_t *            buf_info;
    globus_result_t                     result;
//    globus_xio_file_flag_t              open_flags;
//    globus_bool_t                       delay_release = GLOBUS_FALSE;
    GlobusGFSName(globus_l_gfs_file_dispatch_write);
    GlobusGFSFileDebugEnter();
    
    if(monitor->pending_writes == 0 && !monitor->aborted)
    {
        buf_info = (globus_l_buffer_info_t *)
            globus_priority_q_dequeue(&monitor->queue);
        if(buf_info)
        {
            if(monitor->dirq_info) {
                //if this is directory transfer
                globus_xio_handle_t                 dst_handle;
                mdtm_file_entry_t                   entry;
                int                                 rc;
                char                                filename[PATH_MAX];

                // check if this is data in the buffer. If no, skip writing.
                if(buf_info->length == 0) {
                    mdtm_file_write_cb_kickout_arg_t * data;

                    monitor->pending_writes++;

                    data = globus_malloc(sizeof(mdtm_file_write_cb_kickout_arg_t));
                    data->xio_handle = monitor->file_handle;
                    data->result = GLOBUS_SUCCESS;
                    data->buffer = buf_info->buffer;
                    data->len = buf_info->length;
                    data->nbytes = buf_info->length;
                    data->data_desc = NULL;
                    data->user_arg = (void*) monitor;
                    globus_callback_register_oneshot(
                        NULL,
                        NULL,
                        mdtm_l_gfs_file_write_cb_kickout,
                        data);

                    globus_free(buf_info);

#if 0
                    globus_libc_free(buffer_entry); result = globus_xio_register_write(
                        monitor->file_handle,
                        buf_info->buffer,
                        buf_info->length,
                        buf_info->length,
                        NULL,
        //                globus_l_gfs_file_write_cb,
                        globus_l_gfs_file_write_cb_mt,
                        monitor);
                    static
                    void
                    globus_l_gfs_file_write_cb_mt(
                        globus_xio_handle_t                 xio_handle,
                        globus_result_t                     result,
                        globus_byte_t *                     buffer,
                        globus_size_t                       len,
                        globus_size_t                       nbytes,
                        globus_xio_data_descriptor_t        data_desc,
                        void *                              user_arg)
#endif
                    return GLOBUS_SUCCESS;
                } // no data in buffer
#if 1
                rc = mdtm_dir_queue_find3(
                		monitor->dirq_info->dir_q,
						buf_info->offset,
						&entry);
                if( rc != 0 ) {
                    result = GlobusGFSErrorWrapFailed(
                        "mdtm_dir_queue_find3", GLOBUS_FAILURE);
                	goto error_seek;
                }

                strncpy(filename, monitor->pathroot, PATH_MAX);
                strcpy(strrchr(filename,'/') + 1, entry.path);

                //check if the file is cached
                dst_handle = globus_hashtable_lookup(
                    &monitor->url2handle_hash,
                    (void *)filename);
#else
                // LZ: Debugging code
                {
                  static int                    bCreated = 0;
                  static globus_xio_handle_t    handle;

                  strcpy(filename, "/dev/null");

                  if(!bCreated) {
                      globus_xio_file_flag_t              open_flags;

                      // open file --- blocking call
                      open_flags = GLOBUS_XIO_FILE_BINARY |
                          GLOBUS_XIO_FILE_CREAT |
                          GLOBUS_XIO_FILE_WRONLY;

                      result = mdtm_l_gfs_file_open(
                          &handle,
                          filename,
                          open_flags,
                          monitor
                      );

                      if(result == GLOBUS_SUCCESS) {
                          result = globus_xio_handle_cntl(
                              handle,
                              GLOBUS_XIO_QUERY,
                              GLOBUS_XIO_SEEK,
                              0,
                              GLOBUS_XIO_FILE_SEEK_SET);
                      }
                      bCreated = 1;
                  }
                  dst_handle = handle;
                }
#endif

                // if not cached, open a new file; otherwise adjust its offset
                if(!dst_handle) {
                    mdtm_dispatch_write_t *     info;

                    info = (mdtm_dispatch_write_t *)globus_malloc(sizeof(mdtm_dispatch_write_t));
                    assert(info);
                    info->filename = strdup(filename);
                    info->monitor = monitor;
                    info->buffer = buf_info->buffer;
                    info->length = buf_info->length;
                    info->offset = buf_info->offset;    //used??
//                    info->entry = entry;                // memcpy???
                    memcpy(&info->entry, &entry, sizeof(entry));

                    mdtm_l_gfs_file_dispatch_write_newfile(info);

                    monitor->alloc_size = entry.size;
                    monitor->pending_writes++;
                    globus_free(buf_info);

                    return GLOBUS_SUCCESS;
                } // end of create new
                else {
                    // update current working file
                    if(monitor->file_handle != dst_handle) {
                        mdtm_file_info_t * dst_info;

                        monitor->file_handle = dst_handle;
                        monitor->pathname = globus_hashtable_lookup(
                            &monitor->handle2handle_hash,
                            (void *)filename);

                        // retrieve directio info
                        dst_info = globus_hashtable_lookup(
                            &monitor->handle2info_hash,
                            monitor->file_handle);
                        monitor->directio_on = dst_info->directio_on;
                    }
                    else
                      mdtm_globus_xio_handle_update_space(
                          &monitor->file_handle,
                          mdtm_thread_get_space());

                    // adjust the file offset
                    result = globus_xio_handle_cntl(
                        monitor->file_handle,
                        GLOBUS_XIO_QUERY,
                        GLOBUS_XIO_SEEK,
                        entry.offset,
                        GLOBUS_XIO_FILE_SEEK_SET);
                    assert(result ==GLOBUS_SUCCESS);
                    monitor->alloc_size = entry.size;
                }
            }   // end of directory
            else
              // single file
            if(buf_info->offset != monitor->file_offset)
            {
                globus_off_t            seek_tmp;

                monitor->file_offset = buf_info->offset;
                seek_tmp = monitor->file_offset;

                //LZ: added update to the space
                mdtm_globus_xio_handle_update_space(
                    &monitor->file_handle,
                    mdtm_thread_get_space());

                result = globus_xio_handle_cntl(
                    monitor->file_handle,
                    GLOBUS_XIO_QUERY,
                    GLOBUS_XIO_SEEK,
                    seek_tmp,
                    GLOBUS_XIO_FILE_SEEK_SET);
                if(result != GLOBUS_SUCCESS)
                {
                    result = GlobusGFSErrorWrapFailed(
                        "globus_xio_handle_cntl", result);
                    goto error_seek;
                }
            } // end of single file

            if(monitor->splice_on == GLOBUS_TRUE)
              result = globus_xio_register_write_splice(
                  monitor->file_handle,
                  buf_info->buffer,
                  buf_info->length,
                  buf_info->length,
                  NULL,
                  //                globus_l_gfs_file_write_cb,
                  globus_l_gfs_file_write_cb_mt,
                  monitor);
            else {
                globus_size_t len = buf_info->length;

                if(monitor->directio_on) {
                    len = roundUp(
                        buf_info->length,
//                        monitor->directio_sectorsize> 0?
//                        monitor->directio_sectorsize : MDTM_DIRECTIO_MULTIPLE);
						MDTM_DIRECTIO_MULTIPLE);
                    if(len == buf_info->length) {
                        len = buf_info->length;
                        monitor->directio_bytes_padded = 0;
                    }
                    else {
                        globus_gfs_log_message(
                            GLOBUS_GFS_LOG_INFO,
                            "Adjust %s len from %d to %d: multiple=%d\n",
                            monitor->pathname, (int)buf_info->length, (int)len, MDTM_DIRECTIO_MULTIPLE);
                        monitor->directio_bytes_padded = len - buf_info->length;
                    }
                }

                result = globus_xio_register_write(
                    monitor->file_handle,
                    buf_info->buffer,
                    len,
                    len,
                    NULL,
                    globus_l_gfs_file_write_cb_mt,      //globus_l_gfs_file_write_cb,
                    monitor);
            }

            if(result != GLOBUS_SUCCESS)
            {
                result = GlobusGFSErrorWrapFailed(
                    "globus_xio_register_write", result);
                goto error_register;
            }

            monitor->pending_writes++;

            globus_free(buf_info);
        }
    }

    GlobusGFSFileDebugExit();
    return GLOBUS_SUCCESS;

error_register:
error_seek:
    if(buf_info->buffer)
    {
        globus_memory_push_node(&monitor->mem, buf_info->buffer);
    }
    globus_free(buf_info);

    GlobusGFSFileDebugExitWithError();
    return result;
}

static
void
mdtm_globus_l_gfs_file_dispatch_write_kickout(void *user_arg)
{
  globus_result_t result;
    globus_l_file_monitor_t *           monitor;
//    int                                 rc;
    GlobusGFSName(mdtm_globus_l_gfs_file_dispatch_write);
    GlobusGFSFileDebugEnter();

    monitor = (globus_l_file_monitor_t *) user_arg;

    globus_mutex_lock(&monitor->lock);
    pthread_mutex_lock(&monitor->plock);
//    mdtm_globus_xio_handle_update_space(
//        &monitor->file_handle,
//        mdtm_thread_get_space());

    monitor->pending_writes--;
    result = globus_l_gfs_file_dispatch_write(monitor);
    if(result != GLOBUS_SUCCESS)
      {
        monitor->error = GlobusGFSErrorObjWrapFailed(
            "globus_l_gfs_file_dispatch_write", result);
        goto error_dispatch;
      }

//    if(monitor->pending_reads == 0 && monitor->pending_writes == 0)
//      {
//        globus_assert(monitor->eof || monitor->aborted);
//
//        globus_l_gfs_file_close(monitor, GLOBUS_SUCCESS);
//        pthread_mutex_unlock(&monitor->plock);
//        globus_mutex_unlock(&monitor->lock);
//      }
//    else
    {
      pthread_mutex_unlock(&monitor->plock);
      globus_mutex_unlock(&monitor->lock);
    }

    GlobusGFSFileDebugExit();
    return;

error_dispatch:
    if(monitor->pending_reads != 0 || monitor->pending_writes != 0)
    {
        /* there are still outstanding callbacks, wait for them */
        pthread_mutex_unlock(&monitor->plock);
        globus_mutex_unlock(&monitor->lock);

        GlobusGFSFileDebugExitWithError();
        return;
    }
    globus_l_gfs_file_close(monitor, globus_error_put(monitor->error));
    pthread_mutex_unlock(&monitor->plock);
    globus_mutex_unlock(&monitor->lock);

    GlobusGFSFileDebugExitWithError();
    return;
}

static
void
mdtm_l_gfs_file_dispatch_close_kickout(void *user_arg)
{
//    globus_result_t result;
    globus_l_file_monitor_t *           monitor;
    GlobusGFSName(mdtm_globus_l_gfs_file_dispatch_close_kickout);
    GlobusGFSFileDebugEnter();

    monitor = (globus_l_file_monitor_t *) user_arg;

    globus_mutex_lock(&monitor->lock);
    pthread_mutex_lock(&monitor->plock);

    {
//      globus_assert(monitor->eof || monitor->aborted);

      // In case of folder transfer, replace with the fake file
      // at the end of transfer before closing file.
      if(monitor->dirq_info) {
          char                      filename[] = "/dev/null";
          globus_hashtable_t *      table;
          void *                    entry;
          globus_xio_handle_t       file_handle;

          monitor->file_handle = globus_hashtable_lookup(
              &monitor->url2handle_hash,
              (void *)filename);
          assert(monitor->file_handle);
          monitor->pathname = globus_hashtable_lookup(
              &monitor->handle2handle_hash,
              (void *)filename);

          entry = globus_hashtable_remove(
              &monitor->handle2info_hash,
              (void*)monitor->file_handle);
          free(((mdtm_file_info_t *) entry)->url);
          free(entry);
          globus_hashtable_remove(
              &monitor->url2handle_hash,
              (void*)monitor->pathname);
          globus_hashtable_remove(
              &monitor->handle2handle_hash,
              (void*)monitor->pathname);


          table = &monitor->url2handle_hash;
          for(entry = globus_hashtable_first(table);
              entry != GLOBUS_NULL;
              entry = globus_hashtable_next(table))
          {
              file_handle = (globus_xio_handle_t)entry;

              mdtm_l_gfs_file_close2(
                  file_handle,
                  GLOBUS_SUCCESS);
          }

          monitor->pathname = NULL;
      }

      globus_l_gfs_file_close(monitor, monitor->error? globus_error_put(monitor->error) : GLOBUS_SUCCESS);
      pthread_mutex_unlock(&monitor->plock);
      globus_mutex_unlock(&monitor->lock);
    }

    GlobusGFSFileDebugExit();
    return;
}

/* called locked */
static
void
globus_l_gfs_file_update_concurrency(
    globus_l_file_monitor_t *           monitor)
{
    globus_result_t                     result;
    int                                 optimal_count;
    int                                 extra;
    GlobusGFSName(globus_l_gfs_file_update_concurrency);
    GlobusGFSFileDebugEnter();

    if(!monitor->eof)
    {
        monitor->concurrency_check = monitor->concurrency_check_interval;
        monitor->concurrency_check_interval *= 2;
        if(monitor->concurrency_check_interval > 1024)
        {
            monitor->concurrency_check_interval = 1024;
        }

        globus_gridftp_server_get_optimal_concurrency(
            monitor->op, &optimal_count);
        extra = optimal_count - monitor->optimal_count;

        monitor->optimal_count = optimal_count;
        while(extra-- > 0)
        {
            globus_byte_t *             buffer;

            buffer = globus_memory_pop_node(&monitor->mem);
            result = globus_gridftp_server_register_read(
                monitor->op,
                buffer,
                monitor->block_size,
                globus_l_gfs_file_server_read_cb,
                monitor);
            if(result != GLOBUS_SUCCESS)
            {
                globus_memory_push_node(&monitor->mem, buffer);
                result = GlobusGFSErrorWrapFailed(
                    "globus_gridftp_server_register_read", result);
                goto error_register;
            }

            monitor->pending_reads++;
        }
    }

    GlobusGFSFileDebugExit();
    return;

error_register:
    GlobusGFSFileDebugExitWithError();
}

static
void
globus_l_gfs_file_server_read_cb(
    globus_gfs_operation_t   op,
    globus_result_t                     result,
    globus_byte_t *                     buffer,
    globus_size_t                       nbytes,
    globus_off_t                        offset,
    globus_bool_t                       eof,
    void *                              user_arg)
{
    globus_l_file_monitor_t *           monitor;
    globus_l_buffer_info_t *            buf_info;
    int                                 rc;
    GlobusGFSName(globus_l_gfs_file_server_read_cb);
    GlobusGFSFileDebugEnter();

    monitor = (globus_l_file_monitor_t *) user_arg;

    globus_mutex_lock(&monitor->lock);
    {
        monitor->pending_reads--;
        if(result != GLOBUS_SUCCESS && monitor->error == NULL)
        {
            monitor->error = GlobusGFSErrorObjWrapFailed("callback", result);
        }
        if(monitor->error != NULL)
        {
            goto error;
        }

        if(eof)
        {
            monitor->eof = GLOBUS_TRUE;
        }

        buf_info = (globus_l_buffer_info_t *)
            globus_malloc(sizeof(globus_l_buffer_info_t));
        if(!buf_info)
        {
            monitor->error = GlobusGFSErrorObjMemory("buf_info");
            goto error_alloc;
        }

        /* XXX offset may need some interpretation here depending on type of
         * transfer (may need to remember original start offset)
         */
        buf_info->buffer = buffer;
        buf_info->offset = offset;
        buf_info->length = nbytes;

        monitor->concurrency_check--;
        if(monitor->concurrency_check == 0 && !eof)
        {
            globus_l_gfs_file_update_concurrency(monitor);
        }

        result = globus_l_gfs_file_dispatch_write(monitor);
        if(result != GLOBUS_SUCCESS)
        {
            monitor->error = GlobusGFSErrorObjWrapFailed(
                "globus_l_gfs_file_dispatch_write", result);
            goto error_dispatch;
        }

        rc = globus_priority_q_enqueue(
            &monitor->queue, buf_info, buf_info);
        if(rc != GLOBUS_SUCCESS)
        {
            monitor->error = GlobusGFSErrorObjGeneric(
                "globus_priority_q_enqueue failed");
            goto error_enqueue;
        }
    }
    globus_mutex_unlock(&monitor->lock);
    
    GlobusGFSFileDebugExit();
    return;

error_enqueue:
error_dispatch:
    /* can't free buf_info, its in queue */
    globus_free(buf_info);

error_alloc:
error:
    globus_memory_push_node(&monitor->mem, buffer);
    if(monitor->pending_reads != 0 || monitor->pending_writes != 0)
    {
        /* there are still outstanding callbacks, wait for them */
        globus_mutex_unlock(&monitor->lock);

        GlobusGFSFileDebugExitWithError();
        return;
    }
    globus_l_gfs_file_close(monitor, globus_error_put(monitor->error));
    globus_mutex_unlock(&monitor->lock);

    GlobusGFSFileDebugExitWithError();
}

static
void
globus_l_gfs_file_server_read_cb_mt(
    globus_gfs_operation_t   op,
    globus_result_t                     result,
    globus_byte_t *                     buffer,
    globus_size_t                       nbytes,
    globus_off_t                        offset,
    globus_bool_t                       eof,
    void *                              user_arg)
{
    globus_l_file_monitor_t *           monitor;
    globus_l_buffer_info_t *            buf_info;
    int                                 rc;
    GlobusGFSName(globus_l_gfs_file_server_read_cb);
    GlobusGFSFileDebugEnter();

    monitor = (globus_l_file_monitor_t *) user_arg;

    globus_mutex_lock(&monitor->lock);
    pthread_mutex_lock(&monitor->plock);
    {
        monitor->pending_reads--;
        if(result != GLOBUS_SUCCESS && monitor->error == NULL)
        {
            monitor->error = GlobusGFSErrorObjWrapFailed("callback", result);
        }
        if(monitor->error != NULL)
        {
            goto error;
        }

        if(eof)
        {
            monitor->eof = GLOBUS_TRUE;
        }

        buf_info = (globus_l_buffer_info_t *)
            globus_malloc(sizeof(globus_l_buffer_info_t));
        if(!buf_info)
        {
            monitor->error = GlobusGFSErrorObjMemory("buf_info");
            goto error_alloc;
        }

        /* XXX offset may need some interpretation here depending on type of
         * transfer (may need to remember original start offset)
         */
        buf_info->buffer = buffer;
        buf_info->offset = offset;
        buf_info->length = nbytes;

        monitor->concurrency_check--;
        if(monitor->concurrency_check == 0 && !eof)
        {
            //LZ: TODO
//            globus_l_gfs_file_update_concurrency(monitor);
        }

        rc = globus_priority_q_enqueue(
            &monitor->queue, buf_info, buf_info);
        if(rc != GLOBUS_SUCCESS)
        {
            monitor->error = GlobusGFSErrorObjGeneric(
                "globus_priority_q_enqueue failed");
            goto error_enqueue;
        }

#if 0
        result = globus_l_gfs_file_dispatch_write(monitor);
        if(result != GLOBUS_SUCCESS)
        {
            monitor->error = GlobusGFSErrorObjWrapFailed(
                "globus_l_gfs_file_dispatch_write", result);
            goto error_dispatch;
        }
#else
        // LZ: dispatch the writing tasks to multiple threads
        mdtm_io_thrd_mgr_dispatch_to(
            monitor->writer_mgr,
            monitor->iothreads_idx,
            mdtm_globus_l_gfs_file_dispatch_write_kickout,
            monitor);
        monitor->pending_writes++;
#endif

//        rc = globus_priority_q_enqueue(
//            &monitor->queue, buf_info, buf_info);
//        if(rc != GLOBUS_SUCCESS)
//        {
//            monitor->error = GlobusGFSErrorObjGeneric(
//                "globus_priority_q_enqueue failed");
//            goto error_enqueue;
//        }
    }
    pthread_mutex_unlock(&monitor->plock);
    globus_mutex_unlock(&monitor->lock);

    GlobusGFSFileDebugExit();
    return;

error_enqueue:
//error_dispatch:
    /* can't free buf_info, its in queue */
    globus_free(buf_info);

error_alloc:
error:
    globus_memory_push_node(&monitor->mem, buffer);
    if(monitor->pending_reads != 0 || monitor->pending_writes != 0)
    {
        /* there are still outstanding callbacks, wait for them */
        pthread_mutex_unlock(&monitor->plock);
        globus_mutex_unlock(&monitor->lock);

        GlobusGFSFileDebugExitWithError();
        return;
    }
//    globus_l_gfs_file_close(monitor, globus_error_put(monitor->error));
    mdtm_io_thrd_mgr_dispatch_to(
        monitor->writer_mgr,
        monitor->iothreads_idx,
        mdtm_l_gfs_file_dispatch_close_kickout,
        monitor);
    pthread_mutex_unlock(&monitor->plock);
    globus_mutex_unlock(&monitor->lock);

    GlobusGFSFileDebugExitWithError();
}

#if 0
static
void
globus_l_gfs_file_open_write_cb(
    globus_xio_handle_t                 handle,
    globus_result_t                     result,
    void *                              user_arg)
{
    globus_l_file_monitor_t *           monitor;
    GlobusGFSName(globus_l_gfs_file_open_write_cb);
    GlobusGFSFileDebugEnter();

    monitor = (globus_l_file_monitor_t *) user_arg;
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed(
            "globus_l_gfs_file_open_write_cb", result);
        monitor->file_handle = NULL;
        goto error_open;
    }

    globus_gridftp_server_begin_transfer(
        monitor->op, GLOBUS_GFS_EVENT_TRANSFER_ABORT, monitor);
    
    globus_mutex_lock(&monitor->lock);
    {
        int                             optimal_count;
        globus_size_t                   block_size;
        
        optimal_count = monitor->optimal_count;
        block_size = monitor->block_size;
        while(optimal_count--)
        {
            globus_byte_t *             buffer;

            buffer = globus_memory_pop_node(&monitor->mem);
            result = globus_gridftp_server_register_read(
                monitor->op,
                buffer,
                block_size,
                globus_l_gfs_file_server_read_cb,
                monitor);
            if(result != GLOBUS_SUCCESS)
            {
                globus_memory_push_node(&monitor->mem, buffer);
                result = GlobusGFSErrorWrapFailed(
                    "globus_gridftp_server_register_read", result);
                goto error_register;
            }

            monitor->pending_reads++;
        }
    }
    globus_mutex_unlock(&monitor->lock);

    GlobusGFSFileDebugExit();
    return;

error_register:
    if(monitor->pending_reads != 0)
    {
        /* there are pending reads, need to wait for them */
        monitor->error = globus_error_get(result);
        globus_mutex_unlock(&monitor->lock);

        GlobusGFSFileDebugExitWithError();
        return;
    }
    globus_mutex_unlock(&monitor->lock);

error_open:
    globus_gridftp_server_finished_transfer(monitor->op, result);
    globus_l_gfs_file_monitor_destroy(monitor);

    GlobusGFSFileDebugExitWithError();
}

#else

#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

static
int
mdtm_chown(char * pathname, char * username)
{
  uid_t uid;
  gid_t gid;
  struct passwd *pwd;

  pwd = getpwnam(username);   /* Try getting UID for username */
  if (pwd == NULL) {
      perror("getpwnam");
      return -1;
  }

  uid = pwd->pw_uid;
  gid = pwd->pw_gid;

  if (chown(pathname, uid, gid) == -1) {
      perror("chown");
      return -1;
  }
  return 0;
}

static
void
globus_l_gfs_file_open_write_cb(
    globus_xio_handle_t                 handle,
    globus_result_t                     result,
    void *                              user_arg)
{
    globus_l_file_monitor_t *           monitor;
    GlobusGFSName(globus_l_gfs_file_open_write_cb);
    GlobusGFSFileDebugEnter();

    monitor = (globus_l_file_monitor_t *) user_arg;
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed(
            "globus_l_gfs_file_open_write_cb", result);
        monitor->file_handle = NULL;
        goto error_open;
    }

//    globus_gridftp_server_begin_transfer(
//        monitor->op, GLOBUS_GFS_EVENT_TRANSFER_ABORT, monitor);

    mdtm_chown(monitor->pathname, monitor->session->username);

    globus_mutex_lock(&monitor->lock);
    pthread_mutex_lock(&monitor->plock);
    {
        int                             optimal_count;
        globus_size_t                   block_size;

        optimal_count = monitor->optimal_count;
        block_size = monitor->block_size;
        while(optimal_count--)
        {
            globus_byte_t *             buffer;

            if(monitor->splice_on == GLOBUS_TRUE)
              buffer = mdtm_pipemem_pop_node(&monitor->pipemem);
            else
              buffer = globus_memory_pop_node(&monitor->mem);

#if 0
            result = globus_gridftp_server_register_read_mt(
                monitor->op,
                buffer,
                block_size,
                globus_l_gfs_file_server_read_cb_mt,
                monitor);
            if(result != GLOBUS_SUCCESS)
            {
                globus_memory_push_node(&monitor->mem, buffer);
                result = GlobusGFSErrorWrapFailed(
                    "globus_gridftp_server_register_read", result);
                goto error_register;
            }
#else
            result = mdtm_gridftp_server_register_read_dispatch(
                monitor->op,
                buffer,
                block_size,
                monitor->mt_idx,
                globus_l_gfs_file_server_read_cb_mt,
                monitor);
            if(result != GLOBUS_SUCCESS)
            {
//                globus_memory_push_node(&monitor->mem, buffer);
                result = GlobusGFSErrorWrapFailed(
                    "globus_gridftp_server_register_read", result);
                goto error_register;
            }
#endif

            monitor->pending_reads++;
        }

#if 0
        // LZ: dispatch the writing tasks to multiple threads
        mdtm_io_thrd_mgr_dispatch_to(
            monitor->writer_mgr,
            monitor->iothreads_idx,
            mdtm_globus_l_gfs_file_dispatch_write_kickout,
            monitor);
        monitor->pending_writes++;
#else
        // Nothing to do
#endif
    }
    pthread_mutex_unlock(&monitor->plock);
    globus_mutex_unlock(&monitor->lock);

//    // LZ: dispatch the writing tasks to multiple threads
//    mdtm_io_thrd_mgr_dispatch_to(
//        monitor->writer_mgr,
//        monitor->iothreads_idx,
//        mdtm_globus_l_gfs_file_dispatch_write_kickout,
//        monitor);

    GlobusGFSFileDebugExit();
    return;

error_register:
    if(monitor->pending_reads != 0)
    {
        /* there are pending reads, need to wait for them */
        monitor->error = globus_error_get(result);
        pthread_mutex_unlock(&monitor->plock);
        globus_mutex_unlock(&monitor->lock);

        GlobusGFSFileDebugExitWithError();
        return;
    }
    pthread_mutex_unlock(&monitor->plock);
    globus_mutex_unlock(&monitor->lock);

error_open:
    globus_gridftp_server_finished_transfer(monitor->op, result);
    globus_l_gfs_file_monitor_destroy(monitor);

    GlobusGFSFileDebugExitWithError();
}
#endif

static
void
globus_l_gfs_file_open_read_cb(
    globus_xio_handle_t                 handle,
    globus_result_t                     result,
    void *                              user_arg);

static
globus_result_t
globus_l_gfs_file_open(
    globus_xio_handle_t *               file_handle,
    const char *                        pathname,
    globus_xio_file_flag_t              open_flags,
    void *                              arg)
{
    globus_l_file_monitor_t *           monitor;
    globus_result_t                     result;
    globus_xio_attr_t                   attr;
    globus_xio_stack_t                  stack;
    char *                              perms;
    int                                 timeout;
    GlobusGFSName(globus_l_gfs_file_open);
    GlobusGFSFileDebugEnter();

    monitor = (globus_l_file_monitor_t *) arg;
    
    result = globus_xio_attr_init(&attr);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed("globus_xio_attr_init", result);
        goto error_attr;
    }

#ifdef O_DIRECT
    if(globus_gfs_config_get_bool("direct_io") &&
        monitor->directio_allowed &&
        strcmp(pathname, "/dev/null"))
    {
        open_flags |= O_DIRECT;
        monitor->directio_on = GLOBUS_TRUE;
    }
#endif

    result = globus_xio_attr_cntl(
        attr,
        globus_l_gfs_file_driver,
        GLOBUS_XIO_FILE_SET_FLAGS,
        open_flags);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed("globus_xio_attr_init", result);
        goto error_cntl;
    }

    if(open_flags & GLOBUS_XIO_FILE_CREAT)
    {
        perms = globus_gfs_config_get_string("perms");
        if(perms != NULL)
        {
            int                             p = 0;

            p = strtoul(perms, NULL, 8);
            if(p > 0 ||
                (perms[0] == '0' && perms[1] == '\0'))
            {
                result = globus_xio_attr_cntl(
                    attr,
                    globus_l_gfs_file_driver,
                    GLOBUS_XIO_FILE_SET_MODE,
                    p);
                if(result != GLOBUS_SUCCESS)
                {
                    globus_gfs_log_message(
                        GLOBUS_GFS_LOG_WARN,
                        "Failed to set default permissions to: %s\n", perms);
                }
            }
            else
            {
                globus_gfs_log_message(
                    GLOBUS_GFS_LOG_WARN,
                    "Invalid default permissions: %s\n", perms);
            }
        }
    }

    timeout = globus_gfs_config_get_int("file_timeout");
    if(timeout > 0)
    {
        globus_reltime_t                delay;

        GlobusTimeReltimeSet(delay, timeout, 0);
        result = globus_xio_attr_cntl(
            attr,
            NULL,
            GLOBUS_XIO_ATTR_SET_TIMEOUT_ALL,
            globus_l_gfs_file_timeout_cb,
            &delay,
            NULL);
        if(result != GLOBUS_SUCCESS)
        {
            globus_gfs_log_message(
                GLOBUS_GFS_LOG_WARN,
                "Unable to set file access timeout of %d seconds\n",
                timeout);
        }
    }


    result = globus_xio_stack_init(&stack, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed("globus_xio_stack_init", result);
        goto error_stack;
    }
    result = globus_l_gfs_file_make_stack(
        monitor->op, arg, attr, stack, monitor->session);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_push;
    }

    result = globus_xio_handle_create(file_handle, stack);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed("globus_xio_handle_create", result);
        goto error_create;
    }

    result = globus_xio_register_open(
        *file_handle,
        pathname,
        attr,
        (open_flags & GLOBUS_XIO_FILE_CREAT) ?
            globus_l_gfs_file_open_write_cb :
            globus_l_gfs_file_open_read_cb,
        arg);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed("globus_xio_register_open", result);
        goto error_register;
    }


    globus_xio_attr_destroy(attr);
    globus_xio_stack_destroy(stack);

    GlobusGFSFileDebugExit();
    return GLOBUS_SUCCESS;

error_register:
    globus_xio_register_close(*file_handle, NULL, NULL, NULL);
    *file_handle = NULL;
error_create:
error_push:
    globus_xio_stack_destroy(stack);

error_stack:
error_cntl:
    globus_xio_attr_destroy(attr);

error_attr:
    GlobusGFSFileDebugExitWithError();
    return result;
}

static
void
mdtm_l_gfs_file_open_kickout(void *user_arg)
{
  globus_l_file_monitor_t *     monitor = (globus_l_file_monitor_t *) user_arg;
  globus_result_t               result;
  GlobusGFSName(mdtm_l_gfs_file_open_kickout);
  GlobusGFSFileDebugEnter();

  result = globus_l_gfs_file_open(
      &monitor->file_handle,
      monitor->pathname,
      monitor->open_flags,
      monitor);
  if(result != GLOBUS_SUCCESS)
  {
//      result = GlobusGFSErrorWrapFailed("globus_l_gfs_file_open", result);
      int queue_size;

      globus_gfs_log_message(
          GLOBUS_GFS_LOG_ERR,
          "mdtm_l_gfs_file_open_kickout() failed: %s \n",
          globus_error_print_friendly(globus_error_peek(result)));

      /* must lock/unlock or we can get to the destroy before the
         registeration locak has actually released... crazy threads */
      globus_mutex_lock(&monitor->lock);
      pthread_mutex_lock(&monitor->plock);
      {
          if(monitor->finish_result == GLOBUS_SUCCESS)
          {
              monitor->finish_result = result;
          }
      }
      pthread_mutex_unlock(&monitor->plock);
      globus_mutex_unlock(&monitor->lock);

      // lock for operations on monitor queue
      pthread_mutex_lock(&monitor->monitor_q->lock);
      queue_size = globus_fifo_size(&monitor->monitor_q->q_);

      // lock dirq if existing
      if(monitor->dirq_info != NULL)
        pthread_mutex_lock(&monitor->dirq_info->lock);

      if(queue_size >= (monitor->parallel - 1))
      {
          // Reclaim the monitor
          globus_fifo_enqueue(&monitor->monitor_q->q_, monitor);
          pthread_mutex_unlock(&monitor->monitor_q->lock);

          // All monitors are reclaimed and no files waiting in the directory queue,
          // it is time to finish the current transfer.
          // Dispatch the finish transfer function.
          {
            mdtm_finish_trans_info_t *    info;

            info = (mdtm_finish_trans_info_t *)
                globus_malloc(sizeof(mdtm_finish_trans_info_t));
            assert(info);
            info->op = monitor->op;
            info->finish_result = monitor->finish_result;
            info->xmit_progress = monitor->xmit_progress;

            mdtm_thrd_dispatch_task(
                (mdtm_event_queue_t *)globus_gridftp_server_get_thread_data(monitor->op,0),
                mdtm_server_finished_transfer,
                (void *) info);
          }

          // Release resources
          {
            mdtm_monitor_queue_t *        monitor_q = monitor->monitor_q;
            void *                        m;

            if(monitor->dirq_info) {
            	pthread_mutex_unlock(&monitor->dirq_info->lock);
                pthread_mutex_destroy(&monitor->dirq_info->lock);
                mdtm_dir_queue_destroy(monitor->dirq_info->dir_q);
                if(monitor->dirq_info->pending_segments)
                  globus_free(monitor->dirq_info->pending_segments);
                globus_free(monitor->dirq_info);
                monitor->dirq_info = NULL;
            }

            mdtm_monitor_pool_destroy((mdtm_monitor_pool_t*)(monitor->monitor_pool));
            while( (m = globus_fifo_dequeue(&monitor_q->q_)) != NULL)
              globus_l_gfs_file_monitor_destroy((globus_l_file_monitor_t *)m);

            globus_fifo_destroy(&monitor_q->q_);
            pthread_mutex_destroy(&monitor_q->lock);
            globus_free(monitor_q);
          }
      }
      else {
          // One monitor is done and no files waiting in the directory queue for service,
          // then reclaim the monitor and wait for the rest monitors to finish their jobs

    	  // unlock dirq if existing
    	  if(monitor->dirq_info != NULL)
    		  pthread_mutex_unlock(&monitor->dirq_info->lock);

          // Reclaim the monitor
          globus_fifo_enqueue(&monitor->monitor_q->q_, monitor);
          pthread_mutex_unlock(&monitor->monitor_q->lock);
      }
    goto error_open;
  }

  // if receiving folder, initialize the hash tables
  if (monitor->dirq_info && monitor->mdtm_is_writer)
  {
    mdtm_file_info_t * dst_info;

    dst_info = (mdtm_file_info_t*)globus_malloc(sizeof(mdtm_file_info_t));
    dst_info->bytes_left = 0;
    dst_info->handle = (void*)monitor->file_handle;
    dst_info->url = strdup("");
    dst_info->directio_on = monitor->directio_on;

    assert(globus_hashtable_insert(
        &monitor->handle2handle_hash,
        monitor->pathname,
        (void*)monitor->pathname) == GLOBUS_SUCCESS);
    assert(globus_hashtable_insert(
        &monitor->url2handle_hash,
        monitor->pathname,
        (void*)monitor->file_handle) == GLOBUS_SUCCESS);
    assert(globus_hashtable_insert(
        &monitor->handle2info_hash,
        (void*)monitor->file_handle,
        dst_info) ==  GLOBUS_SUCCESS);
  } // end of receiving folder

  GlobusGFSFileDebugExit();
  return;

error_open:
  GlobusGFSFileDebugExitWithError();
  return;
}

static
void
mdtm_l_gfs_file_open_write_cb(
    globus_xio_handle_t                 handle,
    globus_result_t                     result,
    void *                              user_arg)
{
    mdtm_blocking_info_t *           blocking_info;
    GlobusGFSName(globus_l_gfs_file_open_read_cb);
    GlobusGFSFileDebugEnter();
    
    blocking_info = (mdtm_blocking_info_t *) user_arg;
    if(result != GLOBUS_SUCCESS)
    {
//        result = GlobusGFSErrorWrapFailed(
//            "globus_l_gfs_file_open_write_cb", result);
//        monitor->file_handle = NULL;
        goto error_open;
    }

    mdtm_chown(blocking_info->pathname,
        blocking_info->monitor->session->username);

    globus_mutex_lock(&blocking_info->mutex);
    blocking_info->done = GLOBUS_TRUE;
//    if(error != GLOBUS_NULL)
//    {
//        monitor->err = globus_object_copy(error);
//        monitor->use_err = GLOBUS_TRUE;
//    }
    globus_cond_signal(&blocking_info->cond);
    globus_mutex_unlock(&blocking_info->mutex);

    GlobusGFSFileDebugExit();
    return;
    /* LZ: Hack: End */

error_open:
//    globus_gridftp_server_finished_transfer(monitor->op, result);
//    globus_l_gfs_file_monitor_destroy(monitor);

    GlobusGFSFileDebugExitWithError();
    return;
}

static
void
mdtm_l_gfs_file_open_read_cb(
    globus_xio_handle_t                 handle,
    globus_result_t                     result,
    void *                              user_arg)
{
    GlobusGFSName(globus_l_gfs_file_open_read_cb);
    GlobusGFSFileDebugEnter();

    // NOT USED
    GlobusGFSFileDebugExitWithError();
    return;
}

static
globus_result_t
mdtm_l_gfs_file_open(
    globus_xio_handle_t *               file_handle,
    const char *                        pathname,
    globus_xio_file_flag_t              open_flags,
    void *                              arg)
{
    globus_l_file_monitor_t *           monitor;
    globus_result_t                     result;
    globus_xio_attr_t                   attr;
    globus_xio_stack_t                  stack;
    char *                              perms;
    int                                 timeout;
    GlobusGFSName(globus_l_gfs_file_open);
    GlobusGFSFileDebugEnter();

    monitor = (globus_l_file_monitor_t *) arg;
    
    result = globus_xio_attr_init(&attr);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed("globus_xio_attr_init", result);
        goto error_attr;
    }

#ifdef O_DIRECT
    if(globus_gfs_config_get_bool("direct_io") &&
        monitor->directio_allowed &&
        strcmp(pathname, "/dev/null"))
    {
        open_flags |= O_DIRECT;
        monitor->directio_on = GLOBUS_TRUE;
    }    
#endif

    result = globus_xio_attr_cntl(
        attr,
        globus_l_gfs_file_driver,
        GLOBUS_XIO_FILE_SET_FLAGS,
        open_flags);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed("globus_xio_attr_init", result);
        goto error_cntl;
    }

    if(open_flags & GLOBUS_XIO_FILE_CREAT)
    {
        perms = globus_gfs_config_get_string("perms");
        if(perms != NULL)
        {
            int                             p = 0;
            
            p = strtoul(perms, NULL, 8);
            if(p > 0 || 
                (perms[0] == '0' && perms[1] == '\0'))
            {
                result = globus_xio_attr_cntl(
                    attr,
                    globus_l_gfs_file_driver,
                    GLOBUS_XIO_FILE_SET_MODE,
                    p);
                if(result != GLOBUS_SUCCESS)
                {
                    globus_gfs_log_message(
                        GLOBUS_GFS_LOG_WARN,
                        "Failed to set default permissions to: %s\n", perms);
                }
            }
            else
            {
                globus_gfs_log_message(
                    GLOBUS_GFS_LOG_WARN,
                    "Invalid default permissions: %s\n", perms);
            }
        }
    }

    timeout = globus_gfs_config_get_int("file_timeout");
    if(timeout > 0)
    {
        globus_reltime_t                delay;

        GlobusTimeReltimeSet(delay, timeout, 0);
        result = globus_xio_attr_cntl(
            attr,
            NULL,
            GLOBUS_XIO_ATTR_SET_TIMEOUT_ALL,
            globus_l_gfs_file_timeout_cb,
            &delay,
            NULL);
        if(result != GLOBUS_SUCCESS)
        {
            globus_gfs_log_message(
                GLOBUS_GFS_LOG_WARN,
                "Unable to set file access timeout of %d seconds\n", 
                timeout);
        }
    }
    
    
    result = globus_xio_stack_init(&stack, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed("globus_xio_stack_init", result);
        goto error_stack;
    }
    result = globus_l_gfs_file_make_stack(
        monitor->op, arg, attr, stack, monitor->session);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_push;
    }

    result = globus_xio_handle_create(file_handle, stack);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed("globus_xio_handle_create", result);
        goto error_create;
    }
    
    //LZ: blocking call for file open
    {
      mdtm_blocking_info_t        blocking_info;
      globus_condattr_t           condattr;


      globus_mutex_init(&blocking_info.mutex, GLOBUS_NULL);
      globus_condattr_init(&condattr);
      globus_condattr_setspace(&condattr, mdtm_thread_get_space());
      globus_cond_init(&blocking_info.cond, &condattr);
      blocking_info.done = GLOBUS_FALSE;
      blocking_info.err = GLOBUS_NULL;
      blocking_info.use_err = GLOBUS_FALSE;
      blocking_info.monitor = monitor;
      blocking_info.pathname = strdup(pathname);

      result = globus_xio_register_open(
          *file_handle,
          pathname,
          attr,
          (open_flags & GLOBUS_XIO_FILE_CREAT) ?
              mdtm_l_gfs_file_open_write_cb :
              mdtm_l_gfs_file_open_read_cb,
              (void*) &blocking_info);
      if(result != GLOBUS_SUCCESS)
        {
          result = GlobusGFSErrorWrapFailed("globus_xio_register_open", result);
          globus_free(blocking_info.pathname);
          goto error_register;
        }

      /* wait on cond_wait() for completion */
      globus_mutex_lock(&blocking_info.mutex);

      while(!blocking_info.done)
        {
          globus_cond_wait(&blocking_info.cond, &blocking_info.mutex);
        }

      globus_mutex_unlock(&blocking_info.mutex);

      globus_mutex_destroy(&blocking_info.mutex);
      globus_cond_destroy(&blocking_info.cond);
      globus_free(blocking_info.pathname);
    }
    
    
    globus_xio_attr_destroy(attr);
    globus_xio_stack_destroy(stack);
    
    GlobusGFSFileDebugExit();
    return GLOBUS_SUCCESS;

error_register:
    globus_xio_register_close(*file_handle, NULL, NULL, NULL);
    *file_handle = NULL;
error_create:
error_push:
    globus_xio_stack_destroy(stack);
    
error_stack:
error_cntl:    
    globus_xio_attr_destroy(attr);
    
error_attr:
    GlobusGFSFileDebugExitWithError();
    return result;
}


#if 0
static
void
globus_l_gfs_file_recv(
    globus_gfs_operation_t              op,
    globus_gfs_transfer_info_t *        transfer_info,
    void *                              user_arg)
{
    globus_result_t                     result;
    globus_l_file_monitor_t *           monitor;
    int                                 optimal_count;
    globus_size_t                       block_size;
    globus_xio_file_flag_t              open_flags;
    globus_off_t                        offset;
    globus_off_t                        length;
    GlobusGFSName(globus_l_gfs_file_recv);
    GlobusGFSFileDebugEnter();

    globus_gridftp_server_get_optimal_concurrency(op, &optimal_count);
    globus_gridftp_server_get_block_size(op, &block_size);
    globus_assert(optimal_count > 0 && block_size > 0);


#if 0
    result = globus_l_gfs_file_monitor_init(
        &monitor, block_size, optimal_count);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed(
            "globus_l_gfs_file_monitor_init", result);
        goto error_alloc;
    }

#else
    result = globus_l_gfs_file_monitor_init2(
        &monitor, block_size, optimal_count, transfer_info);
    if(result != GLOBUS_SUCCESS){
        result = GlobusGFSErrorWrapFailed(
            "globus_l_gfs_file_monitor_init", result);
        goto error_alloc;
      }
#endif
    monitor->session = (gfs_l_file_session_t *) user_arg;
    
    globus_gridftp_server_get_write_range(
        op,
        &offset,
        &length);
        
    monitor->op = op;
    monitor->pathname = globus_libc_strdup(transfer_info->pathname);
    
    open_flags = GLOBUS_XIO_FILE_BINARY | 
        GLOBUS_XIO_FILE_CREAT | 
        GLOBUS_XIO_FILE_WRONLY;
    if(transfer_info->truncate)
    {
        open_flags |= GLOBUS_XIO_FILE_TRUNC;
    }
    
    if(transfer_info->expected_checksum)
    {
        monitor->expected_cksm = 
            globus_libc_strdup(transfer_info->expected_checksum);
    }
    if(transfer_info->expected_checksum_alg)
    {
        monitor->expected_cksm_alg = 
            globus_libc_strdup(transfer_info->expected_checksum_alg);
    }
    
    result = globus_l_gfs_file_open(
        &monitor->file_handle, transfer_info->pathname, open_flags, monitor);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed("globus_l_gfs_file_open", result);
        goto error_open;
    }
    
    GlobusGFSFileDebugExit();
    return;

error_open:
    globus_l_gfs_file_monitor_destroy(monitor);
    
error_alloc:
    globus_gridftp_server_finished_transfer(op, result);

    GlobusGFSFileDebugExitWithError();
}

#else

static
mdtm_monitor_pool_t *
mdtm_monitor_pool_create(
    int size)
{
  mdtm_monitor_pool_t *         pool = NULL;

  pool = (mdtm_monitor_pool_t *)globus_malloc(
      sizeof(mdtm_monitor_pool_t) + (size + 1)*sizeof(globus_l_file_monitor_t *));
  if(pool == NULL)
    return NULL;

  if(pthread_mutex_init(&pool->lock, NULL)) {
      globus_free(pool);
      return NULL;
  }
  pool->offset = 0;
  pool->capacity = size + 1;
  pool->monitors = (globus_l_file_monitor_t **)((char*)pool + sizeof(mdtm_monitor_pool_t));
  memset(pool->monitors, 0, (size + 1)*sizeof(globus_l_file_monitor_t *));
  return pool;
}

static
int
mdtm_monitor_pool_add(
    mdtm_monitor_pool_t* pool,
    globus_l_file_monitor_t* monitor)
{
  int rc = -1;

  if(pthread_mutex_lock(&pool->lock) == 0) {
      if(pool->capacity > pool->offset) {
          pool->monitors[pool->offset++] = monitor;
          rc = 0;
      }
      pthread_mutex_unlock(&pool->lock);
  }
  return rc;
}

static
globus_l_file_monitor_t *
mdtm_monitor_pool_entry(
    mdtm_monitor_pool_t * pool,
    int index)
{
  globus_l_file_monitor_t * m = NULL;

  if(pthread_mutex_lock(&pool->lock) == 0) {
      if(index >= 0 && pool->capacity > index + 1) {
          m = pool->monitors[index];
      }
      pthread_mutex_unlock(&pool->lock);
  }
  return m;
}

static
void
mdtm_monitor_pool_destroy(mdtm_monitor_pool_t * pool)
{
  if(pool == NULL)
    return;

  if(pthread_mutex_lock(&pool->lock) == 0) {
      pool->capacity = 0;
      pthread_mutex_unlock(&pool->lock);

      pthread_mutex_destroy(&pool->lock);
      globus_free(pool);
      pool = NULL;
  }
}

static
void
globus_l_gfs_file_recv(
    globus_gfs_operation_t              op,
    globus_gfs_transfer_info_t *        transfer_info,
    void *                              user_arg)
{
    globus_result_t                     result;
    globus_l_file_monitor_t *           monitor;
    int                                 optimal_count;
    globus_size_t                       block_size;
    globus_xio_file_flag_t              open_flags;
    globus_off_t                        offset;
    globus_off_t                        length;
    int                                 i;

    globus_off_t                        size = 0;
    char                                filename[PATH_MAX];
    unsigned int                        recvers, writers;
    mdtm_io_thrd_mgr_t                  *recv_mgr, *writer_mgr;
    int                                 parallel, nrecver, nwriter;
    int                                 rc;
    mdtm_monitor_queue_t *              monitor_q;
    mdtm_dirq_info_t *                  dirq_info = NULL;
    void *                              dirq;

    mdtm_transfer_arg_t *               transfer_arg;
    mdtm_file_entry_t                   entry;
    mdtm_xmit_progress_t *              xmit_prog;
    mdtm_monitor_pool_t *               monitors = 0;
    char *                              file_sys = NULL;
    int                                 numa_of_file = 0;
    int                                 sector_sz = 0;
    globus_bool_t                       splice_on = GLOBUS_FALSE;

    GlobusGFSName(globus_l_gfs_file_recv);
    GlobusGFSFileDebugEnter();

    globus_gridftp_server_get_optimal_concurrency(op, &optimal_count);
    globus_gridftp_server_get_block_size(op, &block_size);
    globus_assert(optimal_count > 0 && block_size > 0);

    // get network io managers
    recv_mgr = mdtm_io_manager(NULL, MDTM_DEVICE_CLASS_NETWORK);
    if ( recv_mgr == 0 ) {
        globus_gfs_log_message(
            GLOBUS_GFS_LOG_ERR,
            "Can not find network io_thread manager.\n");
        result = GlobusGFSErrorGeneric("mdtm network IO manager not found.");
        goto error_alloc;
    }

    // determine the type of file object(regular file or directory),
    // then get the first file to transfer
    dirq = mdtm_i_gfs_data_get_dirq(op);
    if(!dirq) {
        // if single file, pathname is the full path of file;
        size = transfer_info->alloc_size;
        strncpy(filename, transfer_info->pathname, PATH_MAX);
    }
    else if (mdtm_dir_queue_empty(dirq)) {
        globus_gfs_log_message(
            GLOBUS_GFS_LOG_ERR,
            "INFO @%s:%d: EMPTY DIRECTORY OR WRONG META for %s\n", __FILE__,__LINE__,
            transfer_info->pathname);
        result = GlobusGFSErrorGeneric("EMPTY DIRECTORY OR WRONG META.");
        goto error_alloc;
    }
    else {
        if( mdtm_gridftp_server_get_session_type(op) != MDTM_SERVER_SESSION_TYPE_FOLDER)
          goto error_alloc;

        // if directory, pathname is the root path of the directory
        dirq_info = (mdtm_dirq_info_t*)globus_calloc(1, sizeof(mdtm_dirq_info_t));
        pthread_mutex_init(&dirq_info->lock, NULL);
        dirq_info->consumed = 0;
        dirq_info->dir_q = dirq;

        if((rc = mdtm_dir_queue_find3(dirq_info->dir_q, 0, &entry)) != GLOBUS_SUCCESS) {
            globus_gfs_log_message(
                GLOBUS_GFS_LOG_ERR,
                "INFO @%s:%d: mdtm_dir_queue_find3 failed for %s\n", __FILE__,__LINE__,
                transfer_info->pathname);
            result = GlobusGFSErrorGeneric("mdtm_dir_queue_find3.");
        	goto error_alloc;
        }
        strncpy(filename, transfer_info->pathname, PATH_MAX);
        strcpy(strrchr(filename,'/') + 1, entry.path);
        size = entry.size;
    }

    // get disk io managers
    writer_mgr = mdtm_io_manager (filename, MDTM_DEVICE_CLASS_STORAGE);
    if(writer_mgr == NULL) {
        globus_gfs_log_message(
            GLOBUS_GFS_LOG_ERR,
            "Can not find disk io_thread manager for %s.\n", filename);
        result = GlobusGFSErrorGeneric("mdtm writer IO manager not found.");
        goto error_alloc;
    }
    numa_of_file = writer_mgr->devio_info->numaid;

    // get the file system
    if(mdtm_map_getfilesystem(filename, &file_sys) < 0) {
        globus_gfs_log_message(
            GLOBUS_GFS_LOG_ERR,
            "Can not find file system for %s.\n", filename);
        result = GlobusGFSErrorGeneric("file system analysis failed.");
    	goto error_alloc;
    }

    // determine parallel factor by storage type
    result = mdtm_l_get_parallel(recv_mgr, writer_mgr, dirq_info, size, &parallel, 0);
    if(result != GLOBUS_SUCCESS) {
        globus_gfs_log_message(
            GLOBUS_GFS_LOG_ERR,
            "Can not find parallel factor for %s.\n", filename);
        result = GlobusGFSErrorGeneric("mdtm parallel calculation failed.");
    	goto error_alloc;
    }

    //schedule writers and receivers for data transfer
    nwriter  = mdtm_io_thrd_mgr_sched(writer_mgr, parallel, &writers, 0);
    nrecver  = mdtm_io_thrd_mgr_sched(recv_mgr, parallel, &recvers, 0);
    assert(nwriter > 0 && nrecver > 0);
    globus_gfs_log_message(
        GLOBUS_GFS_LOG_INFO,
        "%s: recvers= %s %04x writers= %s %04x\n", __FUNCTION__,
        recv_mgr->devio_info->name, recvers,
        writer_mgr->devio_info->name, writers);

    // save the schedule results
    globus_gridftp_server_set_iothrdmgr_attr(op, recvers, nrecver, recv_mgr);

    // save the client eventq to dc_handle
    globus_gridftp_server_set_client_eventq(op);

    // create monitor arrays
    monitors = mdtm_monitor_pool_create(parallel);
    assert(monitors);

    // start data transfer
    transfer_arg = (mdtm_transfer_arg_t *)globus_malloc(sizeof(mdtm_transfer_arg_t));
    transfer_arg->op = (void*)op;
    transfer_arg->dispatcher = GLOBUS_TRUE;
    transfer_arg->monitors = (void*)monitors;
    globus_gridftp_server_begin_transfer_mt((void*)transfer_arg);
    globus_free(transfer_arg);

    // create transmit progress object
    if(mdtm_monitor_enabled()) {
        xmit_prog = mdtm_xmit_progress_alloc(((gfs_l_file_session_t *)user_arg)->username, op, 0);
        if(!xmit_prog)
          goto error_alloc;
    }
    else
      xmit_prog = 0;

    // initialize monitor queue
    monitor_q = (mdtm_monitor_queue_t*)globus_malloc(sizeof(mdtm_monitor_queue_t));
    globus_fifo_init(&monitor_q->q_);
    pthread_mutex_init(&monitor_q->lock, NULL);

#ifdef HAVE_SPLICE
    /*
     * LZ: Turn on splice for zero-copy data only if the flag is set
     * in the configure file AND the clients require transferring
     * data in non-encryption scheme.
     */
    //TODO: move init pipe pool to mdtm_pipe_get()
    //NOTE: 4* to get around the splice() memory issue
    if(globus_gfs_config_get_bool("splice") &&
        mdtm_gridftp_server_is_clientscheme_encrypted(op) == GLOBUS_FALSE) {
//      assert(mdtm_pipe_init(MDTM_SPLICE_PIPES_PER_NUMA, block_size) == 0);
      splice_on = GLOBUS_TRUE;
    }
#endif

#ifdef O_DIRECT
    if(globus_gfs_config_get_bool("direct_io"))
      sector_sz = mdtm_blockdevice_sectorsize((const char*)transfer_info->pathname);
#endif

    // case #1/2: single file
    for(i = 0; i < parallel && !dirq_info; i++) {
        result = globus_l_gfs_file_monitor_init2(
            &monitor, block_size, optimal_count, numa_of_file);
        if(result != GLOBUS_SUCCESS) {
            result = GlobusGFSErrorWrapFailed(
                "globus_l_gfs_file_monitor_init", result);
            goto error_alloc;
        }

        assert(mdtm_monitor_pool_add(monitors, monitor) == 0);

        if(splice_on == GLOBUS_TRUE) {
            rc = mdtm_pipemem_init(
                &monitor->pipemem,
                block_size,
                optimal_count * 2,
                numa_of_file); //get the pipes of numa
            if(!rc) {
//                globus_free(monitor);
                result = GlobusGFSErrorMemory("pipe");
                goto error_alloc;
            }
        }
        else {
            rc = globus_memory_init2(
                &monitor->mem,
                block_size,
                optimal_count,
                numa_of_file);
            if(!rc) {
//                globus_free(monitor);
                result = GlobusGFSErrorMemory("buffer");
                goto error_alloc;
            }
        }

        monitor->session = (gfs_l_file_session_t *) user_arg;

        globus_gridftp_server_get_write_range(
            op,
            &offset,
            &length);

        monitor->splice_on = splice_on;
        monitor->op = op;
        monitor->pathname = globus_libc_strdup(filename);
        monitor->mt_idx = i;
        monitor->writer_mgr = writer_mgr;
        monitor->iothreads_idx = mdtm_util_nthsetbit(writers, i + 1) - 1;
        monitor->parallel = parallel;
        monitor->monitor_q = monitor_q;
        monitor->monitor_pool = (void*) monitors;
        monitor->transfer_info = (void*)transfer_info;
        monitor->alloc_size = size;
        monitor->xmit_progress = xmit_prog;

        open_flags = GLOBUS_XIO_FILE_BINARY |
            GLOBUS_XIO_FILE_CREAT |
            GLOBUS_XIO_FILE_WRONLY;
//        if(transfer_info->truncate)
//          {
//            open_flags |= GLOBUS_XIO_FILE_TRUNC;
//          }

        monitor->file_system = strdup(file_sys);
        monitor->directio_watermark = MDTM_DIRECTIO_WATERMARK_DEFAULT;

#ifdef O_DIRECT
        if(globus_gfs_config_get_bool("direct_io")) {
        	monitor->directio_sectorsize = sector_sz;
        	if( strstr(file_sys, "tmpfs") == NULL &&
        			size >= monitor->directio_watermark &&
					monitor->splice_on == GLOBUS_FALSE)
        		monitor->directio_allowed = GLOBUS_TRUE;
        	else
        		monitor->directio_allowed = GLOBUS_FALSE;
        }
#endif

        if(transfer_info->expected_checksum)
          {
            monitor->expected_cksm =
                globus_libc_strdup(transfer_info->expected_checksum);
          }
        if(transfer_info->expected_checksum_alg)
          {
            monitor->expected_cksm_alg =
                globus_libc_strdup(transfer_info->expected_checksum_alg);
          }
#if 0
        result = globus_l_gfs_file_open(
            &monitor->file_handle, filename, open_flags, monitor);
        if(result != GLOBUS_SUCCESS)
          {
            result = GlobusGFSErrorWrapFailed("globus_l_gfs_file_open", result);
            goto error_open;
          }
#else
        monitor->open_flags = open_flags;
        assert(
            mdtm_io_thrd_mgr_dispatch_to(
                writer_mgr,
                monitor->iothreads_idx,
                mdtm_l_gfs_file_open_kickout,
                monitor)
             >= 0);
#endif
    }

    // case #2/2: directory
    for(i = 0; i < parallel && dirq_info; i++) {
        result = globus_l_gfs_file_monitor_init2(
            &monitor, block_size, optimal_count, numa_of_file);
        if(result != GLOBUS_SUCCESS) {
            result = GlobusGFSErrorWrapFailed(
                "globus_l_gfs_file_monitor_init", result);
            goto error_alloc;
        }

        assert(mdtm_monitor_pool_add(monitors, monitor) == 0);

        if(splice_on == GLOBUS_TRUE) {
            rc = mdtm_pipemem_init(
                &monitor->pipemem,
                block_size,
                optimal_count * 2,
                numa_of_file); //get the pipes of numa
            if(!rc) {
//                globus_free(monitor);
                result = GlobusGFSErrorMemory("pipe");
                goto error_alloc;
            }
        }
        else {
            rc = globus_memory_init2(
                &monitor->mem,
                block_size,
                optimal_count,
                numa_of_file);
            if(!rc) {
//                globus_free(monitor);
                result = GlobusGFSErrorMemory("buffer");
                goto error_alloc;
            }
        }

        monitor->session = (gfs_l_file_session_t *) user_arg;

        strcpy(filename, "/dev/null");

        monitor->splice_on = splice_on;
        monitor->op = op;
        monitor->pathname = globus_libc_strdup(filename);
        monitor->pathroot = globus_libc_strdup(transfer_info->pathname);
        monitor->multithrd_offset = 0;
        monitor->multithrd_length = -1;
        monitor->mt_idx = i;
        monitor->writer_mgr = writer_mgr;
        monitor->iothreads_idx = mdtm_util_nthsetbit(writers, i + 1) - 1;
        monitor->parallel = parallel;
        monitor->parallel_max = parallel;
        monitor->monitor_q = monitor_q;
        monitor->monitor_pool = (void*) monitors;
        monitor->read_length = -1;      //LZ: TODO: check again
        monitor->dirq_info = dirq_info;
        monitor->transfer_info = (void*)transfer_info;
        monitor->alloc_size = size;
//        monitor->consumed = consumed;
        monitor->mdtm_is_writer = GLOBUS_TRUE;
        monitor->xmit_progress = xmit_prog;

        //Fixme: magic number
        assert(GLOBUS_SUCCESS ==
            globus_hashtable_init(
            &monitor->url2handle_hash,
            MDTM_OPEN_FILE_IN_CACHE_MAX,//GLOBUS_I_CALLOUT_HASH_SIZE,
            globus_hashtable_string_hash,
            globus_hashtable_string_keyeq));

        assert(GLOBUS_SUCCESS ==
            globus_hashtable_init(
                &monitor->handle2info_hash,
                MDTM_OPEN_FILE_IN_CACHE_MAX,//GLOBUS_I_CALLOUT_HASH_SIZE,
                globus_hashtable_voidp_hash,
                globus_hashtable_voidp_keyeq));

        assert(GLOBUS_SUCCESS ==
            globus_hashtable_init(
            &monitor->handle2handle_hash,
            MDTM_OPEN_FILE_IN_CACHE_MAX,//GLOBUS_I_CALLOUT_HASH_SIZE,
            globus_hashtable_string_hash,
            globus_hashtable_string_keyeq));

        open_flags = GLOBUS_XIO_FILE_BINARY |
            GLOBUS_XIO_FILE_CREAT |
            GLOBUS_XIO_FILE_WRONLY;
//        if(transfer_info->truncate)
//            open_flags |= GLOBUS_XIO_FILE_TRUNC;

        monitor->file_system = strdup(file_sys);
        monitor->directio_watermark = MDTM_DIRECTIO_WATERMARK_DEFAULT;

#ifdef O_DIRECT
        if(globus_gfs_config_get_bool("direct_io")) {
        	monitor->directio_sectorsize = sector_sz;
        	if( strstr(file_sys, "tmpfs") == NULL &&
        			size >= monitor->directio_watermark &&
					monitor->splice_on == GLOBUS_FALSE)
        		monitor->directio_allowed = GLOBUS_TRUE;
        	else
        		monitor->directio_allowed = GLOBUS_FALSE;
        }
#endif

#if 0
        result = globus_l_gfs_file_open(
            &monitor->file_handle, filename, open_flags, monitor);
        if(result != GLOBUS_SUCCESS)
          {
            result = GlobusGFSErrorWrapFailed("globus_l_gfs_file_open", result);
            goto error_open;
          }

        dst_info = (mdtm_file_info_t*)globus_malloc(sizeof(mdtm_file_info_t));
        dst_info->bytes_left = entry.size;
        dst_info->handle = (void*)monitor->file_handle;
        dst_info->url = strdup(entry.path);
        if((rc = globus_hashtable_insert(
            &monitor->handle2handle_hash,
            monitor->pathname,
            (void*)monitor->pathname)) < 0)
        {
            assert(0);
        }
        if((rc = globus_hashtable_insert(
            &monitor->url2handle_hash,
            monitor->pathname,
            (void*)monitor->file_handle)) < 0)
        {
            assert(0);
        }
        if((rc = globus_hashtable_insert(
            &monitor->handle2info_hash,
            (void*)monitor->file_handle,
            dst_info)) < 0)
        {
            assert(0);
        }
#else
        monitor->open_flags = open_flags;
        if(mdtm_io_thrd_mgr_dispatch_to(
                writer_mgr,
                monitor->iothreads_idx,
                mdtm_l_gfs_file_open_kickout,
                monitor)
             < 0)
            goto error_open;
#endif
    }


    //LZ: start periodic progress report
    if(mdtm_monitor_enabled())
    {
      globus_reltime_t                  period;
      globus_callback_handle_t *        callback_handle;

      callback_handle = (globus_callback_handle_t *) globus_calloc(1, sizeof(globus_callback_handle_t));
      xmit_prog->callback_handle = (void *) callback_handle;
      xmit_prog->size = 0;
      GlobusTimeReltimeSet(period, 10, 0);
      globus_callback_space_register_periodic(
          xmit_prog->callback_handle,
          0,
          &period,
          mdtm_report_progress_kickout,
          monitor,
          mdtm_thread_get_space());
    }
    // end of progress report

    if(file_sys) free(file_sys);

    GlobusGFSFileDebugExit();
    return;

error_open:
    globus_l_gfs_file_monitor_destroy(monitor);

error_alloc:
    globus_gridftp_server_finished_transfer(op, result);

    if(file_sys) free(file_sys);

    GlobusGFSFileDebugExitWithError();
}
#endif

/**
 * send calls
 */

static
void
globus_l_gfs_file_read_cb(
    globus_xio_handle_t                 xio_handle, 
    globus_result_t                     result,
    globus_byte_t *                     buffer,
    globus_size_t                       len,
    globus_size_t                       nbytes, 
    globus_xio_data_descriptor_t        data_desc,
    void *                              user_arg);


/* called LOCKED */
static
globus_result_t
globus_l_gfs_file_dispatch_read(
    globus_l_file_monitor_t *           monitor)
{
    globus_result_t                     result;
    globus_byte_t *                     buffer;
    globus_size_t                       read_length;
    GlobusGFSName(globus_l_gfs_file_dispatch_read);
    GlobusGFSFileDebugEnter();


    if(monitor->first_read && monitor->pending_reads == 0 && 
        !monitor->eof && !globus_list_empty(monitor->buffer_list) &&
        !monitor->aborted)
    {
        //TODO: LZ: fix op confliction
//        globus_gridftp_server_get_read_range(
//            monitor->op,
//            &monitor->read_offset,
//            &monitor->read_length);
//        monitor->read_offset = 0;
//        monitor->read_length = -1;
//        //END TODO

        if(monitor->read_length == 0)
        {
            monitor->eof = GLOBUS_TRUE;
        }
        else
        {
            // LZ: multithread reading adjust
            monitor->read_offset = monitor->multithrd_offset;
            monitor->read_length = monitor->multithrd_length;
            // End of multithread reading adjust

            if (monitor->file_offset != monitor->read_offset)
            {
                globus_off_t            seek_tmp;
                seek_tmp = monitor->read_offset;
                
                result = globus_xio_handle_cntl(
                    monitor->file_handle,
                    GLOBUS_XIO_QUERY,
                    GLOBUS_XIO_SEEK,
                    seek_tmp,
                    GLOBUS_XIO_FILE_SEEK_SET);
            
                if(result != GLOBUS_SUCCESS)
                {
                    result = GlobusGFSErrorWrapFailed(
                        "globus_xio_handle_cntl", result);
                    goto error_seek;
                }
                monitor->file_offset = monitor->read_offset;
            }
            
        }
        monitor->first_read = GLOBUS_FALSE;
    }             

    if(monitor->pending_reads == 0 && !monitor->eof && 
        !globus_list_empty(monitor->buffer_list) && !monitor->aborted)
    {
        buffer = globus_list_remove(
            &monitor->buffer_list, monitor->buffer_list);
        globus_assert(buffer);

        if(monitor->read_length != -1 &&
            monitor->block_size > monitor->read_length)
        {
            read_length = monitor->read_length;
        }
        else
        {
            read_length = monitor->block_size;
        }

        if(globus_gfs_config_get_bool("mdtm_debug")) {
            gettimeofday(&read_stamp,NULL);
            intervel_stamps[1] += (read_stamp.tv_sec - open_stamp.tv_sec) * 1000 +
                (double)(read_stamp.tv_usec - open_stamp.tv_usec) / 1000;
        }

        if(monitor->splice_on == GLOBUS_TRUE)
          result = globus_xio_register_read_splice(
              monitor->file_handle,
              buffer,
              read_length,
              read_length,
              NULL,
              globus_l_gfs_file_read_cb,
              monitor);
        else
          result = globus_xio_register_read(
              monitor->file_handle,
              buffer,
              read_length,
              read_length,
              NULL,
              globus_l_gfs_file_read_cb,
              monitor);
        if(result != GLOBUS_SUCCESS)
        {
            globus_list_insert(&monitor->buffer_list, buffer);
            result = GlobusGFSErrorWrapFailed(
                "globus_xio_register_read", result);
            goto error_register;
        }

        monitor->pending_reads++;
    }
    
    /* LZ: hack
     *     if buffer is empty, retry later
     */
#if 0
    if(monitor->pending_reads == 0 && !monitor->eof &&
        globus_list_empty(monitor->buffer_list) && !monitor->aborted)
    {
        globus_callback_space_t space;
        globus_callback_handle_t callback_handle;

        globus_reltime_t        delay;
        GlobusTimeReltimeSet(delay, 0, 500);

        //NOTE: globus_callback_space_get() return the latest space in use. Not suitable here
//        globus_callback_space_get(&space);
        space = mdtm_thread_get_space();
        globus_callback_space_register_oneshot(
            &callback_handle,
            &delay, //0,
            mdtm_l_gfs_file_dispatch_read_locked,
//            globus_l_gfs_file_dispatch_read,
            monitor,
            space);
    }
#endif
    /* LZ: hack: end */

    GlobusGFSFileDebugExit();
    return GLOBUS_SUCCESS;

error_seek:
error_register:
    GlobusGFSFileDebugExitWithError();
    return result;
}

//LZ: Added
#if 0
static
void
mdtm_l_gfs_file_dispatch_read_locked(
    void *           arg)
{
  globus_l_file_monitor_t *monitor =(globus_l_file_monitor_t *)arg;

  globus_mutex_lock(&monitor->lock);
  pthread_mutex_lock(&monitor->plock);
  globus_l_gfs_file_dispatch_read(monitor);
  pthread_mutex_unlock(&monitor->plock);
  globus_mutex_unlock(&monitor->lock);

  return;
}

static
void
mdtm_l_gfs_file_dispatch_write_locked(
    void *           arg)
{
  globus_result_t               result;
  globus_l_file_monitor_t *monitor =(globus_l_file_monitor_t *)arg;

  globus_mutex_lock(&monitor->lock);
  pthread_mutex_lock(&monitor->plock);
  result = globus_l_gfs_file_dispatch_write(monitor);
#if 1
  if(result != GLOBUS_SUCCESS)
  {
//      monitor->error = GlobusGFSErrorObjWrapFailed(
//          "globus_l_gfs_file_dispatch_write", result);
//      goto error_dispatch;
  }

  if(monitor->pending_reads == 0 && monitor->pending_writes == 0)
  {
      globus_assert(monitor->eof || monitor->aborted);

      globus_l_gfs_file_close(monitor, GLOBUS_SUCCESS);
      pthread_mutex_unlock(&monitor->plock);
      globus_mutex_unlock(&monitor->lock);
  }
  else
  {
      pthread_mutex_unlock(&monitor->plock);
      globus_mutex_unlock(&monitor->lock);
  }
#else
  pthread_mutex_unlock(&monitor->plock);
  globus_mutex_unlock(&monitor->lock);
#endif

  return;
}
#endif
//LZ: end

static
void
globus_l_gfs_file_server_write_cb(
    globus_gfs_operation_t   op,
    globus_result_t                     result,
    globus_byte_t *                     buffer,
    globus_size_t                       nbytes,
    void *                              user_arg)
{
    globus_l_file_monitor_t *           monitor;
    GlobusGFSName(globus_l_gfs_file_server_write_cb);
    GlobusGFSFileDebugEnter();

    monitor = (globus_l_file_monitor_t *) user_arg;

    globus_mutex_lock(&monitor->lock);
    pthread_mutex_lock(&monitor->plock);        //LZ: added buffer lock
    { 
        monitor->pending_writes--;
        globus_list_insert(&monitor->buffer_list, buffer);

        if(result != GLOBUS_SUCCESS && monitor->error == NULL)
        {
            monitor->error = GlobusGFSErrorObjWrapFailed("callback", result);
        }
        if(monitor->error != NULL)
        {
            goto error;
        }
#if 0
        // LZ: Hack: comment out to disable reading in the current write thread
//        result = globus_l_gfs_file_dispatch_read(monitor);
//        if(result != GLOBUS_SUCCESS)
//        {
//            monitor->error = GlobusGFSErrorObjWrapFailed(
//                "globus_l_gfs_file_dispatch_read", result);
//            goto error;
//        }

//        if(monitor->pending_reads == 0 && monitor->pending_writes == 0)
        if(monitor->pending_reads == 0 && monitor->pending_writes == 0 && (monitor->eof|| monitor->aborted)) //LZ:added
        {
            globus_assert(monitor->eof || monitor->aborted);
//            globus_l_gfs_file_close(monitor, GLOBUS_SUCCESS);
            pthread_mutex_unlock(&monitor->plock);        //LZ: added buffer lock
            globus_mutex_unlock(&monitor->lock);
//            globus_l_gfs_file_close(monitor, GLOBUS_SUCCESS);
            assert(
                mdtm_io_thrd_mgr_dispatch_to(
                    monitor->reader_mgr,
                    monitor->iothreads_idx,
                    globus_l_gfs_file_close_mt,
                    monitor)
                 >= 0);
        }
        else
        {
            pthread_mutex_unlock(&monitor->plock);        //LZ: added buffer lock
            globus_mutex_unlock(&monitor->lock);
        }
#else
        if(mdtm_io_thrd_mgr_dispatch_to(
                            monitor->reader_mgr,
                            monitor->iothreads_idx,
                            mdtm_l_gfs_file_dispatch_read,
                            monitor) >= 0)
        {
            monitor->pending_reads++;                     //LZ: bounce up to prevent racing condition
            pthread_mutex_unlock(&monitor->plock);        //LZ: added buffer lock
            globus_mutex_unlock(&monitor->lock);
        }
        else
          goto error;
#endif
    }
    /* already unlocked */
    GlobusGFSFileDebugExit();
    return;

error:
    if(monitor->pending_reads != 0 || monitor->pending_writes != 0)
    {
        /* there are still outstanding callbacks, wait for them */
        pthread_mutex_unlock(&monitor->plock);        //LZ: added buffer lock
        globus_mutex_unlock(&monitor->lock);
    
        GlobusGFSFileDebugExitWithError();
        return;
    }
    globus_l_gfs_file_close(monitor, globus_error_put(monitor->error));
    pthread_mutex_unlock(&monitor->plock);        //LZ: added buffer lock
    globus_mutex_unlock(&monitor->lock);

    GlobusGFSFileDebugExitWithError();
}

/* LZ: addded */
#if 0
static
void
globus_gridftp_server_register_write2(void* data1, void* data2, int data3)
{
  globus_result_t       result;
  globus_l_file_monitor_t *monitor = (globus_l_file_monitor_t *)data1;
  globus_byte_t * buffer = (globus_byte_t *) data2;
  globus_size_t nbytes = (globus_size_t)data3;

  result = globus_gridftp_server_register_write(
      monitor->op,
      buffer,
      nbytes,
      monitor->file_offset,
      -1,
      globus_l_gfs_file_server_write_cb,
      monitor);
  if(result != GLOBUS_SUCCESS)
  {
      globus_list_insert(&monitor->buffer_list, buffer);
//      monitor->error = GlobusGFSErrorObjWrapFailed(
//          "globus_gridftp_server_register_write", result);
      goto error;
  }

//  monitor->pending_writes++;
//  monitor->file_offset += nbytes;
//  if(monitor->read_length != -1)
//  {
//      monitor->read_length -= nbytes;
//  }
 error:
   return;
}
#endif
/* LZ: end */

static
void
globus_l_gfs_file_read_cb(
    globus_xio_handle_t                 xio_handle, 
    globus_result_t                     result,
    globus_byte_t *                     buffer,
    globus_size_t                       len,
    globus_size_t                       nbytes, 
    globus_xio_data_descriptor_t        data_desc,
    void *                              user_arg)
{
  //LZ: DEBUG
    static int hitcount = 0;
    static struct timeval earlier;
    static struct timeval later;
    static struct timeval interval;

    static int mt_count1 = 0, emt_count1 = 0;
    static int mt_count2 = 0, emt_count2 = 0;
    static int mt_count3 = 0;
    //LZ: END DEBUG

    globus_l_file_monitor_t *           monitor;
    GlobusGFSName(globus_l_gfs_file_read_cb);
    GlobusGFSFileDebugEnter();
    
    monitor = (globus_l_file_monitor_t *) user_arg;
    
    //LZ: debug
    if(globus_gfs_config_get_bool("mdtm_debug")) {
        gettimeofday(&close_stamp,NULL);
        intervel_stamps[2] += (close_stamp.tv_sec - read_stamp.tv_sec) * 1000 +
            (double)(close_stamp.tv_usec - read_stamp.tv_usec) / 1000;

        switch(monitor->mt_idx) {
        case 0:
          mt_count1++;
          if(globus_list_empty(monitor->buffer_list)) emt_count1++;
          break;
        case 1:
          mt_count2++;
          if(globus_list_empty(monitor->buffer_list)) emt_count2++;
          break;
        default:
          mt_count3++;
          break;
        }

        if(hitcount == 0) {
            gettimeofday(&earlier,NULL);
            hitcount++;
        }
        else if(hitcount >= 1000){
            gettimeofday(&later,NULL);
            interval.tv_sec = later.tv_sec - earlier.tv_sec;
            interval.tv_usec = later.tv_usec - later.tv_usec;
            printf("Duration %lld second, %lld usecond: c1=%d,c2=%d,c3=%d, e1=%d, e2=%d list=%d nbytes=%d\n",
                (long long)interval.tv_sec, (long long)interval.tv_sec,
                mt_count1,mt_count2,mt_count3,
                emt_count1, emt_count2,
                globus_list_size(monitor->buffer_list),
                (int)nbytes);
            hitcount = 0;
        }
        else
          hitcount++;
    }
    //LZ: end of debug

    globus_mutex_lock(&monitor->lock);
    pthread_mutex_lock(&monitor->plock);        //LZ: added buffer lock
    {
        monitor->pending_reads--;
        if(result != GLOBUS_SUCCESS && monitor->error == NULL)
        {
            //LZ: hack: comment code to treat everything as eof event
            // TODO: handling differnt error situations
            if(globus_xio_error_is_eof(result))
            {
                monitor->eof = GLOBUS_TRUE;
            }
            else
            {
                monitor->error = GlobusGFSErrorObjWrapFailed(
                    "callback", result);
            }
            //LZ: end
        }
        if(monitor->error != NULL)
        {
            globus_list_insert(&monitor->buffer_list, buffer);
            goto error;
        }

#if 1
        // Normal process
        if(nbytes > 0)
        {
            /* LZ: Hack */
#if 0
            // LZ: hack: use different write cb function
            result = mdtm_gridftp_server_write(monitor, buffer, nbytes, globus_gridftp_server_register_write2);
            if(result != GLOBUS_SUCCESS)
            {
                monitor->error = GlobusGFSErrorObjWrapFailed(
                    "mdtm_gridftp_server_write", result);
                goto error;
            }
            monitor->pending_writes++;
            monitor->file_offset += nbytes;
            if(monitor->read_length != -1)
            {
                monitor->read_length -= nbytes;
            }
#else

#if 0
            result = globus_gridftp_server_register_write(
                monitor->op,
                buffer,
                nbytes,
                monitor->file_offset,
                -1,
                globus_l_gfs_file_server_write_cb,
                monitor);
#else
            // multi-threading
            result = globus_gridftp_server_register_write_mt(
                monitor->op,
                buffer,
                nbytes,
                monitor->consumed + monitor->file_offset,
                -1,
                globus_l_gfs_file_server_write_cb,
                monitor,
                monitor->mt_idx);
#endif
            if(result != GLOBUS_SUCCESS)
            {
                globus_list_insert(&monitor->buffer_list, buffer);
                monitor->error = GlobusGFSErrorObjWrapFailed(
                    "globus_gridftp_server_register_write_mt", result);
                goto error;
            }

            monitor->pending_writes++;
            monitor->file_offset += nbytes;
            if(monitor->read_length != -1)
            {
            	monitor->read_length -= nbytes;
            }
#endif
        }
#else
        // Debug reading process: shortcut sending process by return buffer to list immediately
        if(nbytes > 0)
         {
             /* LZ: Hack */
 #if 0
             // LZ: hack: use different write cb function
             result = mdtm_gridftp_server_write(monitor, buffer, nbytes, globus_gridftp_server_register_write2);
             if(result != GLOBUS_SUCCESS)
             {
                 monitor->error = GlobusGFSErrorObjWrapFailed(
                     "mdtm_gridftp_server_write", result);
                 goto error;
             }
             monitor->pending_writes++;
             monitor->file_offset += nbytes;
             if(monitor->read_length != -1)
             {
                 monitor->read_length -= nbytes;
             }
 #else
//             result = globus_gridftp_server_register_write(
//                 monitor->op,
//                 buffer,
//                 nbytes,
//                 monitor->file_offset,
//                 -1,
//                 globus_l_gfs_file_server_write_cb,
//                 monitor);
//             if(result != GLOBUS_SUCCESS)
//             {
//                 globus_list_insert(&monitor->buffer_list, buffer);
//                 monitor->error = GlobusGFSErrorObjWrapFailed(
//                     "globus_gridftp_server_register_write", result);
//                 goto error;
//             }
//
//             monitor->pending_writes++;
             monitor->file_offset += nbytes;
             if(monitor->read_length != -1)
             {
                 monitor->read_length -= nbytes;
             }

#ifdef HAVE_SPLICE
             if(monitor->splice_on == GLOBUS_TRUE)
             {
               int rc;

               rc = mdtm_pipe_drain(
                     ((mdtm_pipefd_t*)buffer)->pipefd[0],
                     nbytes,
                     MDTM_PIPEDRAIN_DRAIN);
               if(rc < 0)
                 printf("splice error:%d\n", errno);
             }
#endif

             //return buffer
             globus_list_insert(&monitor->buffer_list, buffer);

 #endif
         }
#endif  //Debug reading process
        else
        {
            globus_list_insert(&monitor->buffer_list, buffer);
        }

//        printf("%s %d: monitor=%p file=%s nbytes=%d buffer size=%d\n",
//            __FILE__,__LINE__,
//            monitor, monitor->pathname,nbytes,
//            globus_list_size(monitor->buffer_list));
                    
        if(monitor->read_length == 0)
        {
            monitor->first_read = GLOBUS_TRUE;
        }

//        gettimeofday(&close_stamp,NULL);
//        intervel_stamps[2] += (close_stamp.tv_sec - read_stamp.tv_sec) * 1000 +
//                              (double)(close_stamp.tv_usec - read_stamp.tv_usec) / 1000;
        
        result = globus_l_gfs_file_dispatch_read(monitor);
        if(result != GLOBUS_SUCCESS)
        {
            monitor->error = GlobusGFSErrorObjWrapFailed(
                "globus_l_gfs_file_dispatch_read", result);
            goto error;
        }
       
        if(monitor->pending_reads == 0 && monitor->pending_writes == 0)
        {
            globus_assert(monitor->eof || monitor->aborted);
            globus_l_gfs_file_close(monitor, GLOBUS_SUCCESS);
            pthread_mutex_unlock(&monitor->plock);        //LZ: added buffer lock
            globus_mutex_unlock(&monitor->lock);
//            globus_l_gfs_file_close(monitor, GLOBUS_SUCCESS);
        }
        else
        {
            pthread_mutex_unlock(&monitor->plock);        //LZ: added buffer lock
            globus_mutex_unlock(&monitor->lock);
        }
    }
    /* already unlocked */
    GlobusGFSFileDebugExit();
    return;

error:
//    globus_assert(monitor->pending_reads == 0);
    if(monitor->pending_writes != 0)
    {
        /* there are still outstanding callbacks, wait for them */
        pthread_mutex_unlock(&monitor->plock);        //LZ: added buffer lock
        globus_mutex_unlock(&monitor->lock);
    
        GlobusGFSFileDebugExitWithError();
        return;
    }
    globus_l_gfs_file_close(monitor, globus_error_put(monitor->error));
    pthread_mutex_unlock(&monitor->plock);        //LZ: added buffer lock
    globus_mutex_unlock(&monitor->lock);

    GlobusGFSFileDebugExitWithError();
}

/* LZ: */
static
void
mdtm_l_gfs_file_dispatch_read(void *user_arg)
{
  globus_result_t result;
  globus_l_file_monitor_t *monitor;
  GlobusGFSName(mdtm_l_gfs_file_dispatch_read);
  GlobusGFSFileDebugEnter();

  monitor = (globus_l_file_monitor_t *) user_arg;

  globus_mutex_lock(&monitor->lock);
  pthread_mutex_lock(&monitor->plock);        //LZ: added buffer lock

  monitor->pending_reads--;     //LZ: cancel the increment from globus_l_gfs_file_server_write_cb()
  result = globus_l_gfs_file_dispatch_read(monitor);
  if(result != GLOBUS_SUCCESS)
  {
    monitor->error = GlobusGFSErrorObjWrapFailed(
        "globus_l_gfs_file_dispatch_read", result);
    goto error;
  }

//  if(monitor->pending_reads == 0 && monitor->pending_writes == 0)
  if(monitor->pending_reads == 0 && monitor->pending_writes == 0 && (monitor->eof|| monitor->aborted)) //LZ:added
  {
    globus_assert(monitor->eof || monitor->aborted);
    globus_l_gfs_file_close(monitor, GLOBUS_SUCCESS);
    pthread_mutex_unlock(&monitor->plock);        //LZ: added buffer lock
    globus_mutex_unlock(&monitor->lock);
  }
  else {
      pthread_mutex_unlock(&monitor->plock);        //LZ: added buffer lock
      globus_mutex_unlock(&monitor->lock);
  }
  /* already unlocked */
  GlobusGFSFileDebugExit();
  return;

error:
  if(monitor->pending_reads != 0 || monitor->pending_writes != 0)
  {
      /* there are still outstanding callbacks, wait for them */
      pthread_mutex_unlock(&monitor->plock);        //LZ: added buffer lock
      globus_mutex_unlock(&monitor->lock);

      GlobusGFSFileDebugExitWithError();
      return;
  }
  globus_l_gfs_file_close(monitor, globus_error_put(monitor->error));
  pthread_mutex_unlock(&monitor->plock);        //LZ: added buffer lock
  globus_mutex_unlock(&monitor->lock);

  GlobusGFSFileDebugExitWithError();
}

//static
//void
//globus_l_gfs_file_close_mt(void *user_arg)
//{
//  globus_l_gfs_file_close((globus_l_file_monitor_t *)user_arg, GLOBUS_SUCCESS);
//}

//static
void
globus_l_gfs_file_dispatch2(void *user_arg)
{
  globus_result_t result;
  globus_l_file_monitor_t *monitor;
  GlobusGFSName(globus_l_gfs_file_dispatch2);
  GlobusGFSFileDebugEnter();

  monitor = (globus_l_file_monitor_t *) user_arg;

//  printf("[%lu]%s:%d: %s started\n",syscall(SYS_gettid), __FILE__,__LINE__,__FUNCTION__);

  // LZ: hack change the dc_channel attr space
#if 0
  {
    mdtm_globus_gridftp_server_set_space(monitor->op, 1);//mdtm_thread_get_space());
    globus_gfs_operation_t op;
    op = monitor->op;
    globus_ftp_control_handle_t *handle = &op->data_handle->data_channel;
    globus_i_ftp_dc_handle_t *dc_handle = &handle->dc_handle;
    globus_io_attr_t attr = dc_handle->io_attr;
          attr->space;
          printf("%d\n", 1); // mdtm_thread_get_space();
  }
#endif
  // LZ: end

  globus_mutex_lock(&monitor->lock);
  pthread_mutex_lock(&monitor->plock);  //LZ: TODO: check the mutex everywhere
  monitor->first_read = GLOBUS_TRUE;

  mdtm_globus_xio_handle_update_space(
      &monitor->file_handle,
      mdtm_thread_get_space());

  result = globus_l_gfs_file_dispatch_read(monitor);
  if(result != GLOBUS_SUCCESS)
  {
      monitor->error = GlobusGFSErrorObjWrapFailed(
          "globus_l_gfs_file_dispatch_read", result);
      goto error_dispatch2;
  }

  if(monitor->pending_reads == 0 && monitor->pending_writes == 0)
  {
      globus_assert(monitor->eof || monitor->aborted);

      globus_l_gfs_file_close(monitor, GLOBUS_SUCCESS);
      pthread_mutex_unlock(&monitor->plock);
      globus_mutex_unlock(&monitor->lock);
  }
  else
  {
      pthread_mutex_unlock(&monitor->plock);
      globus_mutex_unlock(&monitor->lock);
  }

  GlobusGFSFileDebugExit();
  return;

error_dispatch2:
  pthread_mutex_unlock(&monitor->plock);
  globus_mutex_unlock(&monitor->lock);


  globus_gridftp_server_finished_transfer(monitor->op, result);
  globus_l_gfs_file_monitor_destroy(monitor);

  GlobusGFSFileDebugExitWithError();

  return;
}

void
globus_gridftp_server_begin_transfer2(void *user_arg)
{
//  globus_result_t result;
  globus_l_file_monitor_t *monitor;
  GlobusGFSName(globus_gridftp_server_begin_transfer2);
  GlobusGFSFileDebugEnter();

  printf("DBG @%s:%d: %s started\n",__FILE__,__LINE__, __FUNCTION__);

  monitor = (globus_l_file_monitor_t *) user_arg;
  globus_gridftp_server_begin_transfer(
        monitor->op, GLOBUS_GFS_EVENT_TRANSFER_ABORT, monitor);

  GlobusGFSFileDebugExit();
}

void
globus_gridftp_server_begin_transfer3(void *user_arg)
{
//  globus_result_t result;
  globus_gfs_operation_t op;
  mdtm_transfer_arg_t   *tran_arg;
  GlobusGFSName(globus_gridftp_server_begin_transfer3);
  GlobusGFSFileDebugEnter();

  printf("DBG @%s:%d: %s started\n",__FILE__,__LINE__, __FUNCTION__);

  tran_arg = (mdtm_transfer_arg_t*) user_arg;
  op = (globus_gfs_operation_t) tran_arg->op;
//  op->data_handle->data_channel.dc_handle = 0;
  globus_gridftp_server_begin_transfer(
        op, GLOBUS_GFS_EVENT_TRANSFER_ABORT, 0);

  GlobusGFSFileDebugExit();
}

//static
//int
//mdtm_file_monitor_spawn(globus_l_file_monitor_t *parent,globus_l_file_monitor_t *child)
//{
//
//  globus_l_file_monitor_t *           monitor;
//  globus_result_t                     result;
//  int                                 rc;
//  int                                 numa;
//
//  numa = parent->file_numa;
//  rc = globus_memory_init2(&child->mem, parent->block_size, parent->optimal_count, numa);
//  if(!rc)
//  {
//      result = GlobusGFSErrorMemory("buffer");
//      goto error_alloc;
//  }
//
//  globus_mutex_init(&child->lock, NULL);
//  globus_priority_q_init(
//      &child->queue, globus_l_gfs_file_queue_compare);
//  child->buffer_list = NULL;
//  child->op = NULL;
//  child->file_handle = NULL;
//  child->pending_reads = 0;
//  child->pending_writes = 0;
//  child->file_offset = 0;
//  child->block_size = parent->block_size;
//  child->optimal_count = parent->optimal_count;
//  child->error = NULL;
//  child->eof = GLOBUS_FALSE;
//  child->aborted = GLOBUS_FALSE;
//  child->concurrency_check = 2;
//  child->concurrency_check_interval = 2;
//  child->expected_cksm = NULL;
//  child->expected_cksm_alg = NULL;
//  child->pathname = NULL;
//
//  child->file_numa = -1;
//  child->parent = parent;
//  child->child = 0;
//  parent->child = child;
//
//  //LZ: added
//  if(pthread_mutex_init(&child->plock, NULL) != 0)
//    printf("DEBUG: %s:%d: mutex failed\n",__FILE__,__LINE__);
//
////  GlobusGFSFileDebugExit();
//  return GLOBUS_SUCCESS;
//
//error_alloc:
////  GlobusGFSFileDebugExitWithError();
//  return result;
//}

//static
//int
//mdtm_gfs_file_split(globus_l_file_monitor_t *monitor, globus_l_file_monitor_t **minimonitors, int ref)
//{
//  int n = 0, result;
//
//  if (ref <= 1 || monitor->read_length < 100000000L) {
//      n = 1;
//      *minimonitors = monitor;
//  }
//  else{
//      globus_l_file_monitor_t *temp;
//      temp = (globus_l_file_monitor_t *) malloc(ref * sizeof(globus_l_file_monitor_t));
//      for(int i=0; i<ref; i++) {
//          result = mdtm_file_monitor_spawn(monitor, &temp[i]);
//      }
//      *minimonitors = temp;
//  }
//  return n;
//}
/* LZ: End */

static
void
globus_l_gfs_file_open_read_cb(
    globus_xio_handle_t                 handle,
    globus_result_t                     result,
    void *                              user_arg)
{
    globus_l_file_monitor_t *           monitor;
    GlobusGFSName(globus_l_gfs_file_open_read_cb);
    GlobusGFSFileDebugEnter();
    
    monitor = (globus_l_file_monitor_t *) user_arg;
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorWrapFailed(
            "globus_l_gfs_file_open_write_cb", result);
        monitor->file_handle = NULL;
        goto error_open;
    }

    // LZ: Hack: comment out the original code
#if 0
    globus_gridftp_server_begin_transfer(
        monitor->op, GLOBUS_GFS_EVENT_TRANSFER_ABORT, monitor);

    globus_mutex_lock(&monitor->lock);
    monitor->first_read = GLOBUS_TRUE;
    result = globus_l_gfs_file_dispatch_read(monitor);
    if(result != GLOBUS_SUCCESS)
    {
        monitor->error = GlobusGFSErrorObjWrapFailed(
            "globus_l_gfs_file_dispatch_read", result);
        goto error_dispatch;
    }
    
    if(monitor->pending_reads == 0 && monitor->pending_writes == 0)
    {
        globus_assert(monitor->eof || monitor->aborted);

        globus_l_gfs_file_close(monitor, GLOBUS_SUCCESS);
        globus_mutex_unlock(&monitor->lock);
    }
    else
    {
        globus_mutex_unlock(&monitor->lock);
    }
    
    GlobusGFSFileDebugExit();
    return;

error_dispatch:
    globus_mutex_unlock(&monitor->lock);
#endif

    // mdtm: added
#if 1
    // multiple thread version
    {
      int                   rc;
      char                  devname[PATH_MAX];
      mdtm_io_thrd_mgr_t    *reader_mgr;

      //look for the reader_mgr based of devname if it is not available
      if(!(reader_mgr = monitor->reader_mgr))
      {
		reader_mgr = mdtm_io_manager(monitor->pathname, MDTM_DEVICE_CLASS_STORAGE);
		if(reader_mgr == NULL) {
			globus_gfs_log_message(
				GLOBUS_GFS_LOG_ERR,
				"Can not find storage io_thread manager for %s.\n", monitor->pathname);
			result = GlobusGFSErrorGeneric("mdtm reader IO manager not found.");
			goto error_open;
		}

//        mdtm_lv_seg_t *segments;
//        rc = mdtm_map_file2(strdup(monitor->pathname), 0, &segments);
//        if(rc < 0) {
//      	  if(rc == MDTM_ERR_INVADL)
//      		  globus_gfs_log_message(
//      				  GLOBUS_GFS_LOG_ERR,
//  					  "mdtm_map_file2 failed @ %s:%d: cannot map %s duo to file system parsing failure.\n",  __FILE__,__LINE__, (monitor->pathname==NULL)? "":monitor->pathname);
//      	  if(rc == MDTM_ERR_NOTSUPPORT)
//      		  globus_gfs_log_message(
//      				  GLOBUS_GFS_LOG_ERR,
//  					  "mdtm_map_file2 failed @ %s:%d: cannot map %s due to lack of system functions like lvm.\n",  __FILE__,__LINE__, (monitor->pathname==NULL)? "":monitor->pathname);
//      	  if(rc == MDTM_ERR_NOMEM)
//      		  globus_gfs_log_message(
//      				  GLOBUS_GFS_LOG_ERR,
//  					  "mdtm_map_file2 failed @ %s:%d: cannot map %s due to insufficient memory.\n",  __FILE__,__LINE__, (monitor->pathname==NULL)? "":monitor->pathname);
//      	  if(rc == MDTM_ERR_NOEXTENT)
//      		  globus_gfs_log_message(
//      				  GLOBUS_GFS_LOG_ERR,
//  					  "mdtm_map_file2 failed @ %s:%d: cannot map %s due to inability to find extents.\n",  __FILE__,__LINE__, (monitor->pathname==NULL)? "":monitor->pathname);
//            result = GlobusGFSErrorGeneric("mdtm_map_file failed");
//            goto error_open;
//        }
//        else if (rc == 0) {
//            // could be sparse file
//            reader_mgr = mdtm_io_thrd_mgr_list_get2(NULL, MDTM_DEVICE_CLASS_STORAGE);
//        }
//        else {
//            strcpy(devname, segments[0].dev);
//            reader_mgr = mdtm_io_thrd_mgr_list_get(devname);
//            mdtm_map_file_free(segments, rc);
//        }
//        if(!reader_mgr) {
//        	globus_gfs_log_message(GLOBUS_GFS_LOG_ERR, "Error @%s:%d: no storage devices in the list. Check mdtmconfig.xml. \n", __FILE__,__LINE__);
//        	result = GlobusGFSErrorGeneric("mdtm reader IO manager not found");
//            goto error_open;
//        }
      }

      if(globus_gfs_config_get_bool("mdtm_debug")) {
          gettimeofday(&open_stamp,NULL);
          intervel_stamps[0] += (open_stamp.tv_sec - start_stamp.tv_sec) * 1000 +
              (double)(open_stamp.tv_usec - start_stamp.tv_usec) / 1000;
      }

      if(mdtm_io_thrd_mgr_dispatch_to(
              reader_mgr,
              monitor->iothreads_idx,
              globus_l_gfs_file_dispatch2,
              monitor) < 0)
      {
          fprintf(stderr, "Error @%s:%d: mdtm_io_thrd_mgr_dispatch_to() failed.\n", __FILE__,__LINE__);
          goto error_open;
      }
    } // end of multiple thread version
#else   // single thread version

    // send transfer request to the event queue of network thread
    mdtm_ftp_dispatch_write(monitor, globus_gridftp_server_begin_transfer2);

    //wait for a period for the connection established
    //TODO: use signaling rather than sleep function
    {
      int seconds = 3;  //TODO: try to use signal rather than sleep()
      printf("DEBUG @%s:%d: Take a %d seconds break for establish stable transfer connections.\n",__FILE__,__LINE__,seconds);
      sleep(seconds);
    }

    /* dispatch file reading to disk thread */
    if(mdtm_file_dispatch_read(monitor, strdup(monitor->pathname), globus_l_gfs_file_dispatch2) != GLOBUS_SUCCESS)
      goto error_open;
#endif

    GlobusGFSFileDebugExit();
    return;
    /* LZ: Hack: End */

error_open:
    globus_gridftp_server_finished_transfer(monitor->op, result);
    globus_l_gfs_file_monitor_destroy(monitor);

    GlobusGFSFileDebugExitWithError();
}


static
void
mdtm_report_progress_kickout(void *data)
{
  // reporting the progress
  in_addr_t                     addr;
  char                          *remote_ip = 0, *local_ip = 0;
  mdtm_io_thrd_mgr_t *          man_mgr;
  globus_l_file_monitor_t *     monitor = (globus_l_file_monitor_t *)data;
  globus_gfs_operation_t        op;

  if(!monitor || !monitor->xmit_progress)
    return;

  op = monitor->op;

  // set current src & dst IPs; return if anyone unavailable
  if((monitor->xmit_progress->src == 0) ||
      (monitor->xmit_progress->dst == 0)) {
      if(globus_gridftp_server_get_remote_ip(op, &remote_ip) &&
          globus_gridftp_server_get_local_ip(op, &local_ip)) {
          if(monitor->xmit_progress->in0out1) {
              inet_pton(AF_INET, local_ip, &addr);
              addr =
                  ((addr & 0xff000000) >> 24) |
                  ((addr & 0x00ff0000) >>  8) |
                  ((addr & 0x0000ff00) <<  8) |
                  ((addr & 0x000000ff) << 24);
              monitor->xmit_progress->src = addr;
              inet_pton(AF_INET,remote_ip, &addr);
              addr =
                  ((addr & 0xff000000) >> 24) |
                  ((addr & 0x00ff0000) >>  8) |
                  ((addr & 0x0000ff00) <<  8) |
                  ((addr & 0x000000ff) << 24);
              monitor->xmit_progress->dst = addr;
          }
          else {
              inet_pton(AF_INET, remote_ip, &addr);
              addr =
                  ((addr & 0xff000000) >> 24) |
                  ((addr & 0x00ff0000) >>  8) |
                  ((addr & 0x0000ff00) <<  8) |
                  ((addr & 0x000000ff) << 24);
              monitor->xmit_progress->src = addr;
              inet_pton(AF_INET, local_ip, &addr);
              addr =
                  ((addr & 0xff000000) >> 24) |
                  ((addr & 0x00ff0000) >>  8) |
                  ((addr & 0x0000ff00) <<  8) |
                  ((addr & 0x000000ff) << 24);
              monitor->xmit_progress->dst = addr;
          }
      }
      else
        // src or dst IP unavailabe, retry later
        goto kickout_done;
  }

  // set transferred bytes
  monitor->xmit_progress->xfr = globus_gridftp_server_get_bytes_transferred(monitor->op);

  // send progress report to the man thread
  man_mgr = mdtm_io_thrd_mgr_list_get("man");
  assert(
      mdtm_io_thrd_mgr_dispatch_to(
          man_mgr,
          0,//monitor->iothreads_idx,
          mdtm_monitor_log_progress,
          monitor->xmit_progress)
          >= 0);

kickout_done:
  if(remote_ip) free(remote_ip);
  if(local_ip) free(local_ip);
  return;
}

static
void
mdtm_report_progress_finished_kickout(void *data)
{
  // reporting the progress
  mdtm_io_thrd_mgr_t *man_mgr;
  mdtm_finish_trans_info_t * info = (mdtm_finish_trans_info_t *)data;

  man_mgr = mdtm_io_thrd_mgr_list_get("man");
  info->xmit_progress->xfr = globus_gridftp_server_get_bytes_transferred(info->op);

  // send progress report to the man thread
  assert(
      mdtm_io_thrd_mgr_dispatch_to(
          man_mgr,
          0,//monitor->iothreads_idx,
          mdtm_monitor_log_progress_finished,
          info->xmit_progress)
          >= 0);
}

static
uid_t name_to_uid(char const *name)
{
  if (!name)
    return -1;
//  long const buflen = sysconf(_SC_GETPW_R_SIZE_MAX);
  long const buflen = 128;
  if (buflen == -1)
    return -1;
  // requires c99
  char buf[buflen];
  struct passwd pwbuf, *pwbufp;
  if (0 != getpwnam_r(name, &pwbuf, buf, buflen, &pwbufp)
      || !pwbufp)
    return -1;
  return pwbufp->pw_uid;
}

static
mdtm_xmit_progress_t *
mdtm_xmit_progress_alloc(
    char* username,
    globus_gfs_operation_t op,
    int direction)
{
  mdtm_rrd_file_t   *           rrdfile;
  in_addr_t                     addr;
  char                          path[512];
  char                          *remote_ip = 0, *local_ip = 0;
  int                           session_id;
  mdtm_xmit_progress_t *        xmit_progress;
  char *                        root;

  if((xmit_progress = (mdtm_xmit_progress_t*)malloc(sizeof(mdtm_xmit_progress_t))) == 0)
    return 0;

  session_id = rand();
  root = mdtm_monitor_getroot(1);
  snprintf(path, 512, "%smdtm%d.rrd", root, session_id);
  if(root) free(root);
  rrdfile = mdtm_rrdfile_open(path);
  mdtm_rrdfile_setformat(rrdfile, "sid,uid,src,dst,size,xfr");
  xmit_progress->file.rrd_indv = rrdfile;
  xmit_progress->data.rrd = mdtm_rrddata_alloc(rrdfile);
  xmit_progress->sid = session_id;
  xmit_progress->uid = name_to_uid(username);
//  xmit_progress->size = size;
  xmit_progress->xfr = 0;
  strcpy(xmit_progress->filepath, path);
  xmit_progress->in0out1 = direction;
  xmit_progress->src = 0;
  xmit_progress->dst = 0;
  xmit_progress->callback_handle = 0;
  if(globus_gridftp_server_get_remote_ip(op, &remote_ip) &&
      globus_gridftp_server_get_local_ip(op, &local_ip)) {
      if(direction) {
          inet_pton(AF_INET, local_ip, &addr);
          addr =
              ((addr & 0xff000000) >> 24) |
              ((addr & 0x00ff0000) >>  8) |
              ((addr & 0x0000ff00) <<  8) |
              ((addr & 0x000000ff) << 24);
          xmit_progress->src = addr;
          inet_pton(AF_INET,remote_ip, &addr);
          addr =
              ((addr & 0xff000000) >> 24) |
              ((addr & 0x00ff0000) >>  8) |
              ((addr & 0x0000ff00) <<  8) |
              ((addr & 0x000000ff) << 24);
          xmit_progress->dst = addr;
      }
      else {
          inet_pton(AF_INET, remote_ip, &addr);
          addr =
              ((addr & 0xff000000) >> 24) |
              ((addr & 0x00ff0000) >>  8) |
              ((addr & 0x0000ff00) <<  8) |
              ((addr & 0x000000ff) << 24);
          xmit_progress->src = addr;
          inet_pton(AF_INET, local_ip, &addr);
          addr =
              ((addr & 0xff000000) >> 24) |
              ((addr & 0x00ff0000) >>  8) |
              ((addr & 0x0000ff00) <<  8) |
              ((addr & 0x000000ff) << 24);
          xmit_progress->dst = addr;
      }
  }

  if(remote_ip) free(remote_ip);
  if(local_ip) free(local_ip);

  return xmit_progress;
}

/**
 * Parse the path of the object to be sent.
 *
 * If it is a directory, return directory list in dirq as well as the name and size of
 * the first file in filename and size parameters;
 * If it is a file, set dirq to null and return its name and size in filename and size parameters.
 *
 * @param pathname		[IN] the path string
 * @param isfolder		[IN] if the path is to folder
 * @param dirq			[OUT] directory queue pointer
 * @param filename		[OUT] the name of file in the path
 * @parm size			[OUT] the size of the file
 */
static
globus_result_t
mdtm_l_path_parse(
		const char * pathname,
		int	isfolder,
		mdtm_dirq_info_t ** dirq,
		char * filename,
		globus_off_t * size)
{
	globus_result_t	result;
    struct stat sb;
    mdtm_dirq_info_t * dirq_info = NULL;
    mdtm_file_segment_info_t * filesegments;

    GlobusGFSName(mdtm_l_path_stat);
    GlobusGFSFileDebugEnter();

    if(!pathname || !dirq || !filename || !size) {
        globus_gfs_log_message(
            GLOBUS_GFS_LOG_ERR,
            "%s: Invalid arguments.\n", __FUNCTION__);
        result = GlobusGFSErrorWrapFailed("Invalid arguments", GLOBUS_SUCCESS);
        goto error_exit;
    }

    if(lstat(pathname, &sb) != 0) {
        globus_gfs_log_message(
            GLOBUS_GFS_LOG_ERR,
            "Warning: %s %d @thrd %lx: %s: pathname=%s stat failed: %s.\n",
            __FILE__,__LINE__,(unsigned long)pthread_self(), __FUNCTION__,
            pathname, strerror(errno));
        result = GlobusGFSErrorSystemError("lstat", errno);
        goto error_exit;
    }
    switch (sb.st_mode & S_IFMT) {
        case S_IFBLK:
        	strncpy(filename, pathname, PATH_MAX);
        	*size = sb.st_size;
        	globus_gfs_log_message(
        			GLOBUS_GFS_LOG_WARN,"%s: %s is block device.\n", __FUNCTION__, pathname);
        	break;
        case S_IFCHR:
        	strncpy(filename, pathname, PATH_MAX);
        	if(!strcmp(filename, "/dev/zero") || !strcmp(filename, "/dev/random")) {
        		*size = 1000000000000UL;
        		globus_gfs_log_message(
        				GLOBUS_GFS_LOG_WARN,"%s: transfer %lld GB from %s .\n", __FUNCTION__, *size/1000000000, pathname);
        	}
        	else {
        		*size = sb.st_size;
        		globus_gfs_log_message(
        				GLOBUS_GFS_LOG_WARN,"%s: %s is character device.\n", __FUNCTION__, pathname);
        	}
        	break;
        case S_IFDIR:

          if(isfolder != MDTM_SERVER_SESSION_TYPE_FOLDER) {
              globus_gfs_log_message(
                  GLOBUS_GFS_LOG_ERR,
                  "Can not transfer file as its url is a directory @ %s:%d.\n",  __FILE__,__LINE__);
              result = GlobusGFSErrorWrapFailed(
                  "source url is directory", GLOBUS_SUCCESS);
              goto error_exit;
          }

          // create pending file segment info
          filesegments = (mdtm_file_segment_info_t*)globus_malloc(sizeof(mdtm_file_segment_info_t));
          if(filesegments == NULL) {
              result = GlobusGFSErrorMemory("filesegments");
              goto error_exit;
          }
          memset(filesegments, 0, sizeof(mdtm_file_segment_info_t));

          // create the directory queue
          dirq_info = (mdtm_dirq_info_t*)globus_malloc(sizeof(mdtm_dirq_info_t));
          pthread_mutex_init(&dirq_info->lock, NULL);
          dirq_info->pending_segments = filesegments;
          dirq_info->consumed = 0;
          dirq_info->dir_q = mdtm_dir_queue_create(pathname);
          if(mdtm_dir_queue_empty(dirq_info->dir_q)) {
              globus_gfs_log_message(
                  GLOBUS_GFS_LOG_ERR,"INFO @%s:%d: EMPTY DIRECTORY: %s\n", __FILE__,__LINE__, pathname);
              //TODO: any clean up work to do here??
              goto error_exit;
          }

          if(mdtm_dir_queue_peek(dirq_info->dir_q, filename) == NULL) {
              globus_gfs_log_message(
                  GLOBUS_GFS_LOG_ERR,"INFO @%s:%d: Invalid file name in the DIRECTORY: %s\n", __FILE__,__LINE__, pathname);
              goto error_exit;
          }

          if(lstat(filename, &sb) < 0) {
        	  globus_gfs_log_message(
        			  GLOBUS_GFS_LOG_ERR,
					  "Warning: %s %d @thrd %lx: %s: filename=%s stat failed: %s.\n",
					  __FILE__,__LINE__,(unsigned long)pthread_self(), __FUNCTION__,
					  filename, strerror(errno));
        	  result = GlobusGFSErrorSystemError("lstat", errno);
        	  goto error_exit;
          }

          *size = sb.st_size;
          *dirq = dirq_info;

          break;
        case S_IFIFO:
        	strncpy(filename, pathname, PATH_MAX);
        	globus_gfs_log_message(
        			GLOBUS_GFS_LOG_WARN,"INFO @%s:%d: %s is FIFO/pipe.\n", __FILE__,__LINE__, pathname);
        	break;
        case S_IFLNK:
        	strncpy(filename, pathname, PATH_MAX);
        	globus_gfs_log_message(
        			GLOBUS_GFS_LOG_WARN,"INFO @%s:%d: %s is symlink.\n", __FILE__,__LINE__, pathname);
        	break;
        case S_IFREG:
          strncpy(filename, pathname, PATH_MAX);
          *size = sb.st_size;
          break;
        case S_IFSOCK:
        	strncpy(filename, pathname, PATH_MAX);
        	globus_gfs_log_message(
        			GLOBUS_GFS_LOG_WARN,"INFO @%s:%d: %s is socket.\n", __FILE__,__LINE__, pathname);
        	break;
        default:
        	strncpy(filename, pathname, PATH_MAX);
        	globus_gfs_log_message(
        			GLOBUS_GFS_LOG_WARN,"INFO @%s:%d: %s is unknown type.\n", __FILE__,__LINE__, pathname);
        	break;
    }

    GlobusGFSFileDebugExit();
    return GLOBUS_SUCCESS;

error_exit:
	return result;
	GlobusGFSFileDebugExitWithError();
}

static
void
globus_l_gfs_file_send(
    globus_gfs_operation_t              op,
    globus_gfs_transfer_info_t *        transfer_info,
    void *                              user_arg)
{
    globus_result_t                     result;
    globus_l_file_monitor_t *           monitor;
    int                                 optimal_count, optimal_count_saved;
    globus_size_t                       block_size;
    globus_xio_file_flag_t              open_flags;
    mdtm_monitor_queue_t                *monitor_q;
    mdtm_dirq_info_t                    *dirq_info = NULL;
    int                                 dirq_size=0;
    mdtm_file_segment_info_t            *filesegments;

    globus_off_t                        size = 0, segsize, off = 0, consumed = 0;
    char                                filename[PATH_MAX];
    unsigned int                        senders, readers;
    mdtm_io_thrd_mgr_t                  *sender_mgr = NULL, *reader_mgr = NULL;
    int                                 parallel, nsender, nreader, i, rc, numa_of_file = 0;
    struct                              stat sb;

    mdtm_transfer_arg_t *               transfer_arg;
    mdtm_xmit_progress_t *              xmit_prog;
    mdtm_monitor_pool_t *               monitors = 0;
    globus_bool_t                       splice_on = GLOBUS_FALSE;

    GlobusGFSName(globus_l_gfs_file_send);
    GlobusGFSFileDebugEnter();

    globus_gridftp_server_get_optimal_concurrency(op, &optimal_count);
    globus_gridftp_server_get_block_size(op, &block_size);
    globus_assert(optimal_count > 0 && block_size > 0);
    
    optimal_count_saved = optimal_count;

    // get info about the object to be sent
    result = mdtm_l_path_parse(
    		transfer_info->pathname,
			mdtm_gridftp_server_get_session_type(op),
			&dirq_info,
			filename,
			&size);
    if( result != GLOBUS_SUCCESS)
    	goto error_alloc;

    // get disk io manager
    reader_mgr = mdtm_io_manager(filename, MDTM_DEVICE_CLASS_STORAGE);
    if(reader_mgr == NULL) {
        globus_gfs_log_message(
            GLOBUS_GFS_LOG_ERR,
            "Can not find disk io_thread manager for %s.\n", filename);
        result = GlobusGFSErrorGeneric("mdtm reader IO manager not found.");
        goto error_alloc;
    }
    numa_of_file = reader_mgr->devio_info->numaid;

    // get network io manager
    sender_mgr = mdtm_io_manager(NULL, MDTM_DEVICE_CLASS_NETWORK);
    if ( sender_mgr == 0 ) {
        globus_gfs_log_message(
            GLOBUS_GFS_LOG_ERR,
            "Can not find network io_thread manager.\n");
        result = GlobusGFSErrorGeneric("mdtm network IO manager not found.");
        goto error_alloc;
    }

    // determine the parallel factor
    result = mdtm_l_get_parallel(sender_mgr, reader_mgr, dirq_info, size, &parallel, &segsize);
    if(result != GLOBUS_SUCCESS)
    	goto error_alloc;

#ifdef HAVE_SPLICE
    /*
     * LZ: Turn on splice for zero-copy data only if the flag is set
     * in the configure file AND the clients require transferring
     * data in non-encryption scheme.
     */
    //TODO: move init pipe pool to mdtm_pipe_get()
    //NOTE: 4* to get around the splice() memory issue
    if(globus_gfs_config_get_bool("splice") &&
        mdtm_gridftp_server_is_clientscheme_encrypted(op) == GLOBUS_FALSE) {
//      assert(mdtm_pipe_init(MDTM_SPLICE_PIPES_PER_NUMA, block_size) == 0);
      splice_on = GLOBUS_TRUE;
    }
#endif

#define MDTM_PIPE_MAXIMUM	(1000)
    if(splice_on == GLOBUS_TRUE) {
    	// estimate if the total number of pipes for splice exceeds the maximum
    	int allocated = mdtm_pipe_allocated(-1);
    	if(allocated + parallel * optimal_count > MDTM_PIPE_MAXIMUM) {
            globus_gfs_log_message(
                GLOBUS_GFS_LOG_ERR,
                "Requesting pipes: allocated %d, parallel %d, optimal_count %d, estimated total %d that exceeds the maximum %d.\n",
				allocated, parallel, optimal_count, allocated + parallel * optimal_count, MDTM_PIPE_MAXIMUM);
            result = GlobusGFSErrorGeneric("Requesting pipes that could exceed the maximum.");
            goto error_alloc;
    	}
    	else
            globus_gfs_log_message(
                GLOBUS_GFS_LOG_INFO,
                "Requesting pipes: allocated %d, parallel %d, optimal_count %d, estimated total %d.\n",
				allocated, parallel, optimal_count, allocated + parallel * optimal_count);
    }

    // schedule io threads
    nreader  = mdtm_io_thrd_mgr_sched(reader_mgr, parallel, &readers, 0);
    nsender  = mdtm_io_thrd_mgr_sched(sender_mgr, parallel, &senders, 0);
    assert(nreader>0 && nsender>0);
    globus_gridftp_server_set_iothrdmgr_attr(op, senders, nsender, sender_mgr);
    globus_gfs_log_message(
        GLOBUS_GFS_LOG_INFO,
        "%s: reads= %s %04x senders= %s %04x\n", __FUNCTION__,
        reader_mgr->devio_info->name, readers,
        sender_mgr->devio_info->name, senders);

    // create monitor arrays
    monitors = mdtm_monitor_pool_create(parallel);
    assert(monitors);

    // start data transfer
    transfer_arg = (mdtm_transfer_arg_t *)globus_malloc(sizeof(mdtm_transfer_arg_t));
    transfer_arg->op = (void*)op;
    transfer_arg->dispatcher = GLOBUS_FALSE;
    sem_init(&transfer_arg->sema, 0, 0);
    transfer_arg->monitors = (void*)monitors;
    mdtm_io_thrd_mgr_dispatch_to(
        sender_mgr,
        mdtm_util_nthsetbit(senders, 1) - 1,
        globus_gridftp_server_begin_transfer_mt,
        (void*)transfer_arg);
    sem_wait(&transfer_arg->sema);
    sem_destroy(&transfer_arg->sema);
    globus_free(transfer_arg);

    // create transmit progress
    if(mdtm_monitor_enabled()) {
        xmit_prog = mdtm_xmit_progress_alloc(((gfs_l_file_session_t *)user_arg)->username, op, 1);
        if(!xmit_prog)
          goto error_alloc;
    }
    else
      xmit_prog = 0;


    // creating monitors and starting to read
    monitor_q = (mdtm_monitor_queue_t*)globus_malloc(sizeof(mdtm_monitor_queue_t));
    globus_fifo_init(&monitor_q->q_);
    pthread_mutex_init(&monitor_q->lock, NULL);

    if(globus_gfs_config_get_bool("mdtm_debug"))
      gettimeofday(&start_stamp,NULL);

    // case #1/2: single file
    for(i = 0; i < parallel && !dirq_info; i++) {
        off = i * segsize;

        optimal_count = optimal_count_saved;
        result = globus_l_gfs_file_monitor_init2(
            &monitor, block_size, optimal_count, numa_of_file);
        if(result != GLOBUS_SUCCESS) {
            result = GlobusGFSErrorWrapFailed(
                "globus_l_gfs_file_monitor_init", result);
            goto error_alloc;
        }
        assert(mdtm_monitor_pool_add(monitors, monitor) == 0);

        if(splice_on == GLOBUS_TRUE) {
            rc = mdtm_pipemem_init(
                &monitor->pipemem,
                block_size,
                optimal_count * 2,
                numa_of_file); //get the pipes of numa
            if(!rc) {
//                globus_free(monitor);
                result = GlobusGFSErrorMemory("pipe");
                goto error_alloc;
            }
        }
        else {
            rc = globus_memory_init2(
                &monitor->mem,
                block_size,
                optimal_count,
                numa_of_file);
            if(!rc) {
//                globus_free(monitor);
                result = GlobusGFSErrorMemory("buffer");
                goto error_alloc;
            }
        }

        if(splice_on == GLOBUS_TRUE)
        {
    		globus_byte_t 	*buffer;

        	while(optimal_count--) {
        		buffer = mdtm_pipemem_pop_node(&monitor->pipemem);
        		if(buffer == NULL)
        			break;
        		globus_list_insert(&monitor->buffer_list, buffer);
        	}
        	if(optimal_count > 0) {
        		while(!globus_list_empty(monitor->buffer_list)) {
        			buffer = globus_list_remove(&monitor->buffer_list, monitor->buffer_list);
        			mdtm_pipemem_push_node(&monitor->pipemem, buffer);
        		}
        		result = GlobusGFSErrorMemory("pipe not allocated");
        		goto error_open;
        	}
        }
        else
          while(optimal_count--) {
              globus_byte_t *                 buffer;

              buffer = globus_memory_pop_node(&monitor->mem);
              globus_list_insert(&monitor->buffer_list, buffer);
          }

        monitor->splice_on = splice_on;
        monitor->session = (gfs_l_file_session_t *) user_arg;
        monitor->op = op;
        monitor->pathname = globus_libc_strdup(filename);
        monitor->multithrd_offset = off;
//        monitor->multithrd_length = ((i == (parallel-1) )? -1 : segsize);
        if(i == (parallel - 1))
            monitor->multithrd_length = size - i * segsize;
        else
            monitor->multithrd_length = segsize;
        monitor->mt_idx = i;
        monitor->reader_mgr = reader_mgr;
        monitor->iothreads_idx = mdtm_util_nthsetbit(readers, i + 1) - 1;
        monitor->parallel = parallel;
        monitor->parallel_max = parallel;
        monitor->monitor_q = monitor_q;
        monitor->read_length = -1;      //LZ: TODO: check again
        monitor->dirq_info = dirq_info;
        monitor->transfer_info = (void*)transfer_info;
        monitor->xmit_progress = xmit_prog;
        monitor->monitor_pool = (void*) monitors;

        open_flags = GLOBUS_XIO_FILE_BINARY | GLOBUS_XIO_FILE_RDONLY;
#if 0
        result = globus_l_gfs_file_open(
            &monitor->file_handle, filename, open_flags, monitor);
        if(result != GLOBUS_SUCCESS)
          {
            result = GlobusGFSErrorWrapFailed("globus_l_gfs_file_open", result);
            goto error_open;
          }
#else
        monitor->open_flags = open_flags;
        assert(
            mdtm_io_thrd_mgr_dispatch_to(
                reader_mgr,
                monitor->iothreads_idx,
                mdtm_l_gfs_file_open_kickout,
                monitor)
             >= 0);
#endif
    } // end of single file

    // case #2/2: directory
    if(dirq_info) {
    	pthread_mutex_lock(&dirq_info->lock);
    	filesegments = dirq_info->pending_segments;
    for(i = 0; i < parallel; i++) {
        optimal_count = optimal_count_saved;
        result = globus_l_gfs_file_monitor_init2(
            &monitor, block_size, optimal_count, numa_of_file);
        if(result != GLOBUS_SUCCESS) {
            result = GlobusGFSErrorWrapFailed(
                "globus_l_gfs_file_monitor_init", result);
            goto error_alloc;
        }

        assert(mdtm_monitor_pool_add(monitors, monitor) == 0);

        if(splice_on == GLOBUS_TRUE) {
            rc = mdtm_pipemem_init(
                &monitor->pipemem,
                block_size,
                optimal_count * 2,
                numa_of_file); //get the pipes of numa
            if(!rc) {
//                globus_free(monitor);
                result = GlobusGFSErrorMemory("pipe");
                goto error_alloc;
            }
        }
        else {
            rc = globus_memory_init2(
                &monitor->mem,
                block_size,
                optimal_count,
                numa_of_file);
            if(!rc) {
//                globus_free(monitor);
                result = GlobusGFSErrorMemory("buffer");
                goto error_alloc;
            }
        }

        if(splice_on == GLOBUS_TRUE)
        {
    		globus_byte_t	*buffer;

        	while(optimal_count--) {
        		buffer = mdtm_pipemem_pop_node(&monitor->pipemem);
        		if(buffer == NULL)
        			break;
        		globus_list_insert(&monitor->buffer_list, buffer);
        	}
        	if(optimal_count > 0) {
        		while(!globus_list_empty(monitor->buffer_list)) {
        			buffer = globus_list_remove(&monitor->buffer_list, monitor->buffer_list);
        			mdtm_pipemem_push_node(&monitor->pipemem, buffer);
        		}
        		result = GlobusGFSErrorMemory("pipe not allocated");
        		goto error_open;
        	}
        }
        else
          while(optimal_count--) {
              globus_byte_t *                 buffer;

              buffer = globus_memory_pop_node(&monitor->mem);
              globus_list_insert(&monitor->buffer_list, buffer);
          }

        monitor->session = (gfs_l_file_session_t *) user_arg;

        consumed = dirq_info->consumed;

        // check if there are pending file segments to transfer...
        if(filesegments->remain_segments == 0) {
            // no pending file segments, then look for the next file in the directory queue
            assert(mdtm_dir_queue_dequeue2(dirq_info->dir_q, filename));
            assert(lstat(filename, &sb) == 0);
            size = sb.st_size;

            // Check if it is a big file...
            if(size >= mdtm_config_get_maxfilesize()) {
                // see a big file, split into segments to transfer in parallel.
                long        count;
                size_t      sum;

                mdtm_util_round((long long)mdtm_config_get_maxfilesize(), (long long*)&filesegments->segment_size);
                for(count = 0, sum = 0; sum < size; count++) {
                    sum += filesegments->segment_size;
                }
                filesegments->total_segments =  count;
                filesegments->remain_segments = count - 1;
                filesegments->file_size = size;
                filesegments->completed_file_size = filesegments->segment_size;
                filesegments->consumed = consumed;
                filesegments->pathname = globus_libc_strdup(filename);

                dirq_info->consumed += filesegments->segment_size;
                monitor->consumed = filesegments->consumed;
                monitor->multithrd_offset = 0;
                monitor->multithrd_length = filesegments->segment_size;
            }
            else {
                // see a normal file, transfer the whole file.
                dirq_info->consumed += size;
                monitor->consumed = consumed;
                monitor->multithrd_offset = 0;
                monitor->multithrd_length = -1;
            }
        }
        else {
            // exist pending file segments, handle those first
            filesegments->remain_segments--;
            if(filesegments->remain_segments) {
                // not last segment
                dirq_info->consumed += filesegments->segment_size;
                monitor->multithrd_offset = filesegments->completed_file_size;
                monitor->multithrd_length = filesegments->segment_size;
                monitor->consumed = filesegments->consumed;
                filesegments->completed_file_size += filesegments->segment_size;
            }
            else {
                dirq_info->consumed += (filesegments->file_size - filesegments->completed_file_size);
                monitor->multithrd_offset = filesegments->completed_file_size;
                monitor->multithrd_length = -1;
                monitor->consumed = filesegments->consumed;
                filesegments->completed_file_size += filesegments->segment_size;
            }
        }
        // end of check

        monitor->splice_on = splice_on;
        monitor->op = op;
        monitor->pathname = globus_libc_strdup(filename);
//        monitor->multithrd_offset = 0;
//        monitor->multithrd_length = -1;
        monitor->mt_idx = i;
        monitor->reader_mgr = reader_mgr;
        monitor->iothreads_idx = mdtm_util_nthsetbit(readers, i + 1) - 1;
        monitor->parallel = parallel;
        monitor->parallel_max = parallel;
        monitor->monitor_q = monitor_q;
        monitor->monitor_pool = (void*) monitors;
        monitor->read_length = -1;      //LZ: TODO: check again
        monitor->dirq_info = dirq_info;
        monitor->transfer_info = (void*)transfer_info;
//        monitor->consumed = consumed;
        monitor->xmit_progress = xmit_prog;

        open_flags = GLOBUS_XIO_FILE_BINARY | GLOBUS_XIO_FILE_RDONLY;
#if 0
        result = globus_l_gfs_file_open(
            &monitor->file_handle, filename, open_flags, monitor);
        if(result != GLOBUS_SUCCESS)
          {
            result = GlobusGFSErrorWrapFailed("globus_l_gfs_file_open", result);
            goto error_open;
          }
#else
        monitor->open_flags = open_flags;
        if( mdtm_io_thrd_mgr_dispatch_to(
                reader_mgr,
                monitor->iothreads_idx,
                mdtm_l_gfs_file_open_kickout,
                monitor) < 0)
        {
        	free(monitor_q);
        	free(monitors);
            pthread_mutex_unlock(&dirq_info->lock);
        	globus_gfs_log_message(
        			GLOBUS_GFS_LOG_ERR,
					"%s: mdtm_io_thrd_mgr_dispatch_to failed.\n",__FUNCTION__);
        	goto error_open;
        }
#endif
    }
    pthread_mutex_unlock(&dirq_info->lock);
    } // end of directory
    
    //mdtmftp: start periodic progress monitor
    if(mdtm_monitor_enabled())
    {
      globus_reltime_t  period;
      globus_callback_handle_t *        callback_handle;

      callback_handle = (globus_callback_handle_t *) globus_calloc(1, sizeof(globus_callback_handle_t));
      if (callback_handle) {
    	  xmit_prog->callback_handle = (void *) callback_handle;
    	  xmit_prog->size = size;
    	  GlobusTimeReltimeSet(period, mdtm_monitor_getperiod(), 0);
    	  globus_callback_space_register_periodic(
    			  xmit_prog->callback_handle,
				  0,
				  &period,
				  mdtm_report_progress_kickout,
				  monitor,
				  mdtm_thread_get_space());
      }
      else {
    	  globus_gfs_log_message(
    			  GLOBUS_GFS_LOG_ERR,
				  "%s: memory allocation failed. can not start transfer monitor.\n",__FUNCTION__);
      }
    }
    // end of progress report

    GlobusGFSFileDebugExit();
    return;

error_open:
    globus_l_gfs_file_monitor_destroy(monitor);
    
error_alloc:  
    globus_gridftp_server_finished_transfer(op, result);

    GlobusGFSFileDebugExitWithError();
}

static
void
globus_l_gfs_file_event(
    globus_gfs_event_info_t *           event_info,
    void *                              user_arg)
{
    globus_l_file_monitor_t *           monitor;
    GlobusGFSName(globus_l_gfs_file_event);
    GlobusGFSFileDebugEnter();
#if 0
    monitor = (globus_l_file_monitor_t *) event_info->event_arg;

    switch(event_info->type)
    {
        case GLOBUS_GFS_EVENT_TRANSFER_ABORT:
            globus_mutex_lock(&monitor->lock);
            {
                monitor->aborted = GLOBUS_TRUE;
            }
            globus_mutex_unlock(&monitor->lock);
            
            globus_xio_handle_cancel_operations(
                monitor->file_handle,
                GLOBUS_XIO_CANCEL_OPEN | 
                GLOBUS_XIO_CANCEL_READ |
                GLOBUS_XIO_CANCEL_WRITE);
            break;
            
        default:
            break;
    }
#else
    // LZ: added to handle mdtm cases
    {
      mdtm_monitor_pool_t *           monitors;
      int                             i = 0;

      monitors = (mdtm_monitor_pool_t *) event_info->event_arg;
      monitor = mdtm_monitor_pool_entry(monitors, i);
      while(monitor) {
          switch(event_info->type)
          {
          case GLOBUS_GFS_EVENT_TRANSFER_ABORT:
            globus_mutex_lock(&monitor->lock);
            pthread_mutex_lock(&monitor->plock);
            {
              monitor->aborted = GLOBUS_TRUE;
            }
            pthread_mutex_unlock(&monitor->plock);
            globus_mutex_unlock(&monitor->lock);

//            globus_xio_handle_cancel_operations(
//                monitor->file_handle,
//                GLOBUS_XIO_CANCEL_OPEN |
//                GLOBUS_XIO_CANCEL_READ |
//                GLOBUS_XIO_CANCEL_WRITE);
            break;

          default:
            break;
          }
          monitor = mdtm_monitor_pool_entry(monitors, ++i);
      }
    }

#endif
    
    GlobusGFSFileDebugExit();
}

static
void
globus_l_gfs_file_init(
    globus_gfs_operation_t              op,
    globus_gfs_session_info_t *         session_info)
{
    gfs_l_file_session_t *              session_h;

    session_h = (gfs_l_file_session_t *) globus_calloc(
        1, sizeof(gfs_l_file_session_t));
    session_h->cred = session_info->del_cred;
    session_h->sbj = globus_libc_strdup(session_info->subject);
    session_h->username = globus_libc_strdup(session_info->username);
    session_h->pw = globus_libc_strdup(session_info->password);

    /* just make it so we can get the cred. */
    globus_gridftp_server_finished_session_start(
        op,
        GLOBUS_SUCCESS,
        session_h,
        NULL,
        NULL);
}

static
void
globus_l_gfs_file_destroy(
    void *                              user_arg)
{
    gfs_l_file_session_t *              session_h;
    session_h = (gfs_l_file_session_t *) user_arg;

    if(session_h)
    {
        if(session_h->sbj != NULL)
        {
            globus_free(session_h->sbj);
        }
        if(session_h->username != NULL)
        {
            globus_free(session_h->username);
        }
        if(session_h->pw != NULL)
        {
            globus_free(session_h->pw);
        }
        globus_free(session_h);
    }
}

static
int
globus_l_gfs_file_activate(void);

static
int
globus_l_gfs_file_deactivate(void);

static globus_gfs_storage_iface_t       globus_l_gfs_file_dsi_iface = 
{
    GLOBUS_GFS_DSI_DESCRIPTOR_SENDER | GLOBUS_GFS_DSI_DESCRIPTOR_HAS_REALPATH,
    globus_l_gfs_file_init,
    globus_l_gfs_file_destroy,
    NULL, /* list */
    globus_l_gfs_file_send,
    globus_l_gfs_file_recv,
    globus_l_gfs_file_event, /* trev */
    NULL, /* active */
    NULL, /* passive */
    NULL, /* data destroy */
    globus_l_gfs_file_command, 
    globus_l_gfs_file_stat,
    NULL,
    NULL,
    globus_l_gfs_file_realpath
};

GlobusExtensionDefineModule(globus_gridftp_server_file) =
{
    "globus_gridftp_server_file",
    globus_l_gfs_file_activate,
    globus_l_gfs_file_deactivate,
    NULL,
    NULL,
    &local_version
};

static
int
globus_l_gfs_file_activate(void)
{
    if(globus_module_activate(GLOBUS_XIO_MODULE) != GLOBUS_SUCCESS)
    {
        goto error_activate;
    }
    
    if(globus_xio_driver_load(
        "file", &globus_l_gfs_file_driver) != GLOBUS_SUCCESS)
    {
        goto error_load_file;
    }
    
    globus_extension_registry_add(
        GLOBUS_GFS_DSI_REGISTRY,
        "file",
        GlobusExtensionMyModule(globus_gridftp_server_file),
        &globus_l_gfs_file_dsi_iface);

    GlobusDebugInit(GLOBUS_GRIDFTP_SERVER_FILE,
        ERROR WARNING TRACE INTERNAL_TRACE INFO STATE INFO_VERBOSE);
    
    return GLOBUS_SUCCESS;
    
error_load_file:
    globus_module_deactivate(GLOBUS_XIO_MODULE);
    
error_activate:
    return GLOBUS_FAILURE;
}

static
int
globus_l_gfs_file_deactivate(void)
{
    globus_extension_registry_remove(
        GLOBUS_GFS_DSI_REGISTRY, "file");
        
    globus_xio_driver_unload(globus_l_gfs_file_driver);
    
    return globus_module_deactivate(GLOBUS_XIO_MODULE);
}
